Require Export HLLLanguage. 
Require Export SemiLat. 

Parameter L:Type. 
Axiom LatL : lattice L. 
Definition FactsL := (@facts _ LatL).

Definition bbot := (@bot L LatL). 
Definition llub := @lub L LatL. 
Definition ssqsub := @leq L LatL. 
Definition llat_bot := @bot L LatL. 
Definition llub_refl := @lub_refl L LatL. 
Definition llub_com := lub_com FactsL.
Definition ssqsub_refl:= @leq_refl L LatL.
Definition ssqsub_trans := @leq_trans L LatL.
Definition llub_sqleft := @lub_leqleft L LatL.
Definition llub_assoc := lub_assoc FactsL.
Definition ssqsub_monoright := @leq_monoright L LatL.
Definition ssq_lubright := @leq_lubright L LatL.
Definition id_bbot := @id_bot L LatL.
Definition bbot_id := bot_id FactsL.
Definition ccompat_llub := compat_lub FactsL. 

Definition CTXT: Type := PVar -> L.
(*
Definition CTXT_MAP : MAP PVar pvar_eq L upd.
  constructor. unfold upd. trivial. Defined.
Existing Instance CTXT_MAP.
*)
(*
Lemma upd_split_at :
forall (s:CTXT) x v y u, (x = y -> v = u) ->
                  (x <> y -> s y = u) ->
                   upd s x v y = u.
apply M_upd_split_at. Qed.
 *)

Definition LatCTXTfacts := funspace_facts PVar (@facts L LatL). 

Definition LatCTXT : lattice CTXT := Lattice LatCTXTfacts. 

Definition BBOT := @bot CTXT LatCTXT. 
Definition LUB := @lub CTXT LatCTXT. 
Definition SQSUB := @leq CTXT LatCTXT. 

Fixpoint GEval (G:CTXT)(E: Expr): L :=
match E with
  valExpr v => (@bot L LatL)
| varExpr x => G x
| bopExpr e1 b e2 => llub (GEval G e1) (GEval G e2)
end. 

Lemma GEval_mono: forall G G' E, SQSUB G G' -> 
ssqsub (GEval G E) (GEval G' E). 
Proof. 
intros. 
induction E; simpl in *. 
  apply ssqsub_refl. 
  apply H. 
eapply ssqsub_trans. 
  apply ssqsub_monoright.  
  apply IHE2. 
  eapply leq_monoleft.  apply IHE1. 
Qed. 

Lemma GEval_CtxtEq: forall e G H, G=H -> GEval G e = GEval H e.
Proof.
intros; subst.
trivial.
Qed. 

Lemma GEval_sep: forall e (G H:CTXT),
GEval (fun x : PVar => llub (G x) (H x)) e = llub (GEval G e) (GEval H e).
Proof.
intros e.
induction e; intros; simpl in *. rewrite bbot_id. trivial.
  reflexivity.
rewrite IHe1. rewrite IHe2.
  rewrite <- llub_assoc. 
  rewrite <- llub_assoc.
  rewrite (llub_assoc (GEval H e1)).
  rewrite (llub_assoc (GEval G e2)).
  rewrite (llub_com (GEval G e2)).
  trivial.
Qed.

Lemma GEval_sep': forall e (G H:CTXT),
GEval (fun x : PVar => @lub L LatL (G x) (H x)) e = @lub L LatL (GEval G e) (GEval H e).
Proof.
intros.
apply GEval_sep.
Qed.

Lemma lubLUB: forall G H, (fun x : PVar => @lub L LatL (G x) (H x)) =
             (fun x => (LUB G H) x).
Proof.
intros.
unfold LUB. apply extensionality. intros.
reflexivity.
Qed.

Lemma LUB_ext: forall G H x, (@lub CTXT LatCTXT G H) x =
             (@lub L LatL (G x) (H x)).
Proof.
intros. reflexivity.
Qed.

Lemma Zero_max_elim: forall (n n':nat), 0 = MinMax.max n n' -> n=0 /\ n'=0.
Proof.
intros.
destruct n; destruct n'; simpl in H; try inversion H.
  split; trivial.
Qed.   