Require Export CTXT. 

Inductive Var :=
  pcVar : Var
| progVar: PVar -> Var. 

Definition var_eq: forall (x y: Var), {x=y} + {x <> y}. 
Proof. 
intros. 
destruct x; destruct y.  
left; trivial. 
right; intros H.  inversion H. 
right; intros H.  inversion H. 
destruct (pvar_eq p p0); subst. 
  left; trivial. 
  right; intros N.  inversion N; subst.  apply n.  trivial. 
Defined. 


Definition var_upd {A:Type} 
  (X:Var -> A) (x:Var) (a:A) : Var -> A :=
fun y => if var_eq x y then a else X y.

Definition VAR_MAP {A:Type}: MAP Var var_eq A var_upd.
  constructor. unfold var_upd. trivial. Defined.
Existing Instance VAR_MAP.

Lemma var_eq_ElimT: forall x p (H:(if var_eq p x then true else false) = true), p=x.
Proof.
 intros.
 remember (var_eq p x) as b; destruct b; subst; trivial. inversion H.
Qed.  
Hint Resolve var_eq_ElimT: varREWRITE.

Lemma var_eq_ElimF: forall x p (H:(if var_eq p x then true else false) = false), p<>x.
Proof.
 intros.
 remember (var_eq p x) as b; destruct b; subst; trivial. inversion H.
Qed. 
Hint Resolve var_eq_ElimF: varREWRITE. 

(*making parameter A implicit here would prevent me from 
  adding the lemma to the hint database polymorphically*)
Lemma var_eq_refl: forall A z (a b:A), 
  (if var_eq z z then a else b) = a.
Proof. intros.
  remember (var_eq z z) as bb; destruct bb; subst; trivial. 
  apply False_ind.  apply n.  trivial.
Qed. 
Hint Rewrite var_eq_refl:varREWRITE.

Lemma var_pvar_eq: forall A x z (a1 a2:A),
  (if var_eq (progVar x) (progVar z) then a1 else a2) =
  (if pvar_eq x z then a1 else a2).
Proof.
intros.
  remember (var_eq (progVar x) (progVar z)) as b; destruct b; subst; simpl.
  inversion e; subst. rewrite pvar_eq_refl. trivial.
remember (pvar_eq x z) as bb; destruct bb; subst; simpl; trivial.
  apply False_ind.  apply n.  trivial. 
Qed.
Hint Rewrite var_pvar_eq:varREWRITE.

Definition FPCCTXT:Type:= Var -> L. 

Definition LatFPCCTXTfacts := funspace_facts Var (@facts L LatL). 

Definition LatFPCCTXT : lattice FPCCTXT := Lattice LatFPCCTXTfacts. 

(*Notation "'fpcupd' G x l" := (@var_upd L G x l) (at level 200).*)
(*
Definition fpcupd (G:FPCCTXT)(x:Var)(l: L): FPCCTXT :=
fun y => if var_eq x y then l else G y.
*)
Definition fpcupd := @var_upd L.

Lemma fpcupd_Same: forall G x v, fpcupd G x v x = v. 
Proof. intros. apply M_upd_Same. Qed. 
Hint Rewrite fpcupd_Same : fpcREWRITE.

Goal forall G x v, fpcupd G x v x = v.
Proof with (autorewrite with fpcREWRITE using try reflexivity).
intros... 
Qed.

Lemma fpcupd_Diff: forall G x v y, x <> y -> fpcupd G x v y = G y. 
Proof. apply M_upd_Diff. Qed. 
(*Hint Rewrite fpcupd_Diff : fpcREWRITE.*)

Lemma fpcupd_split: forall s x v t, 
t x = v ->
(forall y, x <> y -> s y = t y) ->
fpcupd s x v = t.
Proof. apply M_upd_split. Qed.

Lemma fpcupd_split_at: forall s x v y u, 
(x = y -> v = u) ->
(x <> y -> s y = u) ->
fpcupd s x v y = u.
Proof. apply M_upd_split_at. Qed.

Lemma fpcupd_split_at_var1: forall s v y, 
fpcupd s pcVar v (progVar y) = s (progVar y).
Proof.
intros. (* apply M_upd_split_at.*)
apply fpcupd_split_at; intros; trivial.
inversion H.
Qed.

Lemma fpcupd_split_at_var2: forall s x v y u, 
(x = y -> v=u) ->
( x <> y -> s (progVar y) = u) ->
fpcupd s (progVar x) v (progVar y) = u.
Proof.
intros.
apply fpcupd_split_at; intros; trivial.
  inversion H1; subst. apply H. trivial.
apply H0. intros N. apply H1. subst. trivial.
Qed.

Lemma fpcupd_fpcupd: forall s x v u, 
  fpcupd (fpcupd s x u) x v = fpcupd s x v.
Proof. apply M_upd_upd. Qed. 
Hint Rewrite fpcupd_fpcupd : fpcREWRITE.

Lemma fpcupd_fpcupd_swap: forall s x y u v (XY:  x <> y), 
fpcupd (fpcupd s x u) y v = fpcupd (fpcupd s y v) x u.
Proof. apply M_upd_upd_swap. Qed. 

Inductive FPCExpr :=
  pcExpr: FPCExpr
| progExpr : Expr -> FPCExpr. 

Fixpoint FPCGEval (G:FPCCTXT) (E:FPCExpr): L :=
match E with
  pcExpr => G pcVar
| progExpr e => GEval (fun x => G (progVar x)) e
end. 

Definition mkFPCctxt (G:CTXT) (p:L) : FPCCTXT :=
fun y => match y with pcVar => p
                    | progVar x => G x
         end. 

Lemma mkFPCctxt_pcVar: forall G p, mkFPCctxt G p pcVar = p.
Proof.
intros. reflexivity. 
Qed.
  
Lemma mkFPCctxt_progVar: forall G p x, mkFPCctxt G p (progVar x) = G x.
Proof.
intros.
reflexivity.
Qed.

Lemma fpcupd_pcVar: forall G p q, 
     fpcupd (mkFPCctxt G p) pcVar q = mkFPCctxt G q.
Proof.
intros.
apply extensionality. intros x.
destruct x. rewrite fpcupd_Same. reflexivity. 
rewrite fpcupd_Diff. reflexivity.
intros Q. inversion Q.
Qed.
Hint Rewrite fpcupd_pcVar: fpcREWRITE.

Definition FSQSUB := @leq FPCCTXT LatFPCCTXT. 

Definition FPCCTXT2CTXT (G:FPCCTXT): CTXT := fun x => G (progVar x). 

Definition FLUB := @lub FPCCTXT LatFPCCTXT. 

Lemma FLUB_leqleft: forall G H x,  @leq L LatL (G x) (FLUB G H x).
Proof.
intros.
unfold FLUB; simpl.
apply lub_leqleft.
Qed.

Lemma FLUB_leqright: forall G H x,  @leq L LatL (H x) (FLUB G H x).
Proof.
intros.
unfold FLUB; simpl.
apply lub_leqright.
Qed.

Lemma FLUB_ext: forall G H x, (@lub FPCCTXT LatFPCCTXT G H) x =
             (@lub L LatL (G x) (H x)).
Proof.
intros. reflexivity.
Qed.

Lemma mkCtxtInverse: forall G, (mkFPCctxt (FPCCTXT2CTXT G) (G pcVar)) = G. 
Proof. 
intros. 
apply extensionality.  intros z. 
destruct z; trivial. 
Qed. 
Hint Rewrite mkCtxtInverse: fpcREWRITE.

Lemma mkCtxtInverse2: forall G p, 
  FPCCTXT2CTXT (mkFPCctxt G p) = G. 
Proof. 
intros. 
apply extensionality.  intros z.  trivial. 
Qed.
Hint Rewrite mkCtxtInverse2: fpcREWRITE.

Lemma GEval_noPcVar: forall G E p, 
 FPCGEval (mkFPCctxt G p) (progExpr E) = GEval G E. 
Proof. 
intros. 
simpl. 
assert ((fun x : PVar => G x) = G). 
  apply extensionality.  trivial. 
rewrite H; trivial. 
Qed.
Hint Rewrite GEval_noPcVar: fpcREWRITE.

Lemma FPCCTXT2CTXT_upd_pcVar: forall G p,
 FPCCTXT2CTXT (fpcupd G pcVar p) = (FPCCTXT2CTXT G). 
Proof. 
intros. auto. 
Qed. 
Hint Rewrite FPCCTXT2CTXT_upd_pcVar: fpcREWRITE.

Lemma FPCCTXT2CTXT_upd_progVar: forall G x p,
 FPCCTXT2CTXT (fpcupd G (progVar x) p) = 
 upd (FPCCTXT2CTXT G) x p. 
Proof. 
intros. 
unfold FPCCTXT2CTXT.
apply extensionality.  intros z.
apply eq_sym.
apply M_upd_split_at; intros.
  subst. rewrite M_upd_Same. trivial.
  rewrite M_upd_Diff; trivial. congruence. 
Qed.  
Hint Rewrite FPCCTXT2CTXT_upd_progVar: fpcREWRITE.

Lemma fcpupd_pcVar_SQSUB: forall G H p,
   SQSUB G (FPCCTXT2CTXT H) ->
   SQSUB G (FPCCTXT2CTXT(fpcupd H pcVar p)). 
Proof.
 auto. 
Qed. 

Lemma Geval_mono2: forall G G' e (HG: SQSUB (FPCCTXT2CTXT G) (FPCCTXT2CTXT G'))
(Hp: ssqsub (G pcVar) (G' pcVar)),
ssqsub (llub (G pcVar) (GEval (fun x0 : PVar => G (progVar x0)) e))
  (llub (G' pcVar) (GEval (fun x0 : PVar => G' (progVar x0)) e)). 
Proof. 
intros. 
eapply ssqsub_trans. 
        eapply leq_monoleft.  apply Hp. 
        eapply ssqsub_monoright. 
        eapply GEval_mono.  apply HG. 
Qed. 

Inductive FPC : FPCCTXT -> Com -> FPCCTXT -> Prop :=
  FPC_Skip: forall G, FPC G Skip G
| FPC_Assign: forall p x E G H, 
         p = G pcVar ->
         H = fpcupd G (progVar x) (llub p (FPCGEval G (progExpr E))) ->
         FPC G (Assign x E) H
| FPC_Comp: forall G G1 G2 C1 C2, 
         FPC G C1 G1 -> FPC G1 C2 G2 ->
         FPC G (Comp C1 C2) G2
| FPC_Ite: forall p G GG G1 G2 E C1 C2 a H HH,
         a = FPCGEval G (progExpr E) ->
         p = G pcVar ->
         GG = fpcupd G pcVar (llub p a) ->
         FPC GG C1 G1 -> 
         FPC GG C2 G2 ->
         H = FLUB G1 G2 ->
         HH = fpcupd H pcVar p ->
         FPC G (Ite E C1 C2) HH
| FPC_While: forall p G E C H N
         (Hp: p = (G pcVar))
         (pre post : nat -> FPCCTXT)
         (PreBelowN: forall i, i < N -> (pre i) <> pre (S i))
         (PreAboveN: forall i, N <= i -> (pre i) = pre (S i))
         (PreZ: pre 0 = fpcupd G pcVar (llub p (FPCGEval G (progExpr E))))
         (PreN: fpcupd (pre N) pcVar p = H)
         (PrePost: forall i, pre (S i) = fpcupd (FLUB (post i) G) pcVar 
                                (llub ((FLUB (post i) G) pcVar) (FPCGEval (FLUB (post i) G) (progExpr E))))
         (HypC: forall i,  FPC (pre i) C (post i)),
         FPC G (While E C) H.

Lemma FPC_unique: forall C G H (Hyp: FPC G C H) H'
 (Hyp': FPC G C H'), H=H'. 
Proof. 
intros C. 
induction C; intros. 
(*Skip*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. trivial.
(*Assign*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. trivial. 
(*Comp*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. 
    assert (IH1 := IHC1 _ _ H3 _ H4). rewrite IH1 in *.
    apply (IHC2 _ _ H5 _ H7). 
(*If*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.
    assert (IH1 := IHC1 _ _ H6 _ H7).  rewrite IH1 in *.
    assert (IH2 := IHC2 _ _ H10 _ H9). rewrite IH2 in *. trivial.
(*While*) 
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.
    assert (PRE: forall i, N <= i -> pre N = pre i).
      intros i. induction i; intros. destruct N. trivial. apply False_ind. eapply Le.le_Sn_0. apply H.
      destruct N. rewrite IHi. apply PreAboveN. apply Le.le_0_n. apply Le.le_0_n.
      apply Le.le_S_n in H. 
      destruct (Lt.le_lt_or_eq _ _ H). 
         apply Lt.lt_le_S in H0. rewrite (IHi H0).  apply PreAboveN. apply H0. 
         rewrite H0. trivial. 
    assert (PRE': forall i, N0 <= i -> pre0 N0 = pre0 i).
      intros i. induction i; intros. destruct N0. trivial. apply False_ind. eapply Le.le_Sn_0. apply H.
      destruct N0. rewrite IHi. apply PreAboveN0. apply Le.le_0_n. apply Le.le_0_n.
      apply Le.le_S_n in H. 
      destruct (Lt.le_lt_or_eq _ _ H). 
         apply Lt.lt_le_S in H0. rewrite (IHi H0).  apply PreAboveN0. apply H0. 
         rewrite H0. trivial. 
  assert (forall i, (pre i) = (pre0 i) /\ (post i) = (post0 i)).
    intros i.
    induction i; intros; subst; simpl.
      assert (pre 0 = pre0 0).
        rewrite PreZ. rewrite PreZ0. trivial.
      split. assumption.
      apply (IHC (pre 0)). apply HypC. rewrite H. apply HypC0.
    destruct IHi as [IHpre IHpost].
      assert (pre (S i) = pre0 (S i)).
        rewrite PrePost. rewrite PrePost0. repeat rewrite IHpost. trivial.
      split. assumption.
      apply (IHC (pre (S i))). apply HypC. rewrite H. apply HypC0.
  remember (Compare_dec.le_lt_dec N N0) as b; destruct b.
     rewrite (PRE _ l). destruct (H N0). rewrite H0. reflexivity.
     destruct N0. rewrite (PRE' N). destruct (H N). rewrite <- H0. trivial. apply Le.le_0_n.
     clear Heqb.  rewrite (PRE' N). destruct (H N). rewrite <- H0. trivial. apply Lt.lt_le_weak. apply l.
Qed.  

Lemma FPC_Lemma1: forall G C G'
  (Hyp: FPC G C G'),  G pcVar = G' pcVar. 
Proof. 
intros G C. generalize dependent G. 
induction C; simpl; intros; inversion Hyp; clear Hyp; subst; trivial. 
(*Comp*)
rewrite (IHC1 _ _ H2). 
rewrite (IHC2 _ _ H4).
trivial.
Qed.

Lemma FLUB_upd_unfold: forall G H x a y,
fpcupd (FLUB G H) x a y = llub (fpcupd G x a y) (fpcupd H x a y).
Proof.
intros.
apply M_upd_split_at; intros; subst.
  repeat rewrite M_upd_Same. rewrite llub_refl. trivial.
repeat rewrite M_upd_Diff; try assumption. reflexivity.
Qed.

Lemma FPC_mono: forall C G H (Hyp: FPC G C H) G' H'
 (Hyp': FPC G' C H') 
 (HG:FSQSUB G' G), FSQSUB H' H. 
Proof with (autorewrite with fpcREWRITE using try reflexivity).
intros C. 
induction C; intros. 
(*Skip*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.  assumption. 
(*Assign*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.  intros y.
  remember (var_eq (progVar p) y) as xy; destruct xy; subst; simpl in *.
    (autorewrite with fpcREWRITE using try reflexivity).
      eapply ssqsub_trans. 
        apply leq_monoleft.  apply HG. 
        apply ssqsub_monoright. 
          apply GEval_mono. intros y. apply HG.
    repeat rewrite fpcupd_Diff; try assumption. 
      apply HG. 
(*Comp*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. 
    apply (IHC2 _ _ H5 _ _ H7). 
    apply (IHC1 _ _ H3 _ _ H4 HG). 
(*If*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.
   assert (Q: FSQSUB (fpcupd G' pcVar (llub (G' pcVar) (FPCGEval G' (progExpr e))))
                     (fpcupd G pcVar (llub (G pcVar) (FPCGEval G (progExpr e))))).
       clear IHC1 IHC2. intros x.            
       destruct x; simpl.
         (autorewrite with fpcREWRITE using try reflexivity).   
          eapply leq_trans.   
          apply leq_monoleft. apply HG.
          apply leq_monoright. 
          apply GEval_mono. intros y.  apply HG.
        repeat rewrite M_upd_Diff; try discriminate.
          apply HG.  
      
    assert (HH1 := IHC1 _ _ H7 _ _ H6 Q).  clear IHC1. 
    assert (HH2 := IHC2 _ _ H9 _ _ H10 Q).  clear IHC2. 
    intros x. simpl.
    repeat rewrite FLUB_upd_unfold.
    destruct x; simpl.
      repeat rewrite fpcupd_Same. repeat rewrite llub_refl. apply HG.
      repeat rewrite fpcupd_Diff; try discriminate.
        apply leq_mono. apply HH1.  apply HH2.
(*While*)
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. 
  generalize dependent post0. intro post'.
  generalize dependent pre0. intro pre'.
  generalize dependent N0. intro N'.
  intros.
  remember (MinMax.min N N') as M.
  assert (forall j, le j M -> FSQSUB (pre' j) (pre j)).
    intros j. induction j; intros.
       (*apply HG. *) rewrite PreZ. rewrite PreZ0.
           intros y. destruct y. repeat rewrite fpcupd_Same. eapply leq_trans. apply leq_monoleft. apply HG. apply leq_monoright. apply GEval_mono. intros z. apply HG.
           repeat rewrite fpcupd_Diff. apply HG. discriminate. discriminate.
       assert (KK: j <= M). eapply Le.le_trans. apply Le.le_n_Sn. apply H.
       assert (KN: j <= N).
         eapply Le.le_trans. apply KK. rewrite HeqM. apply Min.le_min_l.
       rewrite (PrePost0 j).
       rewrite (PrePost j).
       assert (POST:  FSQSUB (post' j) (post j)).
         eapply IHC. apply (HypC j). apply (HypC0 j). apply IHj. apply KK.
       assert (PG: FSQSUB (FLUB (post' j) G') (FLUB (post j) G)).
         apply leq_mono; assumption.
       intros x. destruct x. repeat rewrite fpcupd_Same. 
                    eapply leq_trans. apply leq_monoleft. apply PG.
                      apply leq_monoright. apply GEval_mono. intros z. apply PG.
                   repeat rewrite fpcupd_Diff. apply PG.
                   discriminate. discriminate.
    assert (EQMin := H M (Le.le_refl _)). clear H.
    destruct (Min.min_dec N N').
    (*min N N' = N*)
     rewrite e0 in HeqM. rewrite HeqM in *. clear HeqM M. 
    assert (K: forall j, N <= j -> pre j = pre N).
      intros. (*clear PreBelowN0. PreZ0*) clear PreAboveN0  PrePost0 HypC0 IHC.
            induction j. 
              assert (XX := Le.le_n_0_eq _ H). rewrite <- XX. trivial.
            destruct (Lt.le_lt_or_eq_iff N (S j)) as [X1 _].
            destruct (X1 H).
              assert (pre j = pre (S j)). apply PreAboveN. apply Le.le_S_n. apply H0. 
              rewrite <- H1. apply IHj. apply Le.le_S_n. apply H0.
             subst. trivial.
       assert (N <= N'). apply (MinMax.MMP.min_l_iff N N'). apply e0.
       rewrite <- (K _ H).
     cut (forall j, N <= j -> j <= N' -> FSQSUB (pre' j) (pre N)).
       intros HH. rewrite (K N').
       intros y.
         destruct  y; simpl. 
         repeat rewrite fpcupd_Same. apply HG.
         repeat rewrite fpcupd_Diff. apply HH. apply H. apply Le.le_refl.
          intros XX; inversion XX. intros XX; inversion XX.
        apply H.
     intros j.
     induction j; intros Nj N'j.
       destruct N. apply EQMin. (*HG rewrite PreZ. rewrite PreZ0. apply leq_refl.*) inversion Nj.
       destruct (Compare_dec.le_lt_eq_dec _ _ Nj).
       (*N < S j*) 
         rewrite (PrePost0 j). 
         rewrite (PreAboveN N (Le.le_refl _)).
         rewrite (PrePost N). 
         assert (POSTG_jN: FSQSUB (FLUB (post' j) G') (FLUB (post N) G)).
            apply leq_mono; try assumption. eapply IHC. apply (HypC N). apply (HypC0 j). apply IHj. apply Le.le_S_n. apply l.
                            eapply Le.le_trans. apply Le.le_n_Sn. apply N'j.
         intros z. destruct z. repeat rewrite fpcupd_Same.  
            eapply leq_trans. eapply leq_monoleft.  apply POSTG_jN. 
                              eapply leq_monoright. apply GEval_mono. intros z. apply POSTG_jN.
             repeat rewrite fpcupd_Diff; try discriminate. apply POSTG_jN.
      (*N = S j*) rewrite e1 in *. 
        clear IHj e1 Nj N'j. apply EQMin.
    (*min N N' = N'*) 
    assert (K': forall j, N' <= j -> pre' N' = pre' j).
      intros. clear IHC HG.
            induction j. 
              assert (XX := Le.le_n_0_eq _ H). rewrite <- XX. trivial.
            destruct (Lt.le_lt_or_eq_iff N' (S j)) as [X1 _].
            destruct (X1 H).
              rewrite <- (PreAboveN0 j). apply IHj. apply Le.le_S_n. apply H0.
              apply Le.le_S_n. apply H0. 
             subst. trivial.
    assert (N'=M). rewrite <- e0. rewrite HeqM. trivial. rewrite H in *. clear H.
    assert (M <= N). apply MinMax.MMP.min_r_iff. apply e0. clear HeqM e0.
    rewrite (K' _ H).
    assert (K: forall j, j <= N -> FSQSUB (pre j) (pre (S j))).
      intros. (*clear PreBelowN0 PreZ0.*) clear HG PreAboveN0 PrePost0 HypC0.
            induction j. (*rewrite PreZ. apply bot_leq.*)
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ (HypC 0)).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). apply FLUB_leqleft.
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff; try discriminate. apply FLUB_leqright.
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply IHC. apply HypC. apply HypC. apply IHj. eapply Le.le_trans. apply Le.le_n_Sn. apply H0.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                          eapply leq_trans. apply leq_monoright. apply GEval_mono. intros z. apply POST_FL.
                          apply leq_monoleft. apply POST_FL.
                      repeat rewrite fpcupd_Diff. apply POST_FL.
                        discriminate. discriminate. 
    cut (forall j, M <= j -> j <= N -> FSQSUB (pre M) (pre j)).
      intros KK.
              intros y.
              destruct y. 
                repeat rewrite fpcupd_Same. apply HG.
              repeat rewrite fpcupd_Diff; try discriminate. 
                 rewrite <- K'. eapply leq_trans. 
                   apply EQMin. apply KK.  apply H. apply Le.le_refl. 
                apply H. 
    intros j.
    induction j; intros HjM HjN. 
    destruct M. apply leq_refl. inversion HjM.
    destruct (Compare_dec.le_lt_eq_dec M (S j) HjM).
      assert (FSQSUB (pre M) (pre j)). apply IHj. apply Le.le_S_n. apply l.
         eapply Le.le_trans. apply Le.le_n_Sn.  apply HjN.
      eapply leq_trans. apply H0.   apply K. eapply Le.le_trans. apply Le.le_n_Sn.  apply HjN.
    rewrite e0. apply leq_refl.
Qed.

Lemma FPC_Lemma2: forall G C H (Hyp: FPC G C H) x,
  ssqsub (H pcVar) (H x) \/ (H x = G x). 
Proof. 
intros G C. generalize dependent G.
induction C; intros; inversion Hyp; clear Hyp; intros; subst; simpl in *.
(*Skip*)
right; trivial.
(*Assign*) 
destruct x. 
rewrite (fpcupd_Diff G (progVar p) _ pcVar).  right; trivial. 
 intros HH.  inversion HH.
remember (var_eq (progVar p) (progVar p0)) as bb; destruct bb; subst; simpl.  rewrite e0 in *. 
   rewrite fpcupd_Same. 
   rewrite (fpcupd_Diff G (progVar p0) _ pcVar). 
     left.  apply llub_sqleft. 
    intros HH.  inversion HH.  
  rewrite (fpcupd_Diff G (progVar p) _ (progVar p0)). 
    rewrite (fpcupd_Diff G (progVar p) _ pcVar).  right; trivial. 
    intros HH.  inversion HH. 
  assumption.
(*Comp*) 
  destruct (IHC2 _ _ H5 x). 
    left; trivial. 
  rewrite H0. 
  rewrite <- (FPC_Lemma1 _ _ _ H5).
  apply IHC1.  assumption.
(*If*)
rewrite fpcupd_Same.
  assert (L1:= FPC_Lemma1 _ _ _ H7). 
  assert (L2 := FPC_Lemma1 _ _ _ H9). 
  destruct x. 
  rewrite (fpcupd_Same _ pcVar _) in *. 
    right.  trivial. 
  rewrite (fpcupd_Diff _ pcVar _ (progVar p)) in *; try discriminate. 
    rewrite (fpcupd_Same _ pcVar _) in L1. 
    rewrite (fpcupd_Same _ pcVar _) in L2.
    destruct (IHC1 _ _ H7 (progVar p)).
       rewrite <- L1 in H.
       destruct (lub_leq_elim _ _ _ H).  
       left. eapply ssqsub_trans.  apply H0.  apply FLUB_leqleft.
    unfold FLUB; simpl. 
    rewrite (fpcupd_Diff _ pcVar _ (progVar p)) in H; try discriminate.  rewrite H in *.   clear H. 
    destruct (IHC2 _ _ H9 (progVar p)).  
       rewrite <- L2 in H.
       destruct (lub_leq_elim _ _ _ H).  
       left. eapply ssqsub_trans.  apply H0.  apply lub_leqright.
    rewrite (fpcupd_Diff _ pcVar _ (progVar p)) in H; try discriminate.  rewrite H in *.   clear H. 
     right.  apply llub_refl. 
(*While*)
rewrite fpcupd_Same.
assert (INCPre: forall i, FSQSUB (pre i) (pre (S i))).
  intros j. induction j. (*rewrite PreZ. apply bot_leq.*)
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ (HypC 0)).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). apply FLUB_leqleft.
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff; try discriminate. apply FLUB_leqright. 
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply FPC_mono. apply HypC. apply HypC. apply IHj.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                         apply leq_mono.
                           apply POST_FL.
                           apply GEval_mono. intros z. apply POST_FL. 
                      repeat rewrite fpcupd_Diff; try discriminate. apply POST_FL.
 assert (forall i , FSQSUB (pre 0) (pre i)).
   intros i. 
   induction i.  apply leq_refl.
   eapply leq_trans. apply IHi. apply INCPre.

 assert (INCPost: forall i, FSQSUB (post i) (post (S i))).
   intros i. eapply FPC_mono. apply HypC. apply HypC. apply INCPre.

  cut (forall i, ssqsub (pre 0 pcVar) (fpcupd (pre i) pcVar (pre 0 pcVar) x) \/
          (fpcupd (pre i) pcVar (pre 0 pcVar) x = (pre 0) x)).
    intros GEN. rewrite PreZ in GEN. 
    destruct (GEN N); clear GEN. rewrite fpcupd_Same in H0.
          destruct (lub_leq_elim _ _ _ H0).
           destruct x. rewrite fpcupd_Same in *. right. trivial. 
           repeat rewrite fpcupd_Diff. rewrite fpcupd_Diff in H1. left. apply H1.
           discriminate. discriminate.
    destruct x. rewrite fpcupd_Same in *. right. trivial.
           repeat rewrite fpcupd_Diff. rewrite fpcupd_Diff in H0. rewrite fpcupd_Diff in H0. right. apply H0.
           discriminate. discriminate. discriminate.

  assert (PP: forall i, ssqsub ((pre 0) pcVar) (post i pcVar)).
    intros i.
    induction i; simpl. rewrite <- (FPC_Lemma1 _ _ _ (HypC 0)). apply leq_refl.
    eapply leq_trans. apply IHi. apply INCPost.

  intros i.
  induction i; simpl.   
    destruct x; simpl. rewrite fpcupd_Same.
       right. trivial.
    rewrite fpcupd_Diff; try discriminate. right. trivial.
  destruct x; simpl. rewrite fpcupd_Same. right. trivial.
    rewrite fpcupd_Diff; try discriminate. rewrite fpcupd_Diff in IHi; try discriminate.
    destruct IHi. left. eapply leq_trans. apply H0. apply INCPre.
    destruct (IHC _ _ (HypC i) (progVar p)); clear IHC HypC.
      rewrite (PrePost i). 
      left.  eapply @leq_trans with (y:=post i pcVar). apply (PP i).   rewrite fpcupd_Diff; try discriminate.
             eapply leq_trans. apply H1. apply FLUB_leqleft.
      rewrite (PrePost i). 
         rewrite fpcupd_Diff; try discriminate. right. unfold FLUB; simpl. rewrite H1. rewrite H0. apply leq_lubleft. rewrite PreZ. rewrite fpcupd_Diff; try discriminate. apply leq_refl.
Qed.

Lemma FPC_Nomod: forall G C H x (Hyp: FPC G C H),
        modVars C x = false -> H (progVar x) = G (progVar x). 
Proof. 
intros G C. 
generalize dependent G. 
induction C; intros; inversion Hyp; clear Hyp; intros; subst; simpl in *.
(*Skip*) 
 trivial. 
(*Assign*)
 apply fpcupd_split_at_var2; intros; subst; auto. rewrite pvar_eq_refl in H0. inversion H0.
(*Comp*)
   apply Bool.orb_false_iff in H0. destruct H0 as [H01 H02]. 
    rewrite (IHC2 _ _ _ H6); trivial.
    rewrite (IHC1 _ _ _ H4); trivial.
(*Ite*)
   apply Bool.orb_false_iff in H0. destruct H0 as [H01 H02].
   rewrite fpcupd_Diff; try discriminate. 
    unfold FLUB. simpl.
    rewrite (IHC1 _ _ x H8); trivial. 
    rewrite (IHC2 _ _ x H10); trivial.
    rewrite fpcupd_Diff; try discriminate.
    apply lub_refl. 
(*While*) 
 rewrite fpcupd_Diff; try discriminate.
   assert (GEN: forall i, pre i (progVar x) = G (progVar x)).
     intros i.
     induction i; simpl. rewrite PreZ. rewrite fpcupd_Diff.  reflexivity. discriminate.
     rewrite (PrePost i). 
     rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl.
          apply leq_lubright. rewrite (IHC _ _ x (HypC i) H0).
                          rewrite IHi. apply leq_refl.
   apply GEN.
Qed. 

Lemma FPC_Mod: forall G C H x (Hyp:FPC G C H),
        modVars C x = true -> @leq L LatL (G pcVar) (H (progVar x)). 
Proof. 
intros G C.
generalize dependent G. 
induction C; intros; inversion Hyp; clear Hyp; intros; subst; simpl in *.
(*Skip*) 
  inversion H0. 
(*Assign*)
 apply pvar_eq_ElimT in H0. subst.
 autorewrite with fpcREWRITE.
 apply lub_leqleft.
(*Comp*)
  assert (M1 := FPC_Lemma1 _ _ _ H4).  
  rewrite M1 in *. 
   remember (modVars C1 x) as b; destruct b; simpl in *; apply eq_sym in Heqb. clear H0.
      assert (X1 := IHC1 _ _ x H4 Heqb).  clear IHC1 H4. 
     remember (modVars C2 x) as b; destruct b; simpl in *; apply eq_sym in Heqb0. 
       apply (IHC2 _ _ x H6 Heqb0).  

      rewrite (FPC_Nomod  _ _ _ _ H6 Heqb0). rewrite <- M1. assumption.  
   
      apply (IHC2 _ _ x H6 H0).
(*Ite*)
 remember (modVars C1 x) as b; destruct b; simpl in *; apply eq_sym in Heqb.  clear H0. 
   clear IHC2 H10. 
      assert (X1 := IHC1 _ _ x H8 Heqb).  clear IHC1 H8 Heqb.
        rewrite fpcupd_Same in X1.  
        rewrite fpcupd_Diff; try discriminate.
          eapply leq_trans. eapply leq_trans. eapply lub_leqleft. apply X1.
             eapply FLUB_leqleft. 
      assert (X2 := IHC2 _ _ x H10 H0).  clear IHC2 H10 H0.
        rewrite fpcupd_Same in X2.  
        rewrite fpcupd_Diff; try discriminate.
          eapply leq_trans. eapply leq_trans. eapply lub_leqleft. apply X2.
             eapply FLUB_leqright. 
(*While*)
assert (INCPre: forall i, FSQSUB (pre i) (pre (S i))).
  intros j. induction j. (*rewrite PreZ. apply bot_leq.*)
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ (HypC 0)).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). apply FLUB_leqleft.
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff; try discriminate. apply FLUB_leqright.
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply FPC_mono. apply HypC. apply HypC. apply IHj.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                          eapply leq_trans. apply leq_monoright. apply GEval_mono. intros z. apply POST_FL.
                          apply leq_monoleft. apply POST_FL.
                      repeat rewrite fpcupd_Diff; try discriminate. apply POST_FL.

  rewrite fpcupd_Diff; try discriminate.
    assert (GEN: forall i, ssqsub (pre 0 pcVar) (pre (S i) (progVar x))).
    intros i; induction i; simpl.
      rewrite (PrePost 0).  
      assert (X:= IHC _ _ x (HypC 0) H0).
      rewrite fpcupd_Diff; try discriminate. eapply leq_trans. apply X. apply FLUB_leqleft.
    eapply (leq_trans IHi). apply INCPre.

    rewrite (PreAboveN _ (Le.le_refl _)).
      eapply @leq_trans with (y:=pre 0 pcVar). rewrite PreZ. rewrite fpcupd_Same. eapply lub_leqleft.
      apply GEN.
Qed. 

Lemma FPC_Mod': forall G C H (Hyp:FPC G C H) x,
        modVars C x = true -> @leq L LatL (G pcVar) (H (progVar x)). 
Proof.
intros.
eapply FPC_Mod. apply Hyp. assumption.
Qed.

Lemma FPC_Mod2: forall G C H x (Hyp:FPC G C H),
        modVars C x = true ->  @leq L LatL (H pcVar) (H (progVar x)). 
Proof. 
intros. 
rewrite <- (FPC_Lemma1 _ _ _ Hyp). 
eapply FPC_Mod.  apply Hyp.  assumption. 
Qed. 

Lemma FPC_triv: forall GG C HH G H, 
FPC GG C HH -> GG=G -> HH=H -> FPC G C H. 
Proof. 
intros; subst; trivial. 
Qed. 


Lemma FPC_Sound2_constructive_strong: 
forall G H C (tw: FPC G C H) s t (Hyp: Sem s C t) x,
ssqsub (H pcVar) (H (progVar x)) \/ 
(s x = t x /\ ssqsub (G (progVar x)) (H (progVar x))). 
Proof. 
intros G C H tw. 
induction tw. 
intros.  inversion Hyp; subst.  right.  split; trivial. 
  apply ssqsub_refl. 
intros.  subst.  inversion Hyp; subst; clear Hyp.
  rewrite M_upd_Diff; try discriminate.
  remember (pvar_eq x x0) as b; destruct b; subst; simpl in *.
    repeat rewrite M_upd_Same in *. 
    left.  apply llub_sqleft. 
  right. repeat rewrite M_upd_Diff; try congruence. split; trivial.  apply ssqsub_refl.
intros.   inversion Hyp; subst; clear Hyp. 
  destruct (IHtw2 _ _ H4 x).  left.  assumption. 
  destruct H as [HH1 HH2].  
  destruct (FPC_Lemma2 _ _ _ tw2 (progVar x)). 
    left.  assumption. 
  rewrite H. 
  rewrite <- (FPC_Lemma1 _ _ _ tw2). 
  destruct (IHtw1 _ _ H2 x).  left.  assumption. 
  right.  rewrite <- HH1.  assumption.   
intros. 
  assert (FL1 := FPC_Lemma1 _ _ _ tw1). 
  assert (FL2 := FPC_Lemma1 _ _ _ tw2). 
  inversion Hyp; subst; clear Hyp.
    repeat rewrite fpcupd_Same in *.
    repeat rewrite fpcupd_Diff; try discriminate.
      destruct (IHtw1 _ _ H11 x). 
        destruct (lub_eq_elim _ _ FL1). 
        left.  eapply ssqsub_trans.  apply H0. 
              eapply ssqsub_trans.  apply H. 
              apply FLUB_leqleft.  
        right.  rewrite fpcupd_Diff in H; try discriminate.  
                 destruct H.  split.  trivial. 
                   eapply ssqsub_trans.  apply H0. 
                   apply FLUB_leqleft.

    repeat rewrite fpcupd_Same in *.
    repeat rewrite fpcupd_Diff; try discriminate.
      destruct (IHtw2 _ _ H11 x). 
        destruct (lub_eq_elim _ _ FL2). 
        left.  eapply ssqsub_trans.  apply H0. 
              eapply ssqsub_trans.  apply H.  
              apply FLUB_leqright. 
        right; rewrite fpcupd_Diff in H; try discriminate.  
                 destruct H.  split.  trivial. 
                   eapply ssqsub_trans.  apply H0. 
                   apply FLUB_leqright. 
(*While*)
assert (INCPre: forall i, FSQSUB (pre i) (pre (S i))).
  intros j. induction j. (*rewrite PreZ. apply bot_leq.*)
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ (HypC 0)).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). apply FLUB_leqleft.
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff; try discriminate. apply FLUB_leqright.
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply FPC_mono. apply HypC. apply HypC. apply IHj.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                          eapply leq_trans. apply leq_monoright. apply GEval_mono. intros z. apply POST_FL.
                          apply leq_monoleft. apply POST_FL.
                      repeat rewrite fpcupd_Diff; try discriminate. apply POST_FL.

 assert (INCPreStrong: forall z i, z <= i -> FSQSUB (pre z) (pre i)).
   intros z. 
   induction z; simpl.
     intros i.
     induction i; intros; simpl. apply leq_refl.
     eapply leq_trans. apply IHi. apply Le.le_0_n. apply INCPre.
   intros i.
   induction i; intros; simpl. inversion H1.
   destruct (Lt.le_lt_or_eq _ _ H1).
     eapply leq_trans. apply IHi. apply Le.le_S_n. apply H2.
            apply INCPre.
     rewrite H2. apply leq_refl.
  (*use of INCPreStrong*)
  intros s t SEM.
  remember (While E C) as D.
  induction SEM; try inversion HeqD; subst.
    intros. assert (IH := IHSEM2 HeqD x). clear IHSEM1 IHSEM2 HeqD.
    rewrite fpcupd_Same in *. 
    rewrite fpcupd_Diff in *; try discriminate.
    destruct IH. left; assumption.
    destruct H as [Y1 Y2].
    assert (X := FPC_Lemma1 _ _ _ (HypC 0)).
    destruct (H0 0 _ _ SEM1 x).
    left. rewrite <- X in H. clear X. rewrite PreZ in H.
          rewrite fpcupd_Same in H.
          destruct (lub_leq_elim _ _ _ H). clear H.
          eapply leq_trans. apply H2.
          rewrite (PreAboveN N).
          eapply @leq_trans with (y:= pre 1 (progVar x)).
            rewrite (PrePost 0). rewrite fpcupd_Diff; try discriminate. apply FLUB_leqleft.
            apply INCPreStrong. apply Le.le_n_S. apply Le.le_0_n.
          apply Le.le_refl.
    right. rewrite <- Y1. destruct H. split. assumption.
      eapply @leq_trans with (y:= pre 0 (progVar x)).
        rewrite PreZ. apply leq_refl.
      eapply leq_trans. apply H2. 
      rewrite (PreAboveN N). 
          eapply @leq_trans with (y:= pre 1 (progVar x)).
            rewrite (PrePost 0). rewrite fpcupd_Diff; try discriminate. apply FLUB_leqleft.
            apply INCPreStrong. apply Le.le_n_S. apply Le.le_0_n.
          apply Le.le_refl.
 intros. rewrite fpcupd_Same.
   rewrite fpcupd_Diff; try discriminate.
   right. split. trivial. 
          eapply @leq_trans with (y:= pre 0 (progVar x)).
             rewrite PreZ. apply leq_refl.
            apply INCPreStrong. apply Le.le_0_n.
Qed. 

Lemma FPC_Sound2_constructive: 
forall G H C (tw: FPC G C H) s t (Hyp: Sem s C t) x,
ssqsub (H pcVar) (H (progVar x)) \/ s x = t x. 
Proof. 
intros. 
destruct (FPC_Sound2_constructive_strong _ _ _ tw _ _ Hyp x). 
  left; assumption. 
destruct H0.  right; assumption.  
Qed.  

Lemma FPC_Sound2: forall G H C (tw: FPC G C H) s t (Hyp: Sem s C t) x, 
~ ssqsub (H pcVar) (H (progVar x)) -> s x = t x. 
Proof. 
intros. 
destruct (FPC_Sound2_constructive _ _ _ tw _ _ Hyp x).  apply False_ind.  apply (H0 H1).  assumption. 
Qed. 

(*This is Definition 2.2 in the paper. 
Theorem 2.3 of the paper is FPC_Sound3 below, and (in polished form,
Theorem FPC_kappaNI at the end of this file.*)
Definition eq_below {A:Type} (k:L) (G:CTXT) (s s': PVar -> A) :=
  forall x, ssqsub (G x) k -> s x = s' x. 

Lemma eq_below_refl: forall {A} kappa G x, @eq_below A kappa G x x.
Proof.
intros. intros z Hz. reflexivity.
Qed.

Lemma eq_below_trans: forall {A} kappa G x y z, 
@eq_below A kappa G x y -> @eq_below A kappa G y z -> @eq_below A kappa G x z.
Proof.
intros. intros v Hv.
rewrite (H v Hv). apply (H0 _ Hv).
Qed.

Lemma eq_below_symm: forall {A} kappa G x y, 
@eq_below A kappa G x y -> @eq_below A kappa G y x.
Proof.
intros. intros v Hv.
rewrite (H v Hv). reflexivity.
Qed.

Lemma eq_below_evalE: forall E G s s' kappa,
eq_below kappa G s s' -> ssqsub (GEval G E) kappa -> eval s E = eval s' E. 
Proof. 
intros E. 
induction E; intros; subst; simpl in *. 
  trivial. 
  apply H.  apply H0. 
  rewrite (IHE1 _ _ _ _ H). 
    rewrite (IHE2 _ _ _ _ H).  trivial. 
    eapply @leq_trans with (y:=llub (GEval G E1) (GEval G E2)). 
      apply lub_leqright.  assumption. 
    eapply @leq_trans with (y:=llub (GEval G E1) (GEval G E2)). 
      apply llub_sqleft.  assumption. 
Qed. 

Lemma eq_below_mono: forall {A} k G s s' k' G', @eq_below A k G s s' -> 
  ssqsub k' k -> SQSUB G G' -> eq_below k' G' s s'. 
Proof. 
unfold eq_below, SQSUB.  intros. 
apply H.  eapply ssqsub_trans.  apply H1.  eapply ssqsub_trans.  apply H2.  apply H0. 
Qed. 

Lemma GevalEqonEval: forall E p GG s s' G,
eq_below p GG s s' -> SQSUB GG G -> ssqsub (GEval G E) p -> 
             eval s E = eval s' E. 
Proof. 
intros. 
eapply eq_below_evalE.  Focus 2.  apply H1. 
eapply eq_below_mono.  apply H.  apply ssqsub_refl.  apply H0. 
Qed. 

Lemma GEval_pc2: forall G e p, 
  FPCGEval (fpcupd G pcVar p) (progExpr e) = FPCGEval G (progExpr e).
Proof.
intros.
induction e; simpl.
  reflexivity.
  rewrite fpcupd_Diff. trivial. discriminate.
simpl in IHe1. rewrite IHe1.
  simpl in IHe2. rewrite IHe2. trivial.
Qed.

Lemma eq_below_GevalE: forall E p G s s',
eq_below p G s s' -> ssqsub (GEval G E) p -> GEval s E = GEval s' E. 
Proof. 
intros E. 
induction E; intros; subst; simpl in *. 
  trivial. 
  apply H.  apply H0. 
  rewrite (IHE1 _ _ _ _ H). 
    rewrite (IHE2 _ _ _ _ H).  trivial. 
    eapply @leq_trans with (y:=llub (GEval G E1) (GEval G E2)). 
      apply lub_leqright.  assumption. 
    eapply @leq_trans with (y:=llub (GEval G E1) (GEval G E2)). 
      apply lub_leqleft.  assumption. 
Qed. 

Lemma FPC_Sound3: forall G H C (tw: FPC G C H) 
s t (Hyp: Sem s C t) s' t' (Hyp': Sem s' C t') kappa,
eq_below kappa (FPCCTXT2CTXT G) s s' -> eq_below kappa (FPCCTXT2CTXT H) t t'. 
Proof. 
intros G H C. 
generalize dependent H. generalize dependent G. 
induction C; intros; subst; simpl in *.
(*Skip*)
inversion tw; clear tw; subst.
inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.  
  assumption. 
(*Assign*)
inversion tw; clear tw; subst.
inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'.  
  rewrite FPCCTXT2CTXT_upd_progVar. 
  intros z Hz.
  apply M_upd_split_at; intros; subst.
    rewrite M_upd_Same in *. 
    rewrite M_upd_Same in Hz. 
    eapply GevalEqonEval.  apply H0.  
        apply leq_refl.  
      eapply leq_trans.  Focus 2.   apply Hz.  
      apply lub_leqright. 
  rewrite M_upd_Diff; try assumption.
    rewrite M_upd_Diff in Hz; try assumption.
    apply H0.  apply Hz. 
(*Comp*)
inversion tw; clear tw; subst.
inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. 
  eapply IHC2.  apply H6.  apply H8. apply H10.
  eapply IHC1.  apply H4.  apply H5.  apply H7.
  apply H0. 
(*If*)
inversion tw; clear tw; subst.
assert (EV:= eq_below_evalE e (FPCCTXT2CTXT G) s s' kappa H0).
inversion Hyp; subst; clear Hyp. 
  (*True*)
    inversion Hyp'; subst; clear Hyp'. 
    (*True*)
       clear IHC2 H10. 
       assert (IH := IHC1 _ _ H8 _ _ H6 _ _ H9 kappa).  clear IHC1 H6 H8 H9. 
         rewrite FPCCTXT2CTXT_upd_pcVar in *. 
         eapply eq_below_mono.  apply (IH H0). 
           apply ssqsub_refl.
           clear IH H5.  intros z.  unfold FPCCTXT2CTXT. apply FLUB_leqleft.  
     (*False*)
        clear IHC1 IHC2. 
        rewrite FPCCTXT2CTXT_upd_pcVar in *. 
        intros z Hz.  unfold FPCCTXT2CTXT in Hz; simpl in Hz. 
        destruct (lub_leq_elim _ _ _ Hz). 
        destruct (FPC_Sound2_constructive_strong _ _ _ H8 _ _ H6 z). 
          assert (L1 := FPC_Lemma1 _ _ _ H8). rewrite M_upd_Same in L1. simpl in L1.
          destruct (@lub_eq_elim L LatL _ _ _ L1). 
          rewrite EV in H5.  rewrite H5 in H7.  inversion H7. 
            eapply ssqsub_trans.  apply H4.  eapply ssqsub_trans.  apply H2.  apply H. 
        destruct H2 as [STR STR1].  
           rewrite fpcupd_Diff in STR1.  Focus 2.  intros HHH.  inversion HHH. 
        rewrite <- STR.  clear STR. 
        destruct (FPC_Sound2_constructive_strong _ _ _ H10 _ _ H9 z).  
            assert (L2 := FPC_Lemma1 _ _ _ H10). rewrite M_upd_Same in L2.  simpl in L2. 
            destruct (@lub_eq_elim L LatL _ _ _ L2). 
            rewrite EV in H5.  rewrite H5 in H7.  inversion H7. 
              eapply ssqsub_trans.  apply H4.  eapply ssqsub_trans.  apply H2.  apply H1. 
        destruct H2 as [STR STR2]. 
           rewrite fpcupd_Diff in STR2.  Focus 2.  intros HHH.  inversion HHH.  
        rewrite <- STR.  clear STR. 
        apply H0. 
        eapply leq_trans. apply STR1. apply H.
  (*False*)
    inversion Hyp'; subst; clear Hyp'. 
    (*True*)
        clear IHC1 IHC2. 
        rewrite FPCCTXT2CTXT_upd_pcVar in *. 
        intros z Hz.  unfold FPCCTXT2CTXT in Hz; simpl in Hz. 
        destruct (lub_leq_elim _ _ _ Hz). 
        destruct (FPC_Sound2_constructive_strong _ _ _ H8 _ _ H9 z). 
          assert (L1 := FPC_Lemma1 _ _ _ H8).  rewrite fpcupd_Same in L1.  simpl in L1. 
          destruct (@lub_eq_elim L LatL _ _ _ L1). 
          rewrite EV in H5.  rewrite H5 in H7.  inversion H7. 
            eapply ssqsub_trans.  apply H4.  eapply ssqsub_trans.  apply H2.  apply H. 
        destruct H2 as [STR STR1].  
           rewrite fpcupd_Diff in STR1.  Focus 2.  intros HHH.  inversion HHH. 
        rewrite <- STR.  clear STR. 
        destruct (FPC_Sound2_constructive_strong _ _ _ H10 _ _ H6 z).  
            assert (L2 := FPC_Lemma1 _ _ _ H10).  rewrite fpcupd_Same in L2.  simpl in L2. 
            destruct (@lub_eq_elim L LatL _ _ _ L2). 
            rewrite EV in H5.  rewrite H5 in H7.  inversion H7. 
              eapply ssqsub_trans.  apply H4.  eapply ssqsub_trans.  apply H2.  apply H1. 
        destruct H2 as [STR STR2]. 
           rewrite fpcupd_Diff in STR2.  Focus 2.  intros HHH.  inversion HHH.  
        rewrite <- STR.  clear STR. 
        apply H0. 
        eapply leq_trans. apply STR1. apply H. 
    (*False*)
       clear IHC1 H8.
       assert (IH := IHC2 _ _ H10 _ _ H6 _ _ H9 kappa).  clear IHC2 H10 H6 H9. 
         rewrite FPCCTXT2CTXT_upd_pcVar in *. 
         eapply eq_below_mono.  apply (IH H0). 
           apply ssqsub_refl. 
           clear IH H0.  intros z.  unfold FPCCTXT2CTXT. apply FLUB_leqright. 
(*While*) 
(*assert (FSound2 := FPC_Sound2_constructive_strong _ _ _ tw).*)
inversion tw; clear tw; subst.
assert (INCPre: forall i, FSQSUB (pre i) (pre (S i))).
  intros j. induction j. 
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ (HypC 0)).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). apply FLUB_leqleft.
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff; try discriminate. apply FLUB_leqright.
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply FPC_mono. apply HypC. apply HypC. apply IHj.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                        apply leq_mono.
                          apply POST_FL.
                          apply GEval_mono. intros z. apply POST_FL.
                      repeat rewrite fpcupd_Diff; try discriminate. apply POST_FL.
(*
 assert (INCPost: forall i, FSQSUB (post i) (post (S i))).
   intros i. eapply FPC_mono. apply HypC. apply HypC. apply INCPre.
*)
 assert (INCPreStrong: forall z i, z <= i -> FSQSUB (pre z) (pre i)).
   intros z. 
   induction z; simpl.
     intros i.
     induction i; intros; simpl. apply leq_refl.
     eapply leq_trans. apply IHi. apply Le.le_0_n. apply INCPre.
   intros i.
   induction i; intros; simpl. inversion H.
   destruct (Lt.le_lt_or_eq _ _ H).
     eapply leq_trans. apply IHi. apply Le.le_S_n. apply H1.
            apply INCPre.
     rewrite H1. apply leq_refl.
rewrite FPCCTXT2CTXT_upd_pcVar in *.

cut (eq_below kappa (FPCCTXT2CTXT (pre N)) s s' -> 
     eq_below kappa (FPCCTXT2CTXT (pre N)) t t').
  intros GEN. apply GEN. clear GEN.
  eapply eq_below_mono. apply H0. apply leq_refl.
  eapply @leq_trans with (y:= (FPCCTXT2CTXT (pre 0))).
    rewrite PreZ. rewrite FPCCTXT2CTXT_upd_pcVar. apply leq_refl.
    intros x. apply INCPreStrong. apply Le.le_0_n.
intros EQ. clear H0.
assert (XX: ssqsub (FPCGEval (pre N) (progExpr e)) (pre N pcVar)).
             rewrite (PreAboveN N). rewrite (PrePost N). rewrite fpcupd_Same. rewrite GEval_pc2. apply lub_leqright.
             apply Le.le_refl.
remember (While e C) as D.
 generalize dependent t'. generalize dependent s'.
 generalize dependent e. generalize dependent C.
 induction Hyp; intros; try inversion HeqD; subst.  
  (*True*)
    clear HeqD. generalize dependent C0. intros C; intros. clear IHHyp1.
    inversion Hyp'; subst; clear Hyp'.
    (*True*)
       generalize dependent r0. intros r' Hyp1' Hyp2'.
       assert (HypCN:=HypC N). 
       assert (HypC0:=HypC 0).
       assert (EQRR: eq_below kappa (FPCCTXT2CTXT (post N)) r r').
         eapply IHC. apply HypCN. apply Hyp1. apply Hyp1'. assumption.
       eapply IHHyp2; clear IHHyp2.
(*         apply FSound2.*)
         apply IHC.
         apply HypC.
         reflexivity.
(*         apply PreZ.*)
         apply PrePost.
         apply PreZ.
         apply XX.
         Focus 2. apply Hyp2'.
         eapply eq_below_mono. apply EQRR. apply leq_refl.
           rewrite (PreAboveN N). rewrite (PrePost N).
             rewrite FPCCTXT2CTXT_upd_pcVar. intros z. unfold FPCCTXT2CTXT. apply FLUB_leqleft.
           apply Le.le_refl.
    (*False*) clear IHHyp2 (*FSound2*) IHC.
       assert (HypS: Sem s (While e C) t). eapply S_WTrue. apply H. apply Hyp1. apply Hyp2.
       intros x Hx.        
        remember (modVars (While e C) x) as b.
        destruct b; apply eq_sym in Heqb; simpl in Heqb.
           rewrite (eq_below_evalE e _ _ _ kappa EQ) in H. rewrite H in H3. inversion H3.   
           eapply leq_trans. apply XX.    
             eapply leq_trans. apply (FPC_Mod _ _ _ x (HypC N) Heqb).
             eapply @leq_trans with (y := FPCCTXT2CTXT (pre (S N)) x).
             rewrite PrePost.  rewrite FPCCTXT2CTXT_upd_pcVar.  unfold FPCCTXT2CTXT. apply FLUB_leqleft.
             rewrite <- (PreAboveN N). apply Hx. apply Le.le_refl.
        rewrite <- (Sem_nomod _ _ _ HypS x). apply EQ. apply Hx.
          apply Heqb.
(*False*)
    clear IHC HeqD. generalize dependent C0. intros C; intros.
    inversion Hyp'; subst; clear Hyp'.
    (*True*)
       generalize dependent r. intros r' Hyp1' Hyp2'.
       assert (HypS: Sem s' (While e C) t'). eapply S_WTrue. apply H2. apply Hyp1'. apply Hyp2'.
       intros x Hx.        
        remember (modVars (While e C) x) as b.
        destruct b; apply eq_sym in Heqb; simpl in Heqb.
           rewrite (eq_below_evalE e _ _ _ kappa EQ) in H. rewrite H in H2. inversion H2.   
           eapply leq_trans. apply XX.    
             eapply leq_trans. apply (FPC_Mod _ _ _ x (HypC N) Heqb).
             eapply @leq_trans with (y := FPCCTXT2CTXT (pre (S N)) x).
             rewrite PrePost.  rewrite FPCCTXT2CTXT_upd_pcVar.  unfold FPCCTXT2CTXT. apply FLUB_leqleft. 
             rewrite <- (PreAboveN N). apply Hx. apply Le.le_refl.
        rewrite <- (Sem_nomod _ _ _ HypS x). apply EQ. apply Hx.
          apply Heqb.
     (*False*) apply EQ.
Qed.

(*The variant of FPC that constructs a type - 
useful for defining the program transformations*)
Inductive FPCT : FPCCTXT -> Com -> FPCCTXT -> Type :=
  FPCT_Skip: forall G, FPCT G Skip G
| FPCT_Assign: forall p x E G H, 
         p = G pcVar ->
         H = fpcupd G (progVar x) (llub p (FPCGEval G (progExpr E))) ->
         FPCT G (Assign x E) H
| FPCT_Comp: forall G G1 G2 C1 C2, 
         FPCT G C1 G1 -> FPCT G1 C2 G2 ->
         FPCT G (Comp C1 C2) G2
| FPCT_Ite: forall p G GG G1 G2 E C1 C2 a H HH,
         a = FPCGEval G (progExpr E) ->
         p = G pcVar ->
         GG = fpcupd G pcVar (llub p a) ->
         FPCT GG C1 G1 -> 
         FPCT GG C2 G2 ->
         H = FLUB G1 G2 ->
         HH = fpcupd H pcVar p ->
         FPCT G (Ite E C1 C2) HH
| FPCT_While: forall p G E C H N
         (Hp: p = (G pcVar))
         (pre post : nat -> FPCCTXT)
         (PreBelowN: forall i, i < N -> (pre i) <> pre (S i))
         (PreAboveN: forall i, N <= i -> (pre i) = pre (S i))
         (PreZ: pre 0 = fpcupd G pcVar (llub (G pcVar) (FPCGEval G (progExpr E))))
         (PreN: fpcupd (pre N) pcVar (G pcVar) = H)
         (PrePost: forall i, pre (S i) = fpcupd (FLUB (post i) G) pcVar 
                                (llub ((FLUB (post i) G) pcVar) (FPCGEval (FLUB (post i) G) (progExpr E))))
         (HypC: forall i,  FPCT (pre i) C (post i)),
         FPCT G (While E C) H.

Lemma FPCT_FPC: forall G C H (Hyp:FPCT G C H), FPC G C H. 
Proof. 
intros.
induction Hyp; intros.
constructor. 
econstructor. apply e. assumption.
econstructor. apply IHHyp1. apply IHHyp2. 
econstructor. apply e. apply e0. apply e1. apply IHHyp1. apply IHHyp2. apply e2. apply e3.  
econstructor. apply Hp. apply PreBelowN. apply PreAboveN. subst. apply PreZ.
  subst. reflexivity.  apply PrePost. apply H0. 
Qed. 

Lemma FPCT_unique: forall C G H (Hyp: FPCT G C H) H'
 (Hyp': FPCT G C H'), H=H'. 
Proof.
intros.
eapply FPC_unique.
apply (FPCT_FPC _ _ _ Hyp). 
apply (FPCT_FPC _ _ _ Hyp'). 
Qed. 

Definition NI {A} (G H:CTXT) (s s' t t':PVar -> A) :=
 forall kappa, eq_below kappa G s s' -> eq_below kappa H t t'.

Definition kappaNI (G:CTXT) C H :=
forall s s' t t' (Hyp: Sem s C t) (Hyp': Sem s' C t'), 
NI G H s s' t t'.

Theorem FPC_kappaNI: forall (G:FPCCTXT) C H, FPC G C H -> 
kappaNI (FPCCTXT2CTXT G) C (FPCCTXT2CTXT H).
Proof.
unfold kappaNI, NI; intros. 
eapply FPC_Sound3. apply H0. apply Hyp. apply Hyp'. apply H1.
Qed.
