Require Export Coq.Lists.List. 
Require Export AXBASE.
Require Export MAP.

Parameter PVar:Type. 
Parameter pvar_eq: forall (x y: PVar), {x=y} + {x <> y}.  

Definition upd {A:Type} 
  (X:PVar -> A) (x:PVar) (a:A) : PVar -> A :=
fun y => if pvar_eq x y then a else X y.

Definition PVAR_MAP {A:Type}: MAP PVar pvar_eq A upd.
  constructor. unfold upd. trivial. Defined.
Existing Instance PVAR_MAP.

Lemma pvar_eq_ElimT: forall x p (H:(if pvar_eq p x then true else false) = true), p=x.
Proof.
 intros.
 remember (pvar_eq p x) as b; destruct b; subst; trivial. inversion H.
Qed.  
Hint Resolve pvar_eq_ElimT: varREWRITE. 

Lemma pvar_eq_ElimF: forall x p (H:(if pvar_eq p x then true else false) = false), p<>x.
Proof.
 intros.
 remember (pvar_eq p x) as b; destruct b; subst; trivial. inversion H.
Qed.  
Hint Resolve pvar_eq_ElimF: varREWRITE. 

(*making parameter A implicit here would prevent me from 
  adding the lemma to the hint database polymorphically*)
Lemma pvar_eq_refl: forall A z (a b:A),
  (if pvar_eq z z then a else b) = a.
Proof. intros.
  remember (pvar_eq z z) as bb; destruct bb; subst; trivial. 
  apply False_ind.  apply n.  trivial.
Qed. 
Hint Rewrite pvar_eq_refl: varREWRITE.

Lemma pvar_eq_diff:
  forall (A : Type) (z x: PVar) (a b : A), 
  z <> x -> (if pvar_eq z x then a else b) = b.
Proof.
intros. remember (pvar_eq z x) as bb; destruct bb; subst; trivial. exfalso. apply H. trivial.
Qed.

Parameter Val:Type.

Definition Store:Type := PVar -> Val.
Definition bop:Type := Val -> Val -> Val.  

Inductive Expr:=
  valExpr: Val -> Expr
| varExpr: PVar -> Expr
| bopExpr: Expr -> bop -> Expr -> Expr.  

Inductive Com:=
  Skip : Com
| Assign: PVar -> Expr -> Com
| Comp: Com -> Com -> Com
| Ite: Expr -> Com -> Com -> Com
| While: Expr -> Com -> Com.  

Fixpoint modVars (C: Com) : PVar -> bool :=
match C with 
   Skip => (fun x => false)
| (Assign x E) => (fun y => if pvar_eq x y then true else false)
| (Comp C1 C2) => (fun x => orb (modVars C1 x) (modVars C2 x))
| (Ite E C1 C2) => (fun x => orb (modVars C1 x) (modVars C2 x))
| (While E C) => modVars C
end.  

Fixpoint modVarsL (C: Com) : list PVar :=
match C with 
   Skip => nil
| (Assign x E) => cons x nil
| (Comp C1 C2) => app (modVarsL C1) (modVarsL C2)
| (Ite E C1 C2) => app (modVarsL C1) (modVarsL C2)
| (While E C) => modVarsL C
end.  

Lemma modVarsChar: forall C x, modVars C x = true <-> In x (modVarsL C).  
Proof.   
intros.  
induction C; simpl; split; intros.  

  inversion H.  contradiction.  

remember (pvar_eq p x) as b; destruct b; simpl in *.  left; assumption. 
  inversion H. 
destruct H; try contradiction.  subst; simpl. 
  remember (pvar_eq x x) as b; destruct b; simpl in *.  trivial. 
  apply False_ind.   apply n.  trivial. 

remember (modVars C1 x) as b; destruct b; simpl in *. 
  apply in_or_app.  left.  apply IHC1.  trivial. 
  apply in_or_app.  right.  apply IHC2.  apply H. 
destruct (in_app_or _ _ _ H). 
  destruct IHC1 as [_ H1].   rewrite (H1 H0).  simpl.  trivial. 
  destruct IHC2 as [_ H2].   rewrite (H2 H0). 
    remember (modVars C1 x) as b; destruct b; simpl in *.  trivial.  trivial. 

remember (modVars C1 x) as b; destruct b; simpl in *. 
  apply in_or_app.  left.  apply IHC1.  trivial. 
  apply in_or_app.  right.  apply IHC2.  apply H. 
destruct (in_app_or _ _ _ H). 
  destruct IHC1 as [_ H1].   rewrite (H1 H0).  simpl.  trivial. 
  destruct IHC2 as [_ H2].   rewrite (H2 H0). 
    remember (modVars C1 x) as b; destruct b; simpl in *.  trivial.  trivial. 

apply IHC.  apply H. 
apply IHC.  apply H.  
Qed.  

Lemma modVarsChar': forall C x, modVars C x = false <-> ~ In x (modVarsL C).  
Proof.  
intros. 
remember (@in_dec PVar pvar_eq x (modVarsL C)) as b; destruct b. 
 destruct (modVarsChar C x). 
 split; intros.  rewrite (H0 i) in H1.  inversion H1. 
  apply False_ind.  apply (H1 i). 
split; intros.  assumption. 
  remember (modVars C x) as b; destruct b; apply eq_sym in Heqb0; trivial. 
  destruct (modVarsChar C x).  apply False_ind.  apply (n (H0 Heqb0)).  
Qed.  

Lemma modVars_app: forall z C1 C2,
 In z (modVarsL C1 ++ modVarsL C2) <-> (modVars C1 z = true \/ modVars C2 z = true).  
Proof.  
intros. 
split; intros. 
  destruct (in_app_or _ _ _ H). 
    left.  apply modVarsChar.  assumption. 
    right.  apply modVarsChar.  assumption. 
destruct H. 
  apply modVarsChar in H; apply in_or_app. 
    left; assumption. 

  apply modVarsChar in H; apply in_or_app. 
    right; assumption.  
Qed.  

Lemma modVars_no_app: forall z C1 C2,
~ In z (modVarsL C1 ++ modVarsL C2) ->
(modVars C1 z || modVars C2 z)%bool = false.  
Proof.  
intros. 
remember (modVars C1 z) as b1; destruct b1; subst; simpl; apply eq_sym in Heqb1. 
  apply (modVarsChar C1 z) in Heqb1. 
  apply False_ind.  apply H.  apply in_or_app.  left; assumption. 

remember (modVars C2 z) as b2; destruct b2; subst; simpl; apply eq_sym in Heqb2. 
  apply (modVarsChar C2 z) in Heqb2. 
  apply False_ind.  apply H.  apply in_or_app.  right; assumption. 
trivial.  
Qed. 

Fixpoint eval (s:PVar -> Val)(E: Expr): Val :=
 match E with
   valExpr v => v
| varExpr x => s x
| bopExpr e1 b e2 => b (eval s e1) (eval s e2)
end. 

Parameter istrue : Val -> bool.  

Inductive Sem  : Store -> Com -> Store -> Prop :=
  S_Skip: forall s t, s=t -> Sem s Skip t
| S_Assign: forall s x e t, 
            t = upd s x (eval s e) -> Sem s (Assign x e) t
| S_Comp: forall s C1 C2 t r, 
          Sem s C1 r -> Sem r C2 t -> Sem s (Comp C1 C2) t
| S_True: forall s E C1 C2 t, istrue (eval s E) = true -> 
          Sem s C1 t -> Sem s (Ite E C1 C2) t
| S_False: forall s E C1 C2 t, istrue (eval s E) = false -> 
           Sem s C2 t -> Sem s (Ite E C1 C2) t
| S_WTrue: forall s b C r t, istrue (eval s b) = true -> 
           Sem s C r -> Sem r (While b C) t -> Sem s (While b C) t
| S_WFalse: forall s b C t, istrue (eval s b) = false -> s=t -> 
           Sem s (While b C) t.  

Lemma Sem_nomod: forall s C t (SEM:Sem s C t) x
 (MV:modVars C x = false), s x = t x.
Proof.
intros s C t SEM.
induction SEM; intros; subst; simpl in *; trivial.
(*Assign*)
unfold upd.
remember (pvar_eq x x0) as b; destruct b; subst; simpl. inversion MV. trivial.
(*Comp*)
remember (modVars C1 x) as b; destruct b; simpl in *. inversion MV.
apply eq_sym in Heqb.
rewrite <- (IHSEM2 x MV).
apply (IHSEM1 _ Heqb).
(*If-True*)
remember (modVars C1 x) as b; destruct b; simpl in *. inversion MV.
apply eq_sym in Heqb.
apply (IHSEM x Heqb).
(*If-False*)
remember (modVars C1 x) as b; destruct b; simpl in *. inversion MV.
apply (IHSEM x MV).
(*While*)
rewrite <- (IHSEM2 x MV).
apply (IHSEM1 _ MV).
Qed.

