Require Export AXBASE.

Set Implicit Arguments. 
Record partialorderfacts (t : Type) 
       (leq: t -> t -> Prop) : Type := 
POFacts{
  po_refl: forall x, leq x x;
  po_trans: forall x y z, leq x y -> leq y z -> leq x z;
  po_antisym: forall x y, leq x y -> leq y x -> x=y
}.  

Record latticefacts (t : Type) (leq: t -> t -> Prop) (bot:t)
       (lub: t -> t -> t): Type := 
LatticeFacts {
   lat_po: partialorderfacts leq;
   lub_com: forall x y, lub x y = lub y x;
   
   lub_assoc: forall x y z, lub x (lub y z) = lub (lub x y) z;

   bot_id: forall x, lub bot x = x;

   compat_lub: forall x y, leq x y <-> (y = lub x y)
}.  

(** A lattice is a type, an element bot, relations lub and sqsub
    on that type, and the above facts about the relations *)
Class lattice (t:Type) : Type :=
  Lattice
  { bot : t;
    lub : t -> t -> t;
    leq: t -> t -> Prop;
    facts : latticefacts leq bot lub
  }.  

Lemma id_bot {A}`{lattice A}: forall {x: A}, 
  lub x bot = x.  
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
  rewrite lub_com. apply bot_id.
Qed. 

Lemma lub_refl {A}`{lattice A}: forall {x: A}, 
  lub x x = x.  
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
apply eq_sym. apply compat_lub.
destruct lat_po. apply po_refl0.
Qed.  

Lemma leq_refl {A}`{lattice A}: forall {x: A}, leq x x. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
destruct lat_po.  apply po_refl0. 
Qed. 

Lemma leq_trans {A}`{lattice A}: forall {x y z: A}, 
leq x y -> leq y z -> leq x z. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
destruct lat_po.  eapply po_trans0.  apply H0.  apply H1. 
Qed. 

Lemma leq_snart {A}`{lattice A}: forall {x y z: A},
       leq y z -> leq x y -> leq x z.
Proof.
intros.
eapply leq_trans. apply H1. assumption.
Qed.

Lemma leq_antisym {A}`{lattice A}: forall {x y: A}, 
leq x y -> leq y x -> x=y. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
destruct lat_po.  eapply po_antisym0.  apply H0.  apply H1. 
Qed. 

Lemma bot_leq {A}`{lattice A}: forall {x: A}, leq bot x. 
Proof. 
intros.  destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
apply compat_lub. 
rewrite bot_id. 
trivial. 
Qed. 


Lemma lub_leqleft {A}`{lattice A}: forall {p q: A}, leq p (lub p q). 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub]. 
apply compat_lub.
rewrite lub_assoc.
rewrite lub_refl.
trivial. 
Qed. 

Lemma lub_leqright {A}`{lattice A}: forall {p q: A}, leq p (lub q p). 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub]. 
rewrite lub_com. 
apply lub_leqleft. 
Qed. 

Lemma lub_triv {A}`{lattice A}: 
forall p p' q q' (Hp: p=p') (Hq:q=q'), lub p q = lub p' q'.
Proof.
intros.
subst. trivial. 
Qed.

Lemma lubleft {A}`{lattice A}: 
forall (p1 p2 q:A), leq p1 p2 -> leq p1 (lub p2 q).
Proof.
intros.
eapply leq_trans. apply H0. eapply lub_leqleft.
Qed.

Lemma lubright {A}`{lattice A}: 
forall (p1 p2 q:A), leq p1 p2 -> leq p1 (lub q p2).
Proof.
intros.
eapply leq_trans. apply H0. apply lub_leqright.
Qed.

Lemma leq_monoright {A}`{lattice A}: forall (p q1 q2:A), 
leq q1 q2 -> leq (lub p q1) (lub p q2). 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
apply compat_lub.
rewrite (lub_com p q1).
rewrite <- lub_assoc.
rewrite (lub_assoc p).
rewrite lub_refl.
apply compat_lub.
apply lubright.
apply H0.
Qed. 

Lemma leq_monoleft {A}`{lattice A}: forall (p q1 q2:A), 
leq q1 q2 -> leq (lub q1 p) (lub q2 p). 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub]. 
rewrite (lub_com q1). 
rewrite (lub_com q2). 
apply (leq_monoright _ _ _ H0). 
Qed. 

Lemma leq_mono {A}`{lattice A}: forall (p1 p2 q1 q2:A), 
leq p1 p2 -> leq q1 q2 -> leq (lub p1 q1) (lub p2 q2). 
Proof.
intros.
eapply leq_trans.
  eapply leq_monoleft. apply H0.
  eapply leq_monoright. apply H1.
Qed. 

Lemma leq_lubright {A}`{lattice A}: forall (p q:A), leq p q -> lub p q = q. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
apply eq_sym.
apply compat_lub.
apply H0.
Qed. 

Lemma leq_lubleft {A}`{lattice A}: forall {p q: A},  leq p q -> lub q p = q. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
rewrite lub_com. 
apply leq_lubright.  apply H0. 
Qed. 

Lemma lub_leq {A}`{lattice A}: forall {a b c: A},
  leq a c -> leq b c -> leq (lub a b) c. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub]. 
apply compat_lub. 
destruct (compat_lub a c) as [K1 _]. 
rewrite (lub_com a). 
rewrite <- lub_assoc. 
rewrite <- (K1 H0). 
destruct (compat_lub b c) as [K2 _]. 
apply (K2 H1).  
Qed. 

Lemma lub_leq_elim {A}`{lattice A}:
  forall (p q r:A), leq (lub p q) r -> leq p r /\ leq q r. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
destruct (compat_lub (lub p q) r) as [K1 _]. 
rewrite (K1 H0). 
split. 
  eapply leq_trans.  eapply lub_leqleft.   eapply lub_leqleft. 
  eapply leq_trans.  eapply lub_leqright.   eapply lub_leqleft. 
Qed. 

Lemma lub_eq_elim {A}`{lattice A}:
 forall (p q r:A), (lub p q) = r -> leq p r /\ leq q r. 
Proof. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
rewrite <- H0 in *. 
split.  apply lub_leqleft.  apply lub_leqright. 
Qed.  
 
Lemma lub_cong {A}`{lattice A}:
  forall l l' m m', l=l' -> m=m' -> lub l m = lub l' m'.
Proof.
intros; subst. trivial.
Qed.

Lemma eq_leq {A}`{lattice A}: forall p q, p=q -> leq p q.
Proof.
intros; subst. apply leq_refl.
Qed.

Lemma lub_assoc1 {A}`{lattice A}: forall p q, lub (lub p q) q = lub p q.
Proof.
intros. 
intros; destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub].
rewrite <- lub_assoc.
rewrite lub_refl.
trivial.
Qed.

Definition po_funspace_facts: forall t leq T
(f: @partialorderfacts t leq), 
  @partialorderfacts (T -> t) 
       (fun G => fun H => forall x, leq (G x) (H x)). 
Proof. 
intros t leq T pofacts. 
destruct pofacts as [po_refl po_trans po_antisym]. 
split; intros. 
(*po_refl*)
apply po_refl. 
(*po_trans*)
eapply po_trans.  apply H.  apply H0. 
(*po_antisym*)
apply extensionality; intros z.  apply po_antisym.  apply H.  apply H0. 
Qed. 

Definition funspace_facts: forall t leq bot lub T
(f: @latticefacts t leq bot lub ), 
  @latticefacts (T -> t)
       (fun G => fun H => forall x, leq (G x) (H x)) 
       (fun t => bot) 
       (fun G => fun G' => fun t => lub (G t) (G' t)). 
Proof. 
intros t leq bot lub T facts. 
assert (LUB_refl := @lub_refl t (@Lattice _ _ _ _ facts)). 
destruct facts as [lat_po lub_com lub_assoc  
                           bot_id compat_lub]. 
assert (P: partialorderfacts (fun G H : T -> t => forall x : T, leq (G x) (H x))). 
  apply po_funspace_facts.  apply lat_po. 
split; intros. 
(*po_facts*)
assumption. 
(*lub_com*)
apply extensionality; intros z.  apply lub_com. 
(*lub_assoc*)
apply extensionality; intros v.  apply lub_assoc. 
(*bot_id*)
apply extensionality; intros v.  apply bot_id. 
(*compat_lub*)
split; intros. 
  apply extensionality; intros v.  apply compat_lub.  apply H. 
  apply compat_lub.  rewrite H. 
    destruct lat_po.  
     rewrite lub_assoc.
     rewrite LUB_refl. trivial.  
Qed.  

Definition FunSpace {A}`{lattice A} (T:Type) : lattice (T -> A). 
apply (Lattice (funspace_facts T facts)). 
Defined. 


(*H is a fixed point of f greater than G:*)
Definition FIX {A}`{lattice A} (f:A -> option A) (G H:A): Prop :=
  leq G H /\ f H = Some H.

(*H is the least fixed point of f above G:*)
Definition LFP {A}`{lattice A} (f:A -> option A) (G H:A): Prop :=
  FIX f G H /\ (forall H', FIX f G H' -> leq H H').

Lemma LFP_unique {A}`{lattice A}: 
forall f G H H', LFP f G H -> LFP f G H' -> H = H'.
Proof.
intros. unfold LFP in *.
destruct H1. destruct H2.
eapply @leq_antisym.
  apply H3. assumption.
  apply H4. assumption.
Qed.

Definition po_prodspace_facts: forall t1 leq1 t2 leq2
(f1: @partialorderfacts t1 leq1)
(f2: @partialorderfacts t2 leq2), 
  @partialorderfacts (prod t1 t2) 
       (fun G => fun H => leq1 (fst G) (fst H) /\ leq2 (snd G) (snd H)). 
Proof. 
intros t1 leq1 t2 leq2 pofacts1 pofacts2. 
split. 
(*po_refl*)
intros G. split. apply pofacts1. apply pofacts2.
(*po_trans*)
intros G H K [LEQ_GH1 LEQ_GH2] [LEQ_HK1 LEQ_HK2].
split; eapply po_trans.
  apply pofacts1. apply LEQ_GH1. apply LEQ_HK1.
  apply pofacts2. apply LEQ_GH2. apply LEQ_HK2.
(*po_antisym*)
intros G H [LEQ_GH1 LEQ_GH2] [LEQ_HG1 LEQ_HG2].
destruct G as [G1 G2]. destruct H as [H1 H2].
assert (G1 = H1).
  apply pofacts1. apply LEQ_GH1. apply LEQ_HG1.
assert (G2 = H2).
  apply pofacts2. apply LEQ_GH2. apply LEQ_HG2.
subst.
trivial.
Qed. 

Definition prodspace_facts: forall 
t1 leq1 bot1 lub1
(f1: @latticefacts t1 leq1 bot1 lub1) 
t2 leq2 bot2 lub2
(f2: @latticefacts t2 leq2 bot2 lub2), 
  @latticefacts (prod t1 t2)
       (fun G => fun H => leq1 (fst G) (fst H) /\ leq2 (snd G) (snd H))
       (bot1,bot2) 
       (fun G => fun G' => (lub1 (fst G) (fst G'), lub2 (snd G) (snd G'))).
Proof. 
intros.
destruct f1 as [lat_po1 lub_com1 lub_assoc1 
                bot_id1 compat_lub1]. 
destruct f2 as [lat_po2 lub_com2 lub_assoc2
                bot_id2 compat_lub2]. 
split; intros.
(*po_facts*)
eapply po_prodspace_facts; assumption.
(*lub_com*)
rewrite lub_com1. rewrite lub_com2. trivial.
(*lub_assoc*)
simpl in *.
rewrite lub_assoc1. rewrite lub_assoc2. trivial.
(*bot_id*)
destruct x; simpl in *. rewrite bot_id1. rewrite bot_id2. trivial.
(*compat_lub*)
destruct x as [x1 x2]. destruct y as [y1 y2]. simpl in *.
split; intros.
  destruct H as [L1 L2].
  destruct (compat_lub1 x1 y1) as [K1 _].
  rewrite <- (K1 L1).
  destruct (compat_lub2 x2 y2) as [K2 _].
  rewrite <- (K2 L2).
  trivial.
inversion H.
  split. apply compat_lub1. rewrite <- H1. apply H1. 
  apply compat_lub2. rewrite <- H2. apply H2. 
Qed.

Definition ProdSpace (T1 T2:Type) 
            (L1:lattice T1) (L2:lattice T2): lattice (prod T1 T2). 
destruct L1. destruct L2.
eapply (Lattice (prodspace_facts facts0 facts1)).
Defined. 

