(*based on SynthesisXi*)
Require Export JoinLanguage. 
Require Export SynthesisIota. 
Require Export SynthesisXi. 

Fixpoint delta (C:Com) : JCom SVar :=
match C with
  Skip => (JSkip SVar)
| Assign x e => (JAssign _ x e)
| Comp C1 C2 => JComp _ (delta C1) (delta C2)
| Ite e C1 C2 => 
        JComp _ (JJoin _ (var_svar pcVar) (e::nil) ((var_svar pcVar)::nil))
                   (JIte _ e (delta C1) (delta C2))
| While e C => 
        let termRegs := (var_svar pcVar) :: nil 
        in JComp _ (JJoin _ (var_svar pcVar) (e::nil) termRegs)
                   (JWhile _ e (JComp _ (delta C)
                                        (JJoin _ (var_svar pcVar) (e::nil) termRegs)))
        end.

Lemma delta_JoinSem_Sem: forall C s G M t D N
(Hyp: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)), 
  Sem s C t. 
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)  
  inversion Hyp; subst; simpl; clear Hyp.
  constructor.  trivial. 
 (*Assign*)
  inversion Hyp; subst; simpl; clear Hyp.
  constructor.  trivial. 
(*Comp*)
  inversion Hyp; subst; simpl; clear Hyp. 
  destruct r as [[r K] P]. 
  assert (IH1 := IHC1 _ _ _ _ _ _ H2). 
  assert (IH2 := IHC2 _ _ _ _ _ _ H4). 
  econstructor.  apply IH1.  assumption. 
(*If*)
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4; simpl in *.
  (*True*) clear IHC2.
     assert (IH1 := IHC1 _ _ _ _ _ _ H6).
     apply S_True; assumption.
  (*False*) clear IHC1.
     assert (IH1 := IHC2 _ _ _ _ _ _ H6).
     apply S_False; assumption.
(*While*)
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil) (var_svar pcVar :: nil)))) as LOOP.
  remember (s, G,zupd SVar SVar_eq M (var_svar pcVar)
         (llub (lab_eval G (e :: nil)) (zeval SVar M (var_svar pcVar :: nil)))) as ST.
  remember  (t, D, N) as TT.
  assert (IH: Sem (fst (fst ST)) (While e C) (fst (fst TT))).
    clear HeqST HeqTT M G s G M.
    induction H4; intros; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
    (*True*) clear H4_0 IHJoinSem1.
             destruct r as [[r K] P].  simpl in *.
             inversion H4_; subst; clear H4_.
             inversion H5; subst; clear H5. destruct s as [[u A] Q].
             assert (IH := IHC _ _ _ _ _ _ H3).
             eapply S_WTrue; try assumption. apply IH. apply (IHJoinSem2 (eq_refl _)).
    (*False*) 
             eapply S_WFalse; trivial.
  subst; simpl in *. apply IH.
Qed.

Lemma modvars_delta: forall C,
modVars (JErase _ (delta C)) = modVars C.
Proof.
intros C. 
induction C; simpl; intros.
(*Skip*)
  reflexivity.
(*Assign*)
  trivial.
(*Comp*)
  apply extensionality. intros z.
  rewrite IHC1. rewrite IHC2. trivial.
(*Ite*)
  apply extensionality. intros z.
  rewrite IHC1. clear IHC1.
  rewrite IHC2. clear IHC2.
  trivial.
(*While*)
  apply extensionality. intros z.
  rewrite IHC. clear IHC.
  repeat rewrite Bool.orb_false_r. trivial.
Qed.

Lemma delta_controlProgVarsPreserved: forall C U
  (HU: U = delta C) 
  s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x,
  M (var_svar (progVar x)) = N (var_svar (progVar x)).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem. trivial.
(*Assign*)
  inversion HRSem; subst; clear HRSem. trivial. 
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  rewrite <- (IHC2 _ (eq_refl _) _ _ _ _ _ _ H4 x).
  apply (IHC1 _ (eq_refl _) _ _ _ _ _ _  H2 x).
(*Ite*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4; simpl in *.
     (*True*) clear H5 IHC2.
        rewrite <- (IHC1 _ (eq_refl _) _ _ _ _ _ _  H6). clear IHC1 H6.
        repeat rewrite supd_Diff; try discriminate. trivial.
     (*False*) clear H5 IHC1.
        rewrite <- (IHC2 _ (eq_refl _) _ _ _ _ _ _  H6). clear IHC2 H6.
        repeat rewrite supd_Diff; try discriminate. trivial.
(*While*)
  inversion HRSem; subst; clear HRSem. 
  inversion H2; subst; clear H2.
  assert (IHCC := IHC _ (eq_refl _)). clear IHC.
  assert (IHW : forall (s : Store) (G : CTXT) (M : ZStore SVar) (t : Store) (D : CTXT)
         (N : ZStore SVar),
       JoinSem SVar SVar_eq (s, G, M) (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil)
                (var_svar pcVar :: nil)))) (t, D, N) ->
       forall x : PVar,
       M (var_svar (progVar x)) = N (var_svar (progVar x))).
    clear H4. clear s G M t D N.
    remember (delta C) as R.
    intros s G M t D N HRS.
    remember (s, G, M) as ST.
    remember (t, D, N) as TT.
    remember (JWhile SVar e
           (JComp SVar R
              (JJoin SVar (var_svar pcVar) (e :: nil)
                 (var_svar pcVar :: nil)))) as PP.
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction HRS; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear IHHRS1. destruct r as [[r A] Q].
       inversion HRS1; subst; clear HRS1.  destruct r0 as [[u B] O].
       inversion H5; subst; clear H5. 
       rewrite (IHCC _ _ _ _ _ _ H3). clear H3 IHCC.
       assert (LP := IHHRS2 (eq_refl _)). clear IHHRS2.
       rewrite <- (LP _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear LP.
       repeat rewrite supd_Diff; try discriminate. trivial.
    (*False*)
       inversion HeqTT; subst. trivial.
  clear IHCC.
    rewrite <- (IHW _ _ _ _ _ _ H4). clear IHW H4.
    repeat rewrite supd_Diff; try discriminate. trivial.
Qed.

Lemma delta_ControlLabelsPreserved: forall C U
  (HR: U = delta C) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N))
  n,
  M (cond_svar n) = N (cond_svar n).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem. trivial.
(*Assign*)
  inversion HRSem; subst; clear HRSem. trivial. 
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  rewrite (IHC1 _ (eq_refl _) _ _ _ _ _ _ H2).
  apply (IHC2 _ (eq_refl _) _ _ _ _ _ _ H4).
(*Ite*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4; simpl in *.
  (*True*)
     clear IHC2 H5.
       rewrite <- (IHC1 _ (eq_refl _) _ _ _ _ _ _ H6). clear IHC1 H6.
       repeat rewrite supd_Diff; try discriminate. trivial.
  (*False*)
     clear IHC1 H5.
       rewrite <- (IHC2 _ (eq_refl _) _ _ _ _ _ _ H6). clear IHC2 H6.
       repeat rewrite supd_Diff; try discriminate. trivial.
(*While*)
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  assert (HC := IHC _ (eq_refl _)). clear IHC.
  remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil)
                (var_svar pcVar :: nil)))) as LOOP.
  assert (IH: forall s G M t D N,
      JoinSem _ SVar_eq (s, G, M) LOOP (t, D, N) ->
      forall n : nat, M (cond_svar n) = N (cond_svar n)).

    clear H4 s G M t D N n.
    intros s G M t D N HRS.
    remember (s, G, M) as ST.
    remember (t, D, N) as TT.
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction HRS; intros; subst; try inversion HeqLOOP; subst; clear HeqLOOP.
    (*True*)
       simpl in *. clear IHHRS1. 
       inversion HRS1; subst; clear HRS1. destruct r0 as [[u A] Q]. rename s0 into s. rename t0 into t.
       inversion H5; subst; clear H5.
       rewrite <- (IHHRS2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear IHHRS2 HRS2.
       repeat rewrite supd_Diff; try discriminate.
         apply (HC _ _ _ _ _ _ H3).
    (*False*)
       inversion HeqTT; subst. trivial.
  clear HC.
    rewrite <- (IH _ _ _ _ _ _ H4). clear IH H4.
    repeat rewrite supd_Diff; try discriminate. trivial.
Qed.

(*The claim that N \approx N in the paragraph before theorem 8 in paper.
  The claims t=t' and D=D' comparing xi with delta are below, 
  lemma delta_xi_equiv*)
Lemma delta_ZStore_sim: forall C s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)),
  ZStore_sim M N.
Proof.
intros.
intros x Hx.
destruct x.
  destruct v. exfalso. apply Hx. trivial.
  eapply (delta_controlProgVarsPreserved C _ (eq_refl _)). apply HRSem.
eapply (delta_ControlLabelsPreserved C _ (eq_refl _)). apply HRSem.
Qed.

Lemma delta_TermLabelsIncreases: forall C s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)),
  ssqsub (M (var_svar pcVar)) (N (var_svar pcVar)).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem.
  apply leq_refl.
(*Assign*)
  inversion HRSem; subst; clear HRSem. apply leq_refl.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  eapply leq_trans.
    apply (IHC1 _ _ _ _ _ _ H2).
  apply (IHC2 _ _ _ _ _ _ H4).
(*Ite*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  (*True*)
     clear IHC2 H5.
       eapply leq_snart. apply (IHC1 _ _ _ _ _ _ H6).
         clear IHC1 H6.
           rewrite supd_Same. apply lubright. apply lub_leqleft.
  (*False*)
     clear IHC1 H5.
       eapply leq_snart. apply (IHC2 _ _ _ _ _ _ H6).
         clear IHC2 H6.
           rewrite supd_Same. apply lubright. apply lub_leqleft.
(*While*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil)
                (var_svar pcVar :: nil)))) as PP.

  assert (IH: forall s G M t D N,
     JoinSem _ SVar_eq (s, G, M) PP (t, D, N) ->
     ssqsub (M (var_svar pcVar)) (N (var_svar pcVar))).
    clear t D N r E M H4. intros.
    remember (s, G, M) as ST. remember (t,D,N) as TT.
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction H; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear H1 IHJoinSem1. destruct r as [[r A] Q].
       eapply leq_snart. eapply IHJoinSem2; try reflexivity. clear IHJoinSem2. 
       inversion H0; subst; clear H0.
       inversion H6; subst; clear H6.
       rewrite supd_Same. 
       eapply leq_trans. apply (IHC _ _ _ _ _ _ H4). clear IHC H4.
       apply lubright. simpl. apply lub_leqleft.
    (*False*)
       inversion HeqTT; subst. apply leq_refl.
(*Use of loop assert *)
  eapply leq_snart. apply (IH _ _ _ _ _ _ H4). clear H4 IH IHC.
  rewrite supd_Same.
    apply lubright. apply lub_leqleft. 
Qed.

Lemma While_delta_TermLabelsIncreases: forall s G M e C t D N
     (HS: JoinSem SVar SVar_eq (s, G, M)
         (JWhile SVar e
            (JComp SVar (delta C)
               (JJoin SVar (var_svar pcVar) (e :: nil)
                  (var_svar pcVar :: nil)))) (t, D, N)),
     ssqsub (M (var_svar pcVar)) (N (var_svar pcVar)).
Proof.
intros.
remember (s, G, M) as ST.
remember (t,D,N) as TT.
remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil)
                (var_svar pcVar :: nil)))) as LOOP.
generalize dependent N. generalize dependent D. generalize dependent t. 
generalize dependent M. generalize dependent G. generalize dependent s.
induction HS; intros; try inversion HeqLOOP; subst; clear HeqLOOP H; simpl in *.
  (*True*)
     clear IHHS1 HS2.  rename s0 into s.  rename t0 into t. 
     inversion HS1; subst; clear HS1.
     inversion H4; subst; clear H4. rename s0 into r. rename G0 into K. rename M0 into P.
     apply delta_TermLabelsIncreases in H2.
     eapply leq_snart. apply (IHHS2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)).
       rewrite supd_Same. simpl.
       apply (leq_trans H2). apply lubright. apply lub_leqleft.
  (*False*)
     inversion HeqTT; subst; clear HeqTT. apply leq_refl.
Qed.

Lemma Theorem_9: forall C s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)) s'
  (EBT: eq_below (N (var_svar pcVar)) G s s'), 
  forall G' M', exists t', exists D', exists N',
    JoinSem _ SVar_eq (s',G',M') (delta C) (t',D',N') /\ 
    forall x
      (EBs: eq_below (D x) G s s'), 
      t x = t' x /\ 
      (eq_below (D x) G G G' ->  D x = D' x).
Proof.
intros C. 
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem.
  eexists; eexists; eexists.
  split. constructor.
  intros. 
  split. apply EBs. apply leq_refl.
  intros EBG.
         apply EBG. apply leq_refl.
(*Assign*)
  rename p into y.
  inversion HRSem; subst; clear HRSem.
  eexists; eexists; eexists.
  split. constructor. reflexivity. reflexivity.
  intros. 
  split. apply M_upd_split_at; intros; subst.
            rewrite M_upd_Same. rewrite M_upd_Same in EBs.
            eapply eq_below_evalE. apply EBs. apply leq_refl.
          rewrite M_upd_Diff; try congruence.
            rewrite M_upd_Diff in EBs; try congruence.
            simpl in *.
            apply EBs. apply leq_refl.
  intros EBG.
  apply M_upd_split_at; intros; subst.
            rewrite M_upd_Same. rewrite M_upd_Same in EBG.
            eapply eq_below_GevalE. apply EBG. apply leq_refl.
          rewrite M_upd_Diff; try congruence.
            rewrite M_upd_Diff in EBG; try congruence.
            simpl in *.
            apply EBG. apply leq_refl.
(*Comp*)
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  assert (IH1:= IHC1 _ _ _ _ _ _ H2). clear IHC1.
  assert (IH2:= IHC2 _ _ _ _ _ _ H4). clear IHC2.
  simpl in *.
  eapply delta_TermLabelsIncreases in H4. 
  eapply delta_TermLabelsIncreases in H2. 
  assert (EBqs: eq_below  (P (var_svar pcVar)) G s s'). eapply eq_below_mono. apply EBT. apply H4. apply leq_refl.
  destruct (IH1 _ EBqs G' M') as [r' [E' [P' [HSem1 EB1]]]]; clear IH1.
  assert (EBo: eq_below (N (var_svar pcVar)) E r r'). intros x Hx. apply EB1. eapply eq_below_mono. apply EBT. apply Hx. apply leq_refl.
  destruct (IH2 _ EBo E' P') as [t' [D' [N' [HSem2 EB2]]]]; clear IH2.
  exists t'. exists D'. exists N'. 
  split. econstructor. apply HSem1. apply HSem2.
  intros x Hx.
  split. apply EB2.
         intros z Hz. apply EB1. apply (eq_below_mono _ _ _ _ _ _ Hx Hz). apply leq_refl. 
  intros EBx.
  apply EB2.
         intros z Hz. apply EB1. apply (eq_below_mono _ _ _ _ _ _ Hx Hz). apply leq_refl.
             intros y Hy. apply EB1.
                            apply (eq_below_mono _ _ _ _ _ _ Hx Hy). apply leq_refl.
                            apply (eq_below_mono _ _ _ _ _ _ EBx Hy). apply leq_refl.
(*Ite*)
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  simpl in *. rewrite id_bbot in *.
  inversion H4; subst; clear H4; simpl in *.
  (*True*) clear IHC2.
    assert (HH:= delta_TermLabelsIncreases _ _ _ _ _ _ _ H6). rewrite supd_Same in HH.
    assert (EV: eval s e = eval s' e).
      clear IHC1.
      eapply eq_below_evalE. apply EBT.
        eapply (leq_snart HH). rewrite lab_eval_GEval. apply lub_leqleft. 
    destruct (IHC1 _ _ _ _ _ _ H6 s' EBT G' (zupd SVar SVar_eq M' (var_svar pcVar) (llub (lab_eval G' (e :: nil)) (zeval SVar M' (var_svar pcVar :: nil))))) as [t' [D' [N' [HSem' Ht']]]]; clear IHC1 H6. 
    exists t'. exists D'. exists N'.
    split. econstructor. econstructor. reflexivity. 
           apply JS_True. rewrite EV in H5; apply H5. apply HSem'.
    apply Ht'. 
  (*False*)
    clear IHC1.
    assert (HH:= delta_TermLabelsIncreases _ _ _ _ _ _ _ H6). rewrite supd_Same in HH.
    assert (EV: eval s e = eval s' e).
      clear IHC2.
      eapply eq_below_evalE. apply EBT.
        eapply (leq_snart HH). rewrite lab_eval_GEval. apply lub_leqleft. 
    destruct (IHC2 _ _ _ _ _ _ H6 s' EBT G' (zupd SVar SVar_eq M' (var_svar pcVar) (llub (lab_eval G' (e :: nil)) (zeval SVar M' (var_svar pcVar :: nil))))) as [t' [D' [N' [HSem' Ht']]]]; clear IHC2 H6. 
    exists t'. exists D'. exists N'.
    split. econstructor. econstructor. reflexivity. 
           apply JS_False. rewrite EV in H5; apply H5. apply HSem'.
    apply Ht'. 
(*While*)
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  simpl in *.
  rewrite id_bbot in *.
  remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil)
                (var_svar pcVar :: nil)))) as LOOP.
  remember (zupd SVar SVar_eq M (var_svar pcVar)
         (llub (lab_eval G (e :: nil)) (M (var_svar pcVar)))) as MM.
  remember (s, G, MM) as ST.
  remember (t, D, N) as TT.
  generalize dependent M'. generalize dependent G'. generalize dependent s'.
  generalize dependent N. generalize dependent D. generalize dependent t. 
  generalize dependent MM.
  generalize dependent M. generalize dependent G. generalize dependent s.
  induction H4; intros; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
  (*True*)
     clear IHJoinSem1.  rename s0 into s.
     inversion H4_; subst; clear H4_.
     inversion H5; subst; clear H5. rename s0 into r. rename G0 into K. rename M0 into P. rename t0 into t.
     assert (HH:= delta_TermLabelsIncreases _ _ _ _ _ _ _ H3). rewrite supd_Same in HH.
     simpl in *.
     rewrite id_bbot in *.
     apply While_delta_TermLabelsIncreases in H4_0. rewrite supd_Same in H4_0.
     assert (EBP : eq_below (P (var_svar pcVar)) G s s').
         eapply eq_below_mono. apply EBT. eapply (leq_snart H4_0). apply lub_leqright. apply leq_refl. 
     assert (IH := IHC _ _ _ _ _ _ H3 s' EBP); clear IHC H3.
     assert (EV: eval s e = eval s' e).
       eapply eq_below_evalE. apply EBT. eapply (leq_snart H4_0). apply lubright. eapply (leq_snart HH). rewrite lab_eval_GEval. apply lub_leqleft.
     destruct (IH G' (zupd SVar SVar_eq M' (var_svar pcVar) (llub (lab_eval G' (e :: nil)) (zeval SVar M' (var_svar pcVar :: nil))))) 
             as [r' [K' [P' [HSemr' Hr']]]]; clear IH.
     assert (EBr': eq_below (N (var_svar pcVar)) K r r').
       intros z Hz. apply Hr'.
         eapply (eq_below_mono _ _ _ _ _ _ EBT).
           apply Hz.
           apply leq_refl.
     assert (IHW:= IHJoinSem2 (eq_refl _) _ _ _ _ (eq_refl _) (eq_refl _) _ _ _ (eq_refl _) r' EBr' K' P'); clear IHJoinSem2. 
     destruct IHW as [t' [D' [N' [HSemt' Ht']]]].
     inversion HSemt'; subst; clear HSemt'.
     inversion H3; subst; clear H3.
     eexists. eexists. eexists.
       split. econstructor. econstructor.  reflexivity.
       eapply JS_WTrue.
         rewrite EV in H; apply H.
         econstructor.   
           apply HSemr'.
           constructor. reflexivity.
         apply H5.
     intros.
     assert (EBr2: eq_below (D x) K r r').
        intros z Hz. apply Hr'.
          eapply eq_below_mono. apply EBs. apply Hz. apply leq_refl. 
      split; intros; apply Ht'.
        apply EBr2. apply EBr2.
        intros z Hz. apply Hr'. 
          eapply eq_below_mono. apply EBs. apply Hz. apply leq_refl.
          eapply eq_below_mono. apply H0. apply Hz. apply leq_refl.
(*False*)
  inversion HeqTT; subst; clear HeqTT. rewrite supd_Same in EBT.
  rename t0 into t. clear IHC.
  assert (EV: eval t e = eval s' e).
       eapply eq_below_evalE. apply EBT. rewrite lab_eval_GEval. apply lub_leqleft.
  eexists. eexists. eexists.
  split. econstructor. econstructor; reflexivity.
         eapply JS_WFalse. rewrite EV in H. apply H.
                           reflexivity.
  intros.
  split. apply EBs. apply leq_refl. 
  intros. apply H0. apply leq_refl.
Qed.    

(*A variation that includes a ZStore-sim guarantee*)
Lemma Theorem_9_SZtore_sim: forall C s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)) s'
  (EBT: eq_below (N (var_svar pcVar)) G s s'), 
  forall G' M', exists t', exists D', exists N',
    JoinSem _ SVar_eq (s',G',M') (delta C) (t',D',N') /\ 
    (ZStore_sim M M' -> ZStore_sim N N') /\
    forall x
      (EBs: eq_below (D x) G s s'), 
      t x = t' x /\ 
      (eq_below (D x) G G G' ->  D x = D' x).
Proof.
intros. subst.
destruct (Theorem_9 _ _ _ _ _ _ _ HRSem _ EBT G' M') as [t' [D' [N' [HRSem' H']]]].
exists t'. exists D'. exists N'.
split; trivial.
split; trivial.
intros.
apply delta_ZStore_sim in HRSem.
apply delta_ZStore_sim in HRSem'.
eapply ZStore_sim_trans. eapply ZStore_sim_sym. apply HRSem.
eapply ZStore_sim_trans. apply H. apply HRSem'.
Qed.

Lemma Theorem_9_Sem: forall C s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)) s'
  (EBT: eq_below (N (var_svar pcVar)) G s s'), 
  exists t', 
   (Sem s' C t' /\ 
    forall x
      (EBX: eq_below (D x) G s s'),
      t x = t' x).
Proof.
intros. 
destruct (Theorem_9 _ _ _ _ _ _ _ HRSem _ EBT G M) as [t' [D' [N' [HJSem' HT]]]].
exists t'.
split. eapply delta_JoinSem_Sem. apply HJSem'.
apply HT.
Qed.   

(*The result of Corollary2*)
Lemma Theorem_9_Sem_kappa: forall C s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N)) s'
  (EBT: eq_below (N (var_svar pcVar)) G s s'), 
  exists t', 
   Sem s' C t' /\ 
    forall kappa
      (EB: eq_below kappa G s s'),eq_below kappa D t t'.
Proof.
intros. 
destruct (Theorem_9_Sem _ _ _ _ _ _ _ HRSem _ EBT) as [t' [HJSem' HT]].
exists t'.
split. assumption.
intros kappa Hkappa x Hx.
  apply HT. eapply eq_below_mono. apply Hkappa. apply Hx. apply leq_refl.
Qed.

Definition lift (kappa:L) : ZStore SVar := fun z =>
match z with
  var_svar v => match v with
                  pcVar => kappa
                | progVar x => bbot
                end
| cond_svar n => bbot
end.

Lemma lift_mkSVars: forall p e, zeval SVar (lift p) (mkSVars e) = bbot.
Proof.
intros.
induction e; simpl; try reflexivity.
  apply bbot_id.
rewrite zeval_app. rewrite IHe1. rewrite IHe2.
  apply bbot_id.
Qed.

Lemma lift_zupd: forall G e p,
zupd SVar SVar_eq (lift p) (var_svar pcVar)
         (llub (lab_eval G (e :: nil)) p) 
= lift (llub (GEval G e) p).
Proof.
intros. 
apply extensionality. intros z.
destruct z.  
   destruct v. rewrite supd_Same. unfold lab_eval. simpl. repeat rewrite id_bbot. reflexivity.
               rewrite supd_Diff; simpl. reflexivity. congruence.
   rewrite supd_Diff; simpl. reflexivity. congruence.
Qed.

Lemma lift_zupd': forall G e p,
zupd SVar SVar_eq (lift p) (var_svar pcVar)
         (llub (lab_eval G (e :: nil)) (llub p bbot)) 
= lift (llub (GEval G e) p).
Proof.
intros.
rewrite id_bbot.
apply lift_zupd.
Qed.

Lemma lift_preserved: forall C s G p t D N
  (HRSem: JoinSem _ SVar_eq (s,G,lift p) (delta C) (t,D,N)),
  exists q, N = lift q /\ ssqsub p q.
Proof.
intros.
exists (N (var_svar pcVar)).
assert (N1 := delta_TermLabelsIncreases C _ _ _ _ _ _ HRSem). simpl in N1.
assert (N2:= delta_ControlLabelsPreserved _ _ (eq_refl _) _ _ _ _ _ _ HRSem).
assert (N3:= delta_controlProgVarsPreserved _ _ (eq_refl _) _ _ _ _ _ _ HRSem).
clear HRSem.
split; trivial.
apply extensionality. intros z.
destruct z; simpl.
  destruct v; simpl; trivial.
  rewrite <- N3. reflexivity.
rewrite <- N2. reflexivity.
Qed.

Lemma While_delta_TermLabelsIncreases_lift: forall s G q e C t D N
     (HS: JoinSem SVar SVar_eq (s, G, lift q)
         (JWhile SVar e
            (JComp SVar (delta C)
               (JJoin SVar (var_svar pcVar) (e :: nil)
                  (var_svar pcVar :: nil)))) (t, D, N)),
     ssqsub q (N (var_svar pcVar)).
Proof.
intros.
apply While_delta_TermLabelsIncreases in HS. simpl in HS. assumption.
Qed.

Lemma Theorem_9_Sem_lift: forall C s G p t D N
  (HRSem: JoinSem _ SVar_eq (s,G,lift p) (delta C) (t,D,N)) s'
  (EBT: eq_below (N (var_svar pcVar)) G s s'), 
  exists t', 
   (Sem s' C t' /\ 
    forall x
      (EBX: eq_below (D x) G s s'),
      t x = t' x).
Proof.
intros.
apply (Theorem_9_Sem _ _ _ _ _ _ _ HRSem _ EBT). 
Qed.

(*Here's a slightly weaker form of Theorem 8_Sem_lift*) 
Lemma Theorem_9_Sem_lift_weak: forall C s G p t D N
  (HRSem: JoinSem _ SVar_eq (s,G,lift p) (delta C) (t,D,N)) s'
  (EBT: eq_below (N (var_svar pcVar)) G s s'), 
  exists t', 
   (Sem s' C t' /\ 
    forall x
      (EBX: eq_below (llub (D x)  (N (var_svar pcVar))) G s s'),
      t x = t' x).
Proof.
intros.
destruct (Theorem_9_Sem _ _ _ _ _ _ _ HRSem s' EBT) as [t' [Semt Ht]].
exists t'.
split; trivial.
intros x Hx. apply Ht.
  eapply eq_below_mono. apply Hx. apply lub_leqleft. apply leq_refl. 
Qed.


(*Agreement of delta and xi on all variables and data taints.
  (Comment in paper just before Theorem8)*)
Theorem delta_xi_equiv: forall C F s G M t D N
  (Hyp: JoinSem _ SVar_eq (s,G,M) (delta C) (t,D,N))
   M' t' D' N'
  (Hyp': JoinSem _ SVar_eq (s, G, M') (xi C F) (t', D', N')), 
  t=t' /\ D=D'.
Proof.
intros C.
induction C; simpl; intros; subst.
(*Skip*)
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'.
  split; trivial.
(*Assign*)
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  split; trivial.
(*Comp*)
  inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [[r' K'] P'].
  destruct (IHC1 _ _ _ _ _ _ _ H2 _ _ _ _ H3) as [IH1a IH1b]. subst.
  apply (IHC2 _ _ _ _ _ _ _ H4 _ _ _ _ H6). 
(*Ite*)
  remember (fresh F) as Fr. destruct Fr as [n Fn].
  inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
  inversion H2; subst; clear H2.
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [[r' K'] P'].
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. 
  inversion H2; subst; clear H2. simpl in *.
  inversion H4; subst; clear H4; simpl in *.
  (*True*) clear IHC2.
     inversion H6; subst; clear H6; simpl in *.
     (*True*)
        inversion H8; subst; clear H8. destruct r as [[u' G'] P'].
        destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
        apply (IHC1 _ _ _ _ _ _ _ H7 _ _ _ _ H2).
     (*False*) rewrite H4 in H5. inversion H5.
  (*False*) clear IHC1.
     inversion H6; subst; clear H6; simpl in *.
     (*True*) rewrite H4 in H5. inversion H5.
     (*False*)
        inversion H8; subst; clear H8. destruct r as [[u' G'] P'].
        destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
        apply (IHC2 _ _ _ _ _ _ _ H7 _ _ _ _ H2).
(*While*)
  remember (fresh F) as Fr. destruct Fr as [n Fn].
  inversion Hyp; subst; clear Hyp.
  inversion H2; subst; clear H2.
  inversion Hyp'; subst; clear Hyp'. destruct r as [[r' K'] P'].
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r as [[u' E'] Q'].
  inversion H2; subst; clear H2.
  inversion H6; subst; clear H6. destruct r as [[r' K'] P'].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
  rename r' into t'. rename K' into D'. rename P' into N'.
  remember (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e)))))) as LOOP'.
  remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil) (var_svar pcVar :: nil)))) as LOOP.
  assert (IH: forall s G M t D N (HL: JoinSem SVar SVar_eq (s, G, M) LOOP (t, D, N))
                 M' t' D' N' (HL': JoinSem SVar SVar_eq (s, G, M') LOOP' (t', D', N')),
                 t=t' /\ D=D').
     clear H2 H4 u' E' M M' t D N t' D' N'.
     intros s G M t D N Hyp.
     remember (s,G,M) as ST. remember (t,D,N) as TT.
     generalize dependent N. generalize dependent D. generalize dependent t.
     generalize dependent M. generalize dependent G. generalize dependent s.
     induction Hyp; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
     (*True*) clear IHHyp1 Hyp2. destruct s as [[s G] M]. destruct r as [[u K] P]. destruct t as [[t D] N]. simpl in *.
              intros.
              apply eq_sym in HeqST. inversion HeqST; subst; clear HeqST.
              apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT.
              assert (IH := IHHyp2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear IHHyp2.
              inversion Hyp1; subst; clear Hyp1.
              inversion H5; subst; clear H5. rename M0 into P.
              inversion HL'; subst; clear HL'; simpl in *.
              (*True*) 
                 inversion H5; subst; clear H5.
                 destruct r0 as [[v' A'] Q'].
                 inversion H9; subst; clear H9.
                 inversion H10; subst; clear H10. 
                 inversion H8; subst; clear H8.
                 inversion H11; subst; clear H11. 
                 destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
                 destruct (IHC _ _ _ _ _ _ _ H3 _ _ _ _ H6) as [Z1 Z3]; clear IHC H3 H6.
                 apply eq_sym in Z1; apply eq_sym in Z3. subst.
                 apply (IH _ _ _ _ H7); clear IH H7.
              (*False*) rewrite H4 in H. inversion H.
     (*False*) intros; subst. apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT. simpl in *.
               inversion HL'; subst; clear HL'; simpl in *.
               (*True*) rewrite H2 in H. inversion H.
               (*False*) apply eq_sym in H5; inversion H5; subst; clear H5.
                         split; trivial.
  clear IHC. subst.
  apply (IH _ _ _ _ _ _ H4 _ _ _ _ H2). 
Qed. 


Definition ZStore_sub (M M':ZStore SVar) :=
forall z, ssqsub (M z) (M' z).

Lemma ZStore_sub_refl: forall M, ZStore_sub M M.
Proof.
intros. intros z. apply leq_refl.
Qed.

Lemma ZStore_sub_trans: forall {M P} N 
(HMN:ZStore_sub M N) (HNP: ZStore_sub N P),
ZStore_sub M P.
Proof.
intros. intros z.
eapply leq_trans. apply (HMN z). apply (HNP z).
Qed.

Lemma MERGE_lift: forall G p, MERGE G (lift p) = G.
Proof.
intros. 
unfold MERGE; simpl. apply extensionality. intros z. apply id_bbot.
Qed.

Lemma MERGE_pvar_bbot: forall G N (H:forall z, N (var_svar (progVar z)) = bbot),
MERGE G N = G.
Proof.
intros.
apply extensionality. intros.
unfold MERGE; simpl. rewrite H. apply id_bbot.
Qed.

Lemma lift_zupd_pcVar: forall p q, zupd _ SVar_eq (lift p) (var_svar  pcVar) q = lift q.
Proof.
intros.
apply extensionality. intros z.
apply supd_split_at; intros. subst. reflexivity. 
destruct z; simpl; trivial.
destruct v; simpl; trivial. congruence.
Qed.

Lemma zeval_lift_mkSVars: forall p e, zeval _ (lift p) (mkSVars e) = bbot.
Proof.
intros.
induction e; simpl. reflexivity.
apply bbot_id.
rewrite zeval_app. rewrite IHe1. rewrite IHe2. apply bbot_id.
Qed.

Lemma zeval_lift_allPCs: forall p F, zeval _ (lift p) (allPCs F)  = bbot.
Proof.
intros.
induction F; simpl. reflexivity.
rewrite IHF.
apply bbot_id.
Qed.


(*Agreement of delta and xi on all variables and data taints, extended
  by claims regarding improved precision of path-data tracking than
  path-control-data tracking ert eqbelow.*)
Theorem Theorem8: forall C F s G pc t D N
  (Hyp: JoinSem _ SVar_eq (s,G,lift pc) (delta C) (t,D,N))
   M' t' D' N'
  (Hyp': JoinSem _ SVar_eq (s, G, M') (xi C F) (t', D', N'))
  (pcSub: ssqsub pc (M' (var_svar (pcVar)))), 
  t=t' /\ D=D' /\ ZStore_sub N N' /\
  (forall A u u' kappa, 
      @eq_below A kappa D u u' = @eq_below A kappa (MERGE D N) u u' /\
      (@eq_below A kappa D u u' -> @eq_below A kappa (MERGE D' N') u u')).
Proof.
intros C.
induction C; simpl; intros; subst.
(*Skip*)
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'.
  rewrite MERGE_lift.
  split. trivial.
  split. trivial.
  split. intros z. destruct z; simpl.
              destruct v; simpl. assumption.
                  apply bot_leq.
            apply bot_leq.      
  intros. split; trivial. 
          intros. eapply eq_below_mono. apply H. apply leq_refl. apply MERGE_L.
(*Assign*)
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  rewrite MERGE_lift.
  split. trivial.
  split. trivial.
  split. intros z. destruct z; simpl.
              destruct v; simpl. rewrite supd_Diff. assumption. congruence.
                  apply bot_leq.
            apply bot_leq.
  intros. split; trivial. 
          intros. eapply eq_below_mono. apply H. apply leq_refl. apply MERGE_L.
(*Comp*)
  inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [[r' K'] P'].
  destruct (IHC1 _ _ _ _ _ _ _ H2 _ _ _ _ H3 pcSub) as [IH1a [IH1b [IH1c IH1d]]]. subst. clear IHC1.
  destruct (lift_preserved _ _ _ _ _ _ _ H2) as [q [Hq1 Hq2]]; subst.
  destruct (IHC2 _ _ _ _ _ _ _ H4 _ _ _ _ H6) as [IH2a [IH2b [IH2c IH2d]]]. clear IHC2. 
  assert (Z:= IH1c (var_svar pcVar)). apply Z.
  subst.
  split. trivial.
  split. trivial.
  split; trivial.
(*Ite*)
  remember (fresh F) as Fr. destruct Fr as [n Fn].
  inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
  inversion H2; subst; clear H2. simpl in H4. rewrite id_bbot in H4.
  rewrite lift_zupd in H4.
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [[r' K'] P'].
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. 
  inversion H2; subst; clear H2. simpl in *.
  inversion H4; subst; clear H4; simpl in *.
  (*True*) clear IHC2.
     inversion H6; subst; clear H6; simpl in *.
     (*True*)
        inversion H8; subst; clear H8. destruct r as [[u' G'] P'].
        destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
        destruct (IHC1 _ _ _ _ _ _ _ H7 _ _ _ _ H2) as [Z1 [Z2 [Z3 Z4]]]; clear H2 H7 IHC1. 
          rewrite supd_Diff; try congruence.
          rewrite supd_Same. rewrite lab_eval_GEval. apply lub_leq. apply lub_leqleft. apply lubright. apply lubleft; assumption.
        subst. 
        split; trivial.
        split; trivial.
        split. intros z; destruct z; simpl. destruct v. apply Z3.
                      eapply leq_trans. apply Z3.
                      remember (in_dec pvar_eq p (modVarsL C2)) as b; destruct b; subst.
                           apply lub_leqright.
                           apply leq_refl.
                  apply Z3.        
        intros A s s' kappa.
          destruct (Z4 _ s s' kappa) as [HEB Z6]; clear Z4.
          split; trivial.
          intros. 
            eapply eq_below_mono. apply (Z6 H). apply leq_refl.
            apply MERGE_monoright.
            intros. destruct z; try apply leq_refl.
                    destruct v; try apply leq_refl.
                      remember (in_dec pvar_eq p (modVarsL C2)) as b; destruct b; subst.
                           apply lub_leqright.
                           apply leq_refl.  
     (*False*) rewrite H4 in H5. inversion H5.
  (*False*) clear IHC1.
     inversion H6; subst; clear H6; simpl in *.
     (*True*) rewrite H4 in H5. inversion H5.
     (*False*)
        inversion H8; subst; clear H8. destruct r as [[u' G'] P'].
        destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
        destruct (IHC2 _ _ _ _ _ _ _ H7 _ _ _ _ H2) as [Z1 [Z2 [Z3 Z4]]]; clear H2 H7 IHC2. 
          rewrite supd_Diff; try congruence.
          rewrite supd_Same. rewrite lab_eval_GEval. apply lub_leq. apply lub_leqleft. apply lubright. apply lubleft; assumption.
        subst. 
        split; trivial.
        split; trivial.
        split. intros z; destruct z; simpl. destruct v. apply Z3.
                      eapply leq_trans. apply Z3.
                      remember (in_dec pvar_eq p (modVarsL C1)) as b; destruct b; subst.
                           apply lub_leqright.
                           apply leq_refl.
                  apply Z3.         
        intros A s s' kappa.
          destruct (Z4 _ s s' kappa) as [Z5 Z6]; clear Z4.
          split; trivial.
          intros.
            eapply eq_below_mono. apply (Z6 H). apply leq_refl.
            apply MERGE_monoright.
            intros. destruct z; try apply leq_refl.
                    destruct v; try apply leq_refl.
                      remember (in_dec pvar_eq p (modVarsL C1)) as b; destruct b; subst.
                           apply lub_leqright.
                           apply leq_refl.       
(*While*)
  remember (fresh F) as Fr. destruct Fr as [n Fn].
  inversion Hyp; subst; clear Hyp.
  inversion H2; subst; clear H2.
  simpl in H4. rewrite lift_zupd' in H4.
  inversion Hyp'; subst; clear Hyp'. destruct r as [[r' K'] P'].
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r as [[u' E'] Q'].
  inversion H2; subst; clear H2.
  inversion H6; subst; clear H6. destruct r as [[r' K'] P'].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
  rename r' into t'. rename K' into D'. rename P' into N'.
  remember (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e)))))) as LOOP'.
  remember (JWhile SVar e
          (JComp SVar (delta C)
             (JJoin SVar (var_svar pcVar) (e :: nil) (var_svar pcVar :: nil)))) as LOOP.
  assert (IH: forall s G p t D N (HL: JoinSem SVar SVar_eq (s, G, lift p) LOOP (t, D, N))
                 M' t' D' N' (HL': JoinSem SVar SVar_eq (s, G, M') LOOP' (t', D', N'))
                 (Hp : ssqsub p (M' (var_svar pcVar))), t=t' /\ D=D' /\ ZStore_sub N N' /\
                (forall (A : Type) (u u' : PVar -> A) (kappa : L),
                    eq_below kappa D u u' = eq_below kappa (MERGE D N) u u' /\
                    (eq_below kappa D u u' -> eq_below kappa (MERGE D' N') u u'))).
     clear H2 H4 u' E' pcSub pc M' t D N t' D' N'.
     intros s G p t D N Hyp.
     remember (s,G,lift p) as ST. remember (t,D,N) as TT.
     generalize dependent N. generalize dependent D. generalize dependent t.
     generalize dependent p. generalize dependent G. generalize dependent s.
     induction Hyp; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
     (*True*) clear IHHyp1 Hyp2. destruct s as [[s G] M]. destruct r as [[u K] P]. destruct t as [[t D] N]. simpl in *.
              intros.
              apply eq_sym in HeqST. inversion HeqST; subst; clear HeqST.
              apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT.
              inversion Hyp1; subst; clear Hyp1.
              inversion H5; subst; clear H5.
              destruct (lift_preserved _ _ _ _ _ _ _ H3) as [q [Hq1 Hq2]]. subst.
              simpl in IHHyp2. rewrite lift_zupd' in IHHyp2.
              assert (IH := IHHyp2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear IHHyp2.
              inversion HL'; subst; clear HL'; simpl in *.
              (*True*) 
                 inversion H5; subst; clear H5.
                 destruct r0 as [[v' A'] Q'].
                 inversion H9; subst; clear H9.
                 inversion H10; subst; clear H10. 
                 inversion H8; subst; clear H8.
                 inversion H11; subst; clear H11. 
                 destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
                 destruct (IHC _ _ _ _ _ _ _ H3 _ _ _ _ H6 Hp) as [Z1 [Z2 [Z3 Z4]]]; clear IHC H3 H6.
                 apply eq_sym in Z1; apply eq_sym in Z2. subst.
                 apply (IH _ _ _ _ H7); clear IH H7.
                 rewrite supd_Diff; try congruence.
                 rewrite supd_Same. rewrite lab_eval_GEval.
                 apply lub_leq. apply lub_leqleft.
                 assert (Z:= Z3 (var_svar pcVar)). simpl in Z.
                 eapply leq_trans. apply Z. apply lubright. apply lub_leqleft.
              (*False*) rewrite H4 in H. inversion H.
     (*False*) intros; subst. apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT. simpl in *.
               inversion HL'; subst; clear HL'; simpl in *.
               (*True*) rewrite H2 in H. inversion H.
               (*False*) apply eq_sym in H5; inversion H5; subst; clear H5.
                         split; trivial.
                         split; trivial.
                         split. intros z. destruct z; simpl. destruct v; simpl. apply Hp. apply bot_leq. apply bot_leq.
                         rewrite MERGE_lift. 
                         intros. split; trivial.
                            intros. eapply eq_below_mono. apply H0. apply leq_refl. apply MERGE_L.
  clear IHC. subst.
  destruct (IH _ _ _ _ _ _ H4 _ _ _ _ H2) as [Z1 [Z2 [Z3 Z4]]]; clear IH H4 H2.
     rewrite supd_Diff; try congruence. rewrite supd_Same.
     rewrite lab_eval_GEval; simpl.
     apply leq_monoright. apply lubleft. assumption. 
  subst.
  split; trivial.
  split; trivial.
  split. intros z; destruct z; simpl. destruct v. apply Z3.
                      eapply leq_trans. apply Z3.
                      remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; subst.
                           apply lub_leqright.
                           apply leq_refl.
                  apply Z3.       
  intros A s s' kappa.
  destruct (Z4 _ s s' kappa) as [Z5 Z6].
  split; trivial.
  intros. eapply eq_below_mono. apply (Z6 H). apply leq_refl.
          apply MERGE_monoright.
            intros. destruct z; try apply leq_refl.
                    destruct v; try apply leq_refl.
                      remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; subst.
                           apply lub_leqright.
                           apply leq_refl.       
Qed. 
