Require Export TaintLanguage.
Require Export FPC.

Section SynthesisTheta. (*based on mkRifleProg.v, where theta was called mkR*)
Fixpoint theta (C:Com) (G H:FPCCTXT) (tw:FPCT G C H): TCom. 
destruct tw. 
  apply TSkip. 
(* mkRifleProg has this: apply (TAssign (cons (FPCGEval G (progExpr E)) (cons p nil)) x E). 
    But it suffices to only annotate with p*)
  apply (TAssign (cons p nil) x E). 
  apply (TComp (theta _ _ _ tw1) (theta _ _ _ tw2)). 
  apply (TComp (TIte E (theta _ _ _ tw1) (theta _ _ _ tw2)) (CompCode (llub p a) (Ite E C1 C2))).
  apply (TComp (TWhile E (theta _ _ _ (HypC N)))
            (CompCode (llub p (FPCGEval (pre N) (progExpr E))) C)).
Defined. 

Lemma theta_unique: forall C G H (tw:FPCT G C H) G' C' H' 
      (tw':FPCT G' C' H') (HG:G' = G) (HC: C' = C), 
      theta C G H tw = theta C' G' H' tw'.
Proof.
intros C G H tw.
induction tw; simpl.
(*Skip*)
intros G' C' H' tw'.
induction tw'; simpl; intros; try inversion HC. subst. trivial.

(*Assign*)
intros G' C' H' tw'. 
induction tw'; simpl; intros; try inversion HC. subst. trivial.

(*Comp*)
intros G' C' H'. 
induction tw'; simpl; intros; try inversion HC. subst.
clear IHtw'1 IHtw'2.
assert (G3 = G1). eapply FPCT_unique. apply tw'1. apply tw1. subst.
assert (G4 = G2). eapply FPCT_unique. apply tw'2. apply tw2. subst.
rewrite (IHtw1 _ _ _ tw'1 (eq_refl _) (eq_refl _)).
rewrite (IHtw2 _ _ _ tw'2 (eq_refl _) (eq_refl _)).
trivial.

(*Ite*)
intros G' C' H'. 
induction tw'; simpl; intros; try inversion HC. subst.
clear IHtw'1 IHtw'2.
assert (G3 = G1). eapply FPCT_unique. apply tw'1. apply tw1. subst.
assert (G4 = G2). eapply FPCT_unique. apply tw'2. apply tw2. subst.
rewrite (IHtw1 _ _ _ tw'1 (eq_refl _) (eq_refl _)).
rewrite (IHtw2 _ _ _ tw'2 (eq_refl _) (eq_refl _)).
trivial.

(*While*)
intros G' C' H'. 
induction tw'; simpl; intros; try inversion HC. subst.
clear H2.
assert (HN := H0 N). clear H0.
  (*Following 3 asserts are from FPC_uniqe*)
    assert (PRE: forall i, N <= i -> pre N = pre i). 
      intros i. induction i; intros. destruct N. trivial. apply False_ind. eapply Le.le_Sn_0. apply H.
      destruct N. rewrite IHi. apply PreAboveN. apply Le.le_0_n. apply Le.le_0_n.
      apply Le.le_S_n in H. 
      destruct (Lt.le_lt_or_eq _ _ H). 
         apply Lt.lt_le_S in H0. rewrite (IHi H0).  apply PreAboveN. apply H0. 
         rewrite H0. trivial. 
    assert (PRE': forall i, N0 <= i -> pre0 N0 = pre0 i).
      intros i. induction i; intros. destruct N0. trivial. apply False_ind. eapply Le.le_Sn_0. apply H.
      destruct N0. rewrite IHi. apply PreAboveN0. apply Le.le_0_n. apply Le.le_0_n.
      apply Le.le_S_n in H. 
      destruct (Lt.le_lt_or_eq _ _ H). 
         apply Lt.lt_le_S in H0. rewrite (IHi H0).  apply PreAboveN0. apply H0. 
         rewrite H0. trivial. 
    assert (forall i, (pre i) = (pre0 i) /\ (post i) = (post0 i)).
      intros i.
      induction i; intros; subst; simpl.
        assert (pre 0 = pre0 0).
          rewrite PreZ. rewrite PreZ0. trivial.
        split. assumption.
        eapply FPCT_unique. apply HypC. rewrite H. apply HypC0.
      destruct IHi as [IHpre IHpost].
        assert (pre (S i) = pre0 (S i)).
          rewrite PrePost. rewrite PrePost0. repeat rewrite IHpost. trivial.
        split. assumption.
        eapply FPCT_unique. apply HypC. rewrite H. apply HypC0.
    assert (POST: forall i, N <= i -> post N = post i).
      intros i Hi. eapply FPCT_unique. apply HypC. rewrite (PRE i Hi). apply HypC.
    assert (POST': forall i, N0 <= i -> post0 N0 = post0 i).
      intros i Hi. eapply FPCT_unique. apply HypC0. rewrite (PRE' i Hi). apply HypC0.  
    assert (preMX : pre N = pre (MinMax.max N N0)).
      apply PRE. apply MinMax.MMP.le_max_l. 
    assert (preMX' : pre0 N0 = pre0 (MinMax.max N N0)).
      apply PRE'. apply MinMax.MMP.le_max_r.
    assert (postMX : post N = post (MinMax.max N N0)).
      apply POST. apply MinMax.MMP.le_max_l. 
    assert (postMX' : post0 N0 = post0 (MinMax.max N N0)).
      apply POST'. apply MinMax.MMP.le_max_r.
    remember  (llub (G pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) as p.
    remember (llub (G pcVar) (GEval (fun x : PVar => pre0 N0 (progVar x)) E)) as p'.
    assert (p = p').
      rewrite preMX' in Heqp'. rewrite preMX in Heqp. subst.
      assert ((fun x : PVar => pre (MinMax.max N N0) (progVar x)) = (fun x : PVar => pre0 (MinMax.max N N0) (progVar x))).
        apply extensionality.  intros z. destruct (H (MinMax.max N N0)). rewrite H0. trivial. 
      rewrite H0. trivial. 
    rewrite H0. 

    cut (theta C (pre N) (post N) (HypC N) = theta C (pre0 N0) (post0 N0) (HypC0 N0)).
      intros X. rewrite X. trivial.
    clear H0.
    apply HN. rewrite preMX. rewrite preMX'.  destruct (H (MinMax.max N N0)). rewrite H0. trivial.
    trivial.
Qed. 

Lemma modVars_theta: forall C G H tw x, modVars (TErase (theta C G H tw)) x = modVars C x. 
Proof. 
intros. 
induction tw; simpl in *; trivial. 
(*Comp*)
rewrite IHtw1.  rewrite IHtw2.  trivial.
(*Ite*) 
rewrite IHtw1.  rewrite IHtw2. clear IHtw1 IHtw2 tw1 tw2. 
  rewrite CompCode_modVars. simpl.
  repeat rewrite <- Bool.orb_assoc.
  remember (modVars C1 x) as b; destruct b; simpl; trivial.
  remember (modVars C2 x) as bb; destruct bb; simpl; trivial.
rewrite (H0 N).
  rewrite CompCode_modVars.
  apply orb_refl.
Qed.

Lemma theta_TntSem_Sem: forall C Gamma Delta (tw:FPCT Gamma C Delta) s G t D
(Hyp: TntSem (s,G) (theta C Gamma Delta tw) (t,D)), 
  Sem s C t. 
Proof.
intros C Gamma Delta tw. 
induction tw; intros; subst; rename G into Gamma; rename G0 into G;
  inversion Hyp; subst; simpl; clear Hyp.  
(*Skip*)  
  inversion H; subst; clear H. 
  constructor.  trivial. 
 (*Assign*)
  constructor.  trivial. 
(*Comp*)
  destruct r as [r H]. 
  assert (IH1 := IHtw1 _ _ _ _ H2). 
  assert (IH2 := IHtw2 _ _ _ _ H4). 
  econstructor.  apply IH1.  assumption. 
(*If*)
  destruct r as [r H]. 
  cut (r = t).  intros Htr; subst.  clear H4. 
    inversion H2; subst; simpl in *; clear H2. 
    assert (IH1 := IHtw1 _ _ _ _ H7). 
    apply S_True.  apply H6.  assumption. 
  
    assert (IH1 := IHtw2 _ _ _ _ H7). 
    apply S_False.  apply H6.  assumption. 

  apply extensionality.  intros z.  
  destruct (CompCode_Lemma0 _ _ _ _ _ _ H4). subst. trivial.
(*While*)
  destruct r as [r H]. 
  assert (r = t).  
    apply extensionality.  intros z.  
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H5). subst. trivial.
 subst. clear H5 D.
 assert (HN := H0 N).
 remember (t, H) as tt. generalize dependent H. generalize dependent t.
 remember (s,G) as ss. generalize dependent G. generalize dependent s.
 remember (TWhile E (theta C (pre N) (post N) (HypC N))) as CC.
 generalize dependent C.
 clear PreBelowN.
 
 induction H3; intros; subst; try inversion HeqCC; subst; clear HeqCC; rename H into Cond.
 (*WhileTrue*) 
    clear IHTntSem1. rename s0 into s. rename t0 into t. rename H1 into H. simpl in*.
    destruct r as [r D].
    eapply S_WTrue. apply Cond. 
      eapply HN. apply H3_.
      eapply IHTntSem2.
        apply H0.
        trivial.
        apply HN.
        reflexivity.
        reflexivity.
  (*WhileFalse*)
    inversion Heqtt; subst; clear Heqtt.
    eapply S_WFalse. apply Cond. trivial.
Qed.


(*Tainted execution yields a terminal tag component that respects static post-type*)
Lemma Lemma2: forall C Gamma Delta (tw:FPCT Gamma C Delta) s G t D
(Hyp: TntSem (s,G) (theta C Gamma Delta tw) (t,D))
(HG: SQSUB G (FPCCTXT2CTXT Gamma)),  SQSUB D (FPCCTXT2CTXT Delta). 
Proof. 
intros C Gamma Delta tw. 
induction tw; intros; subst; simpl in *. 
(*Skip*)
inversion Hyp; subst; simpl; clear Hyp. 
  inversion H; subst; clear H. assumption. 
(*Assign*)
inversion Hyp; subst; simpl; clear Hyp.
  autorewrite with fpcREWRITE. rewrite id_bbot. 
  rewrite (llub_com (G pcVar)).
  intros y.
  remember (pvar_eq x y) as b; destruct b; subst; simpl.
    repeat rewrite M_upd_Same.
    apply leq_monoleft. apply GEval_mono. apply HG.
  repeat rewrite M_upd_Diff; trivial.
(*Comp*)
inversion Hyp; subst; simpl; clear Hyp.  destruct r as [r H]. 
  assert (HH1 := IHtw1 _ _ _ _ H2 HG).  clear IHtw1. 
  eapply IHtw2.  apply H4.  assumption. 
(*Ite*)
inversion Hyp; subst; simpl; clear Hyp.  destruct r as [r H]. 
  inversion H2; subst; clear H2. 
  (*True*) clear IHtw2. 
     assert (HH1 := IHtw1 _ _ _ _ H8 HG); clear IHtw1.
     intros y.
     assert (Z:= CompCode_Lemma _ _ _ _ _ y _ H4); clear H4.
     destruct Z. 
     destruct H1.
         clear H8. destruct H1. rewrite H1; clear H1 H0.
         autorewrite with fpcREWRITE.
         unfold FPCCTXT2CTXT, FLUB; simpl.
         remember (modVars C1 y) as b; destruct b; subst; simpl in *; apply eq_sym in Heqb.
           apply lubleft.   
           apply lub_leq. apply HH1.                 
           apply (FPC_Mod _ _ _ y (FPCT_FPC _ _ _ tw1) Heqb). clear tw1.
         rewrite Heqb in *; simpl in *.
         apply lub_leq.
            apply lubleft. apply HH1.
            apply lubright. apply (FPC_Mod _ _ _ y (FPCT_FPC _ _ _ tw2) H2).
       destruct H1; simpl in *. apply Bool.orb_false_iff in H2. destruct H2.
         rewrite H1. 
         destruct (TSem1 _ _ _ _ _ H8 y) as [_ Z].
            rewrite modVars_theta. assumption.
          rewrite <- Z. clear H8 tw2 Z. 
          assert (Z:= FPCT_FPC _ _ _ tw1).
          assert (ZZ:= FPC_Nomod _ _ _ y Z H2).
          rewrite FPCCTXT2CTXT_upd_pcVar.
          rewrite M_upd_Diff in ZZ; try discriminate.
          eapply leq_trans. apply HG. unfold FLUB, FPCCTXT2CTXT. rewrite <- ZZ.
          simpl. apply lub_leqleft.
(*False*) clear IHtw1. 
     assert (HH1 := IHtw2 _ _ _ _ H8 HG); clear IHtw2.
     intros y.
     assert (Z:= CompCode_Lemma _ _ _ _ _ y _ H4); clear H4.
     destruct Z. 
     destruct H1.
         clear H8. destruct H1. rewrite H1; clear H1 H0.
         autorewrite with fpcREWRITE.
         unfold FPCCTXT2CTXT, FLUB; simpl.
         remember (modVars C2 y) as b; destruct b; subst; simpl in *; apply eq_sym in Heqb.
           apply lubright.   
           apply lub_leq. apply HH1.                 
           apply (FPC_Mod _ _ _ y (FPCT_FPC _ _ _ tw2) Heqb). clear tw2.
         rewrite Heqb in *. rewrite Bool.orb_false_r in H2.
         apply lub_leq.
            apply lubright. apply HH1.
            apply lubleft. apply (FPC_Mod _ _ _ y (FPCT_FPC _ _ _ tw1) H2).
       destruct H1; simpl in *. apply Bool.orb_false_iff in H2. destruct H2.
         rewrite H1. 
         destruct (TSem1 _ _ _ _ _ H8 y) as [_ Z].
            rewrite modVars_theta. assumption.
          rewrite <- Z. clear H8 tw1 Z. 
          assert (Z:= FPCT_FPC _ _ _ tw2).
          assert (ZZ:= FPC_Nomod _ _ _ y Z H3).
          rewrite FPCCTXT2CTXT_upd_pcVar.
          rewrite M_upd_Diff in ZZ; try discriminate.
          eapply leq_trans. apply HG. unfold FLUB, FPCCTXT2CTXT. rewrite <- ZZ.
          simpl. apply lub_leqright. 

(*While*)
clear PreBelowN.
  rewrite FPCCTXT2CTXT_upd_pcVar.
inversion Hyp; subst; simpl; clear Hyp.  destruct r as [r H].
destruct (CompCode_Lemma0 _ _ _ _ _ _ H5) as [MC1 MC2]. clear H5.
intros x.
destruct (MC2 x); clear MC2.
Focus 2. (*notmod x C*)
  destruct H1. subst. rewrite H1 in *. clear H1.
  assert (modVars (TErase (TWhile E (theta C (pre N) (post N) (HypC N)))) x = false).
    simpl. rewrite modVars_theta. assumption. 
  destruct (TSem1 _ _ _ _ _ H3 x H1) as [RS1 RS2].
    rewrite <- RS2.
    eapply leq_trans. apply HG.
    rewrite (PreAboveN N). rewrite PrePost. rewrite FPCCTXT2CTXT_upd_pcVar.
      unfold FLUB; simpl. apply lub_leqright. 
    apply Le.le_refl.
(*mod x C*)
  destruct H1. rewrite H1. clear H1.
  remember (s,G0) as ss. remember (r,H) as rr.
  assert (HGG: SQSUB G0 (FPCCTXT2CTXT (pre N))).
    eapply leq_trans. apply HG. rewrite (PreAboveN N).  rewrite PrePost. 
           intros z. unfold FLUB; simpl. rewrite FPCCTXT2CTXT_upd_pcVar.
           apply lub_leqright. apply Le.le_refl.
  remember (HypC N) as HypCN. assert (K0 := H0 N). rewrite <- HeqHypCN in *.
    clear HeqHypCN. clear H0 PreZ HypC HG. 
  remember ((TWhile E (theta C (pre N) (post N) (HypCN)))) as DD.
  rewrite MC1 in *. clear MC1 r.
  generalize dependent H. generalize dependent t.
  generalize dependent G0. generalize dependent s.
  generalize dependent C. generalize dependent E. (* generalize dependent G.  *)
  induction H3; intros; try inversion HeqDD; subst; clear HeqDD.
    (*WhileTrue*)
    destruct r as [u K]. simpl in *. clear H IHTntSem1.
    assert (X1 : SQSUB G0 (FPCCTXT2CTXT (pre N))).
      clear IHTntSem2 H3_0. apply HGG.
    assert (IH := IHTntSem2 _ PrePost _ HypCN (eq_refl _) H2 K0).
       clear IHTntSem2.
    assert (XX: SQSUB K (FPCCTXT2CTXT (pre N))). clear IH.
      apply @leq_trans with (y:= FPCCTXT2CTXT (post N)).
        eapply K0. apply H3_. apply X1.
      rewrite (PreAboveN N). rewrite PrePost. rewrite FPCCTXT2CTXT_upd_pcVar.
        unfold FLUB; simpl. intros z. apply lub_leqleft.
      apply Le.le_refl.
    apply (IH _ _ (eq_refl _) XX _ _ (eq_refl _)).
  (*WhileFalse*) 
     inversion Heqrr; subst; clear Heqrr H. simpl in *.
     rewrite (PreAboveN N). rewrite (PrePost N).   
       rewrite FPCCTXT2CTXT_upd_pcVar. unfold FPCCTXT2CTXT, FLUB; simpl.
       apply lub_leq.
         eapply leq_trans. apply HGG. rewrite (PreAboveN N).  rewrite PrePost. rewrite FPCCTXT2CTXT_upd_pcVar. apply leq_refl. apply Le.le_refl.
         assert ((GEval
        (fun x0 : PVar =>
         fpcupd (fun t : Var => llub (post N t) (G t)) pcVar
           (llub (llub (post N pcVar) (G pcVar))
              (GEval
                 (fun x1 : PVar => llub (post N (progVar x1)) (G (progVar x1)))
                 E)) (progVar x0)) E)
   = FPCGEval (FLUB (post N) G) (progExpr E)).
                 rewrite <- (GEval_pc2 _ _  (llub (llub (post N pcVar) (G pcVar))
        (GEval (fun x1 : PVar => llub (post N (progVar x1)) (G (progVar x1)))
           E))). reflexivity.

           unfold llub in *. rewrite H. clear H.
           assert (XN := FPCT_FPC _ _ _ HypCN).
           rename H2 into MV.
           assert (Y:= FPC_Mod2 _ _ _ x XN MV).
           assert (X1 := FPC_Lemma1 _ _ _ XN).
             rewrite <- X1 in *; clear X1 XN MV.
             eapply @leq_trans with (y:=post N (progVar x)).
               eapply @leq_trans with (y:=pre N pcVar).
                rewrite (PreAboveN N). rewrite PrePost. rewrite fpcupd_Same.
                  apply leq_monoleft. unfold FLUB; simpl. apply lub_leqright.
                 apply Le.le_refl.
               apply Y.
              apply lub_leqleft.
            apply Le.le_refl.
Qed.

Lemma TntSem_FPC:  forall C Gamma Delta (tw:FPCT Gamma C Delta) s G t D
(Hyp: TntSem (s,G) (theta C Gamma Delta tw) (t,D))
(HG: SQSUB G (FPCCTXT2CTXT Gamma)),  SQSUB D (FPCCTXT2CTXT Delta).
Proof. apply Lemma2. Qed. 

End SynthesisTheta.

Definition sec {A} (G H:CTXT) (s s' t t': PVar -> A):=
  forall x, eq_below (H x) G s s' -> t x = t' x.

Definition static_sec (G H:CTXT) C := forall s s' t t' D D'
(Hyp: TntSem (s,G) C (t,D)) (Hyp': TntSem (s',G) C (t',D')),
sec G H s s' t t'.

(*The two halves of Lemma 3:*)
Lemma sec_NI: forall G H (s s' t t': Store), 
  sec G H s s' t t' -> NI G H s s' t t'.
Proof.
intros. unfold sec in H0. intros kappa EB x Hx.
eapply H0.
intros z Hz. apply EB. apply (leq_trans Hz Hx).
Qed.
 
Lemma NI_sec: forall G H (s s' t t': Store), 
  NI G H s s' t t' -> sec G H s s' t t'.
Proof.
intros. unfold NI in H0. intros x EB.
eapply H0. apply EB. apply leq_refl.
Qed. 

Lemma Theorem2: forall C Gamma Delta (tw: FPCT Gamma C Delta),
    static_sec (FPCCTXT2CTXT Gamma) (FPCCTXT2CTXT Delta) (theta _ _ _ tw).
Proof.
intros.
assert (FP:= FPCT_FPC _ _ _ tw).
assert (Z:= FPC_kappaNI _ _ _ FP).
unfold static_sec. intros.
eapply NI_sec.
unfold kappaNI in Z.
apply theta_TntSem_Sem in Hyp.
apply theta_TntSem_Sem in Hyp'.
apply Z; trivial.
Qed.


Lemma FPC_Sound3_RSem: forall Gamma Delta C (tw: FPCT Gamma C Delta) 
s G t D (Hyp: TntSem (s,G) (theta C Gamma Delta tw) (t,D))
s' G' t' D' (Hyp': TntSem (s',G') (theta C Gamma Delta tw) (t',D'))
(SUB: SQSUB G (FPCCTXT2CTXT Gamma))
(SUB': SQSUB G' (FPCCTXT2CTXT Gamma)) 
kappa
(EB_Ctxt: eq_below kappa (FPCCTXT2CTXT Gamma) G G')
(EB_State: eq_below kappa (FPCCTXT2CTXT Gamma) s s'),
(eq_below kappa (FPCCTXT2CTXT Delta) t t' /\ 
 eq_below kappa (FPCCTXT2CTXT Delta) D D'). 
Proof.
intros Gamma Delta C tw. 
induction tw; intros; subst; simpl in *; rename G into Gamma; rename G0 into G. 
(*Skip*)
inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. 
  inversion H; subst; clear H. 
  inversion H0; subst; clear H0. 
  split; assumption. 
(*Assign*)
inversion Hyp; subst; clear Hyp. autorewrite with fpcREWRITE in *.
  inversion Hyp'; subst; clear Hyp'.
  split; intros z Hz. 
  (*store*)
    remember (pvar_eq x z) as b; destruct b; subst; simpl in *.
      rewrite M_upd_Same in Hz.  
      repeat rewrite M_upd_Same in *.  
      eapply GevalEqonEval.  apply EB_State.   apply leq_refl.
      eapply (leq_snart Hz). apply lub_leqright.
    rewrite M_upd_Diff in Hz; trivial.
      repeat rewrite M_upd_Diff; trivial. apply EB_State.  apply Hz. 
  (*dyn context*)
    remember (pvar_eq x z) as b; destruct b; subst; simpl in *. 
      rewrite M_upd_Same in Hz.  
      repeat rewrite M_upd_Same. 
      assert (EQ: GEval G E = GEval G' E). 
        eapply eq_below_GevalE. apply EB_Ctxt. eapply (leq_snart Hz). apply lub_leqright. 
      rewrite EQ; trivial.
    rewrite M_upd_Diff in Hz; trivial. 
      repeat rewrite M_upd_Diff; trivial.
      apply EB_Ctxt.  apply Hz. 
(*Comp*)
inversion Hyp; subst; clear Hyp. 
  destruct r as [r H]. 
  inversion Hyp'; subst; clear Hyp'. 
  destruct r0 as [r' H']. 
  destruct (IHtw1 _ _ _ _ H2 _ _ _ _ H5 SUB SUB' kappa EB_Ctxt EB_State) as [K1 K2]. 
  eapply (IHtw2 _ _ _ _ H4 _ _ _ _ H7).
     apply (TntSem_FPC _ _ _ tw1 _ _ _ _ H2 SUB). 
     apply (TntSem_FPC _ _ _ tw1 _ _ _ _ H5 SUB'). 
     apply K2. 
     apply K1. 
(*Ite*)
  assert (EB:= eq_below_evalE E (FPCCTXT2CTXT Gamma) s s' kappa EB_State). 
  autorewrite with fpcREWRITE in *.
  inversion Hyp; subst; clear Hyp. 
  inversion Hyp'; subst; clear Hyp'. 
  destruct r as [r H]. 
  destruct r0 as [r' H'].  
  destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [RT MC]; subst; clear H4.
  destruct (CompCode_Lemma0 _ _ _ _ _ _ H6) as [RT' MC']; subst; clear H6.
  split.  (*claim about store*)
    intros z Hz. 
    autorewrite with fpcREWRITE in *.
    destruct (MC z); simpl in *. 
    (*In  z (modVarsL C1 ++ modVarsL C2)*) 
        destruct H0 as [HD HZ].  clear MC' MC HD. 
        assert (EV: eval s E = eval s' E).  clear IHtw1 IHtw2 H2 H3. 
           apply EB.
           eapply (leq_snart Hz); clear Hz. 
           unfold FLUB; simpl.
           apply Bool.orb_true_iff in HZ.
            destruct HZ. 
                assert (Q:= FPC_Mod _ _ _ _ (FPCT_FPC _ _ _ tw1) H0). 
                rewrite fpcupd_Same in Q.  
                destruct (lub_leq_elim _ _ _ Q). apply lubleft. apply H2.
             assert (Q:= FPC_Mod _ _ _ _ (FPCT_FPC _ _ _ tw2) H0). 
                rewrite fpcupd_Same in Q.  
                destruct (lub_leq_elim _ _ _ Q). apply lubright. apply H2.
        inversion H2; subst; clear H2. 
        (*True*) clear IHtw2. 
           inversion H3; subst; clear H3. 
           (*True*)
             destruct (IHtw1 _ _ _ _ H8 _ _ _ _ H9 SUB SUB' _ EB_Ctxt EB_State) as [K1 _].  clear IHtw1 H8 H9. 
             apply K1. apply (leq_snart Hz).
                         unfold FLUB.  simpl.  apply lub_leqleft.
           (*False*) clear IHtw1. 
             simpl in *.  rewrite EV in H7. rewrite H7 in H6.   inversion H6. 
        (*False*) clear IHtw1. 
           inversion H3; subst; clear H3. 
           (*True*) clear IHtw2. 
             simpl in *.  rewrite EV in H7.  rewrite H7 in H6.  inversion H6. 
           (*False*)
             destruct (IHtw2 _ _ _ _ H8 _ _ _ _ H9 SUB SUB' _ EB_Ctxt EB_State) as [K1 _].  clear IHtw2 H8 H9. 
             apply K1.  apply (leq_snart Hz).
                         unfold FLUB.  simpl.  apply lub_leqright.
    (* ~ In z (modVarsL C1 ++ modVarsL C2)*)
        clear IHtw1 IHtw2 MC MC'. 
        assert (RS1 := TSem1 _ _ _ _ _ H2 z).  clear H2. simpl in *. repeat rewrite modVars_theta in RS1.
        
        assert (RS2 := TSem1 _ _ _ _ _ H3 z).  clear H3. simpl in *. repeat rewrite modVars_theta in RS2.
        destruct H0 as [ _ K]. 
        destruct (RS1 K) as [K1 K2].  rewrite <- K1.  clear K1. 
        destruct (RS2 K) as [K1' K2'].  rewrite <- K1'.  clear K1'. 
        apply EB_State.
        eapply (leq_snart Hz).
          unfold FPCCTXT2CTXT, FLUB; simpl. apply Bool.orb_false_iff in K. destruct K as [K _].
          rewrite (FPC_Nomod _ _ _ z (FPCT_FPC _ _ _ tw1) K).
          apply lub_leqleft.  
  (*claim about dynamic CTXT - very similar proof, but we need to do them
        separate since both clauses quantify over all program variables. *)
    intros z Hz.
    autorewrite with fpcREWRITE in *.
    destruct (MC z); simpl in *. 
    (*In  z (modVarsL C1 ++ modVarsL C2)*) 
        destruct H0 as [HD HZ].
        destruct (MC' z); simpl in *; destruct H0 as [HD' HZ']; rewrite HZ in HZ'. Focus 2. inversion HZ'. clear HZ'.
        assert (EV: eval s E = eval s' E).  clear IHtw1 IHtw2 H2 H3. 
           apply EB.
           eapply (leq_snart Hz); clear Hz. 
           unfold FLUB; simpl.
           apply Bool.orb_true_iff in HZ.
            destruct HZ. 
                assert (Q:= FPC_Mod _ _ _ _ (FPCT_FPC _ _ _ tw1) H0). 
                rewrite fpcupd_Same in Q.  
                destruct (lub_leq_elim _ _ _ Q). apply lubleft. apply H2.
             assert (Q:= FPC_Mod _ _ _ _ (FPCT_FPC _ _ _ tw2) H0). 
                rewrite fpcupd_Same in Q.  
                destruct (lub_leq_elim _ _ _ Q). apply lubright. apply H2.
        rewrite HD in *; clear HD. rewrite HD' in *; clear HD'.
        assert (H z = H' z).
          inversion H2; subst; clear H2. 
          (*True*) clear IHtw2. 
             inversion H3; subst; clear H3. 
             (*True*)
               destruct (IHtw1 _ _ _ _ H8 _ _ _ _ H9 SUB SUB' _ EB_Ctxt EB_State) as [K1 K2].  clear IHtw1 H8 H9. 
               apply K2. apply (leq_snart Hz).
                         unfold FLUB.  simpl.  apply lub_leqleft.
             (*False*) clear IHtw1. 
               simpl in *.  rewrite EV in H7. rewrite H7 in H6.   inversion H6. 
          (*False*) clear IHtw1. 
             inversion H3; subst; clear H3. 
             (*True*) clear IHtw2. 
               simpl in *.  rewrite EV in H7.  rewrite H7 in H6.  inversion H6. 
             (*False*)
               destruct (IHtw2 _ _ _ _ H8 _ _ _ _ H9 SUB SUB' _ EB_Ctxt EB_State) as [K1 K2].  clear IHtw2 H8 H9. 
               apply K2.  apply (leq_snart Hz).
                         unfold FLUB.  simpl.  apply lub_leqright.
        rewrite H0; trivial.
    (* ~ In z (modVarsL C1 ++ modVarsL C2)*)
        destruct H0 as [HD HZ].
        destruct (MC' z); simpl in *; destruct H0 as [HD' HZ']; rewrite HZ in HZ'. inversion HZ'. clear HZ'.
        rewrite HD in *; clear HD.
        rewrite HD' in *; clear HD'.
        clear IHtw1 IHtw2 MC MC'.           
        assert (RS1 := TSem1 _ _ _ _ _ H2 z).  clear H2. simpl in *. repeat rewrite modVars_theta in RS1.        
        assert (RS2 := TSem1 _ _ _ _ _ H3 z).  clear H3. simpl in *. repeat rewrite modVars_theta in RS2.
        destruct (RS1 HZ) as [_ K2].  rewrite <- K2.  clear K2. 
        destruct (RS2 HZ) as [_ K2'].  rewrite <- K2'.  clear K2'. 
        apply EB_Ctxt.
        eapply (leq_snart Hz).
          unfold FPCCTXT2CTXT, FLUB; simpl. apply Bool.orb_false_iff in HZ. destruct HZ as [K _].
          rewrite (FPC_Nomod _ _ _ z (FPCT_FPC _ _ _ tw1) K).
          apply lub_leqleft.  
(*While*)
autorewrite with fpcREWRITE in *.
assert (HG : FSQSUB Gamma (pre N)).
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. apply lubleft. unfold FLUB; simpl. apply lub_leqright.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqright.
  apply Le.le_refl.
assert (SUB_G_N : SQSUB G (FPCCTXT2CTXT (pre N))).
  eapply leq_trans. apply SUB. intros z. apply HG.
assert (SUB_G'_N : SQSUB G' (FPCCTXT2CTXT (pre N))).
  eapply leq_trans. apply SUB'. intros z. apply HG.
assert (PostNPreN : FSQSUB (post N) (pre N)).
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. apply lubleft. unfold FLUB; simpl. apply lub_leqleft. 
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl.  apply lub_leqleft.
  apply Le.le_refl.
remember (llub (Gamma pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) as p.  
assert (p_le: ssqsub p (pre N pcVar)).
  clear Hyp Hyp'. rewrite (PreAboveN N). rewrite PrePost. rewrite Heqp.
             unfold FLUB; simpl. rewrite fpcupd_Same. 
             apply lub_leq. eapply @leq_trans with (y:=llub (post N pcVar) (Gamma pcVar)). apply lub_leqright. apply lub_leqleft.
             eapply @leq_trans with (y:=(GEval (fun x : PVar => llub (post N (progVar x)) (Gamma (progVar x))) E)).
                  eapply GEval_mono. intros y. rewrite (PreAboveN N). rewrite PrePost. rewrite fpcupd_Diff; try discriminate. apply leq_refl. apply Le.le_refl. 
                eapply lub_leqright. 
            apply Le.le_refl.
remember (Gamma pcVar) as GpcVar.
remember  (GEval (fun x : PVar => (pre N) (progVar x)) E) as GEvalE.
assert (P1 : ssqsub GEvalE p). subst. apply lub_leqright.
assert (HN := H0 N). clear H0.
assert (EQ_C : eq_below kappa (FPCCTXT2CTXT (pre N)) G G').
  eapply eq_below_mono. apply EB_Ctxt. apply leq_refl. 
      intros z. apply HG. 
assert (EQ_S : eq_below kappa (FPCCTXT2CTXT (pre N)) s s').
  eapply eq_below_mono. apply EB_State. apply leq_refl. 
      intros z. apply HG.
clear HeqGpcVar PreZ PrePost PreBelowN PreAboveN SUB SUB' HG EB_Ctxt EB_State Gamma.
inversion Hyp; clear Hyp. inversion Hyp'; clear Hyp'.
  destruct r as [r F]. 
  destruct r0 as [r' F'].
  split.  (*claim about store*)
    intros z Hz. 
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [MC1 MC2].  rewrite <- MC1.  clear MC1.  
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H10) as [MC1' MC2'].  rewrite <- MC1'.  clear MC1'. 
    clear H4 H10 H0 C2 H6 C3.       
    destruct (MC2 z); destruct H0 as [HD HZ]. 
    (*In  z (modVarsL C)*) 
        clear MC2' MC2 HD.
        assert (subKappa: ssqsub (GEval (FPCCTXT2CTXT (pre N)) E) kappa).
             unfold FPCCTXT2CTXT. rewrite <-  HeqGEvalE.
             eapply leq_trans. apply P1.
             eapply leq_trans. apply p_le. 
             eapply leq_trans. apply (FPC_Mod _ _ _ _ (FPCT_FPC _ _ _ (HypC N)) HZ).
             eapply leq_trans. apply PostNPreN. apply Hz. 
        assert (EV: eval s E = eval s' E).
           eapply eq_below_evalE. apply EQ_S. apply subKappa. 
        clear H C1. clear H5 C0. clear H3 H9 t D t' D' t0 t1.
        remember (TWhile E (theta C (pre N) (post N) (HypC N))) as R.
        rewrite <- H1 in H2. remember (r, F) as t. clear H7 s1.
        generalize dependent F'. generalize dependent r'.
        generalize dependent G'. generalize dependent s'.
        generalize dependent F. generalize dependent r.
        generalize dependent G. generalize dependent s.
        induction H2; intros; try inversion HeqR.
        (*WhileTrue*) clear IHTntSem1.
           inversion H8; subst; clear H8.
           (*True*) clear HeqR EV. simpl in *.
             destruct r as [t HH]. destruct r1 as [t' HH']. 
             assert (Z: SQSUB HH (FPCCTXT2CTXT (pre N))).
                clear IHTntSem2. 
                 eapply @leq_trans with (y:=(FPCCTXT2CTXT (post N))).
                  eapply TntSem_FPC. apply H2_. apply SUB_G_N.
                  intros y. apply PostNPreN. 
             assert (Z': SQSUB HH' (FPCCTXT2CTXT (pre N))).
                clear IHTntSem2. 
                 eapply @leq_trans with (y:=(FPCCTXT2CTXT (post N))).
                  eapply TntSem_FPC. apply H7. apply SUB_G'_N.
                 intros y. apply PostNPreN.
             destruct (HN _ _ _ _ H2_ _ _ _ _ H7 SUB_G_N SUB_G'_N kappa EQ_C EQ_S) as [K1 K2]. 
             clear H2_ H7. 
             assert (EQ_CC : eq_below kappa (FPCCTXT2CTXT (pre N)) HH HH').
                eapply eq_below_mono. apply K2. apply leq_refl. intros y. apply PostNPreN.
             assert (EQ_SS : eq_below kappa (FPCCTXT2CTXT (pre N)) t t').
                eapply eq_below_mono. apply K1. apply leq_refl. intros y. apply PostNPreN.
             clear K1 K2.
             assert (EV_t: eval t E = eval t' E).
                eapply eq_below_evalE. apply EQ_SS. apply subKappa.
             eapply (IHTntSem2 (eq_refl _) _ _ Z (eq_refl _ ) _ _ (eq_refl _) _ EQ_SS EV_t _ Z' EQ_CC _ _ H10).
            (*False*) clear IHTntSem2. 
             simpl in *. rewrite EV in H.  rewrite H in H6.  inversion H6. 
        (*False*) clear HN. 
           inversion H8; subst; clear H8. 
           (*True*) 
             simpl in *. rewrite EV in H.  rewrite H in H6.  inversion H6.
           (*False*)
             inversion H10; subst. inversion Heqt; subst. 
             apply EQ_S. apply Hz.
    (* ~ In z (modVarsL C1 ++ modVarsL C2)*)
        clear HN MC2 MC2'.
        assert (RS1 := TSem1 _ _ _ _ _ H2 z).  clear H2. simpl in RS1. rewrite modVars_theta in RS1.
        destruct (RS1 HZ) as [K _]. rewrite <- K. clear RS1 K.

        assert (RS1 := TSem1 _ _ _ _ _ H8 z).  clear H8.  simpl in RS1. rewrite modVars_theta in RS1.
        destruct (RS1 HZ) as [K _]. rewrite <- K. clear RS1 K.

        apply EQ_S. apply Hz.
(*claim about dynamic context*)
    intros z Hz. 
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [MC1 MC2]. rewrite <- MC1 in *.  clear MC1.  
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H10) as [MC1' MC2']. rewrite <- MC1' in *.  clear MC1'.  
    clear H4 H10 H0 C2 H6 C3.       
    destruct (MC2 z); destruct H0 as [HD HZ]. 
    (*In  z (modVarsL C)*) 
        destruct (MC2' z); destruct H0 as [HD' HZ']; rewrite HZ in HZ'. Focus 2. inversion HZ'. clear HZ'. 
        clear MC2' MC2. rewrite HD in *. rewrite HD' in *. clear HD HD'.
        assert (HFF: (F z) = (F' z)).
          assert (subKappa: ssqsub (GEval (FPCCTXT2CTXT (pre N)) E) kappa).
             unfold FPCCTXT2CTXT. rewrite <-  HeqGEvalE.
             eapply leq_trans. apply P1.
             eapply leq_trans. apply p_le. 
             eapply leq_trans. apply (FPC_Mod _ _ _ _ (FPCT_FPC _ _ _ (HypC N)) HZ).
             eapply leq_trans. apply PostNPreN. apply Hz. 
          assert (EV: eval s E = eval s' E).
             eapply eq_below_evalE. apply EQ_S. apply subKappa. 
          clear H C1. clear H5 C0. clear H3 H9 t t' t0 t1.
          remember (TWhile E (theta C (pre N) (post N) (HypC N))) as R.
          rewrite <- H1 in H2. remember (r, F) as t. clear H7 s1.
          generalize dependent F'. generalize dependent r'.
          generalize dependent G'. generalize dependent s'.
          generalize dependent F. generalize dependent r.
          generalize dependent G. generalize dependent s.
          induction H2; intros; try inversion HeqR.
          (*WhileTrue*) clear IHTntSem1.
             inversion H8; subst; clear H8.
             (*True*) clear HeqR EV. simpl in *.
               destruct r as [t HH]. destruct r1 as [t' HH']. 
               assert (Z: SQSUB HH (FPCCTXT2CTXT (pre N))).
                 clear IHTntSem2. 
                 eapply @leq_trans with (y:=(FPCCTXT2CTXT (post N))).
                  eapply TntSem_FPC. apply H2_. apply SUB_G_N.
                  intros y. apply PostNPreN. 
               assert (Z': SQSUB HH' (FPCCTXT2CTXT (pre N))).
                 clear IHTntSem2. 
                 eapply @leq_trans with (y:=(FPCCTXT2CTXT (post N))).
                  eapply TntSem_FPC. apply H7. apply SUB_G'_N.
                 intros y. apply PostNPreN.
               destruct (HN _ _ _ _ H2_ _ _ _ _ H7 SUB_G_N SUB_G'_N kappa EQ_C EQ_S) as [K1 K2]. 
               clear H2_ H7. 
               assert (EQ_CC : eq_below kappa (FPCCTXT2CTXT (pre N)) HH HH').
                  eapply eq_below_mono. apply K2. apply leq_refl. intros y. apply PostNPreN.
               assert (EQ_SS : eq_below kappa (FPCCTXT2CTXT (pre N)) t t').
                  eapply eq_below_mono. apply K1. apply leq_refl. intros y. apply PostNPreN.
               clear K1 K2.
               assert (EV_t: eval t E = eval t' E).
                  eapply eq_below_evalE. apply EQ_SS. apply subKappa.
               eapply (IHTntSem2 (eq_refl _) _ _ Z (eq_refl _ ) _ _ (eq_refl _) _ EQ_SS EV_t _ Z' EQ_CC _ _ H10).
            (*False*) clear IHTntSem2. 
             simpl in *. rewrite EV in H.  rewrite H in H6.  inversion H6. 
        (*False*) clear HN. 
           inversion H8; subst; clear H8. 
           (*True*) 
             simpl in *. rewrite EV in H.  rewrite H in H6.  inversion H6.
           (*False*)
             inversion H10; subst; clear H10. inversion Heqt; subst; clear Heqt. 
             apply EQ_C. apply Hz.
       rewrite HFF; trivial.
    (* ~ In z (modVarsL C1 ++ modVarsL C2)*)
        clear HN.
        destruct (MC2' z); destruct H0 as [HD' HZ']; rewrite HZ in HZ'. inversion HZ'. clear HZ'. 
        clear MC2' MC2. rewrite HD in *. rewrite HD' in *. clear HD HD'.
        assert (RS1 := TSem1 _ _ _ _ _ H2 z).  clear H2. simpl in RS1. rewrite modVars_theta in RS1.
        destruct (RS1 HZ) as [_ K]. rewrite <- K. clear RS1 K.

        assert (RS1 := TSem1 _ _ _ _ _ H8 z).  clear H8.  simpl in RS1. rewrite modVars_theta in RS1.
        destruct (RS1 HZ) as [_ K]. rewrite <- K. clear RS1 K.

        apply EQ_C. apply Hz.
Qed.

Lemma Theorem2_explicitProof: forall C Gamma Delta (tw: FPCT Gamma C Delta),
    static_sec (FPCCTXT2CTXT Gamma) (FPCCTXT2CTXT Delta) (theta _ _ _ tw).
Proof.
unfold static_sec, sec. intros.
assert (SUB: SQSUB (FPCCTXT2CTXT Gamma) (FPCCTXT2CTXT Gamma)). apply leq_refl.
destruct (FPC_Sound3_RSem _ _ _ tw _ _ _ _ Hyp _ _ _ _ Hyp' SUB SUB _ (eq_below_refl _ _ _) H).
apply H0. apply leq_refl.
Qed.

Definition dynamic_sec (G:CTXT) C := forall s s' t t' D D'
(Hyp: TntSem (s,G) C (t,D)) (Hyp': TntSem (s',G) C (t',D')),
sec G D s s' t t'.

Definition sim {A:Type} (k:L) (G G': CTXT) (s s' : PVar -> A) :=
forall x, ssqsub (G x) k -> (s x = s' x /\ G x = G' x). 

Lemma sim_eq_below: forall {A} G G' s s' k, @sim A k G G' s s' <->
eq_below k G s s' /\ eq_below k G G G'. 
Proof. 
intros.  unfold sim, eq_below. 
split; intros. 
  split; intros.  apply H.  assumption.  apply H.  assumption. 
destruct H. 
  split.  apply H.  assumption. 
  apply H1.  assumption. 
Qed. 

Lemma sim_GEval: forall {A} E k G G' s s' 
(Hyp: @sim A k G G' s s') (SS:ssqsub (GEval G E) k),
GEval G E = GEval G' E. 
Proof. 
intros A E. 
induction E; intros; simpl in *. 
  trivial. 

  apply Hyp.  assumption. 

  destruct (lub_leq_elim _ _ _ SS) as [K1 K2]. 
  rewrite (IHE1 _ _ _ _ _ Hyp K1). 
  rewrite (IHE2 _ _ _ _ _ Hyp K2). 
  trivial. 
Qed. 

Lemma sim_eval: forall E k G G' s s' 
(Hyp: sim k G G' s s') (SS:ssqsub (GEval G E) k),
eval s E = eval s' E. 
Proof. 
intros E. 
induction E; intros; simpl in *. 
  trivial. 

  apply Hyp.  assumption. 

  destruct (lub_leq_elim _ _ _ SS) as [K1 K2]. 
  rewrite (IHE1 _ _ _ _ _ Hyp K1). 
  rewrite (IHE2 _ _ _ _ _ Hyp K2). 
  trivial. 
Qed. 

Lemma sim_mono: (*mono only on state component: G/G/ unchanged*)
  forall (A : Type) (k : L) (G G': CTXT) (s s' : PVar -> A) (k' : L),
  sim k G G' s s' -> ssqsub k' k -> sim k' G G' s s'.
Proof.
unfold sim; intros.
apply H. eapply leq_trans. apply H1. apply H0. 
Qed. 

(*Lemma is called FPCRifle_SoundAux_explicitProof in HLRifle.v*)
Lemma Lemma4: forall Gamma Delta C (tw: FPCT Gamma C Delta) 
s G t D (Hyp: TntSem (s,G) (theta C Gamma Delta tw) (t,D)) 
s' G' t' D' (Hyp': TntSem (s',G') (theta C Gamma Delta tw) (t',D'))
(HG: SQSUB G (FPCCTXT2CTXT Gamma)) 
(HG': SQSUB G' (FPCCTXT2CTXT Gamma))
x
(SIM: sim (D x) G G' s s'),
t x = t' x /\ D x = D' x.
Proof.
intros Gamma Delta C tw. 
induction tw; intros; subst; simpl in *; rename G into Gamma; rename G0 into G. 
(*Skip*)
inversion Hyp; subst; clear Hyp. 
  inversion H; subst; clear H.  
  inversion Hyp'; subst; clear Hyp'. 
  inversion H; subst; clear H. 
  apply (SIM x). apply leq_refl. 
(*Assign*)
inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'. simpl in *. rewrite id_bbot in SIM. 
  remember (pvar_eq x x0) as b; destruct b; subst; simpl.
    repeat rewrite M_upd_Same. rewrite M_upd_Same in SIM.  
    split.  eapply sim_eval. 
             apply SIM. 
             apply llub_sqleft. 
    assert (GEval G E = GEval G' E). 
      eapply sim_GEval.  apply SIM.  apply llub_sqleft. 
      rewrite H.  trivial.
  repeat rewrite M_upd_Diff; trivial. rewrite M_upd_Diff in SIM; trivial.  
    apply SIM. apply ssqsub_refl.  
(*Comp*)
inversion Hyp; subst; clear Hyp. 
  destruct r as [r F]. 
  inversion Hyp'; subst; clear Hyp'. 
  destruct r0 as [r' F']. 
  eapply IHtw2; clear IHtw2.  
    apply H4. 
    apply H6. 
    clear IHtw1. 
      eapply TntSem_FPC.  apply H2. apply HG.
      eapply TntSem_FPC.  apply H3. apply HG'.
  clear H4 H6. 
  intros z Hz. 
    eapply IHtw1; clear IHtw1. 
      apply H2. 
      apply H3. 
      assumption. 
      assumption.
      eapply sim_mono. apply SIM. apply Hz. 
 (*If*)
autorewrite with fpcREWRITE in *.
inversion Hyp; subst; clear Hyp. destruct r as [r F].
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [r' F'].
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [MC1 MC2]. subst.
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H6) as [MC1' MC2']. subst.
    clear H4 H6.
    destruct (MC2 x) as [[HD HZ] | [HD HZ]]. 
    (*In  z (modVarsL C1 ++ modVarsL C2)*)
       destruct (MC2' x) as [[HD' _] | [HD' HZ']]. Focus 2. rewrite HZ in HZ'. inversion HZ'. 
        rewrite HD in *.  rewrite HD' in *. clear HD HD' MC2 MC2'. 
        assert (EV: eval s E = eval s' E). 
            clear IHtw1 IHtw2 tw1 tw2 H2 H3. 
            eapply sim_eval.  apply SIM. apply lubright. apply lubright.
                eapply GEval_mono. apply HG.
          inversion H2; subst; clear H2; simpl in *. 
          (*True*) clear IHtw2. 
            inversion H3; subst; clear H3; simpl in *. 
            (*True*) 
               destruct (IHtw1 _ _ _ _ H7 _ _ _ _ H8 HG HG' x) as [IH_S IH_C]; clear IHtw1 H7 H8 tw1 tw2.
                 eapply sim_mono. apply SIM. apply lub_leqleft.
               rewrite <- IH_S; clear IH_S. 
               rewrite <- IH_C; clear IH_C.
               split; trivial. 
            (*False*)
              clear IHtw1 H7 H8 tw1 tw2. 
              rewrite EV in H6.  rewrite H6 in H5.  inversion H5. 
          (*False*) clear IHtw1. 
            inversion H3; subst; clear H3; simpl in *. 
            (*True*) 
              clear IHtw2 H7 H8 tw1 tw2. 
              rewrite EV in H6.  rewrite H6 in H5.  inversion H5. 
            (*False*)
               destruct (IHtw2 _ _ _ _ H7 _ _ _ _ H8 HG HG' x) as [IH_S IH_C]; clear IHtw2 H7 H8 tw1 tw2.
                 eapply sim_mono. apply SIM. apply lub_leqleft.
               rewrite <- IH_S; clear IH_S. 
               rewrite <- IH_C; clear IH_C.
               split; trivial. 
    (* ~ In  z (modVarsL C1 ++ modVarsL C2)*)
      destruct (MC2' x) as [[HD' HZ'] | [HD' _]]. rewrite HZ in HZ'. inversion HZ'. 
      rewrite HD in *.  rewrite HD' in *. clear HD HD' MC2 MC2'. 
 (*      simpl in HZ. apply Bool.orb_false_iff in HZ. destruct HZ as [MV1 MV2].*)
      assert (noMod1: s x = t x /\ G x = F x). 
        eapply TSem1.  apply H2. simpl. repeat rewrite modVars_theta. apply HZ.
      destruct noMod1 as [Y1 Y2] .  rewrite <- Y1 in *.  rewrite <- Y2 in *.  clear Y1 Y2 H2. 
      assert (noMod2: s' x = t' x /\ G' x = F' x). 
        eapply TSem1.  apply H3. simpl. repeat rewrite modVars_theta. apply HZ.
      destruct noMod2 as [Y1 Y2] .  rewrite <- Y1 in *.  rewrite <- Y2 in *.  clear Y1 Y2 H3. 

      clear IHtw1 IHtw2 tw1 tw2. 
      apply SIM. apply ssqsub_refl. 
(*While*)
clear PreBelowN.
assert (HGamma : FSQSUB Gamma (pre N)).
  clear Hyp Hyp'.
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. apply lubleft.
        unfold FLUB; simpl.  apply lub_leqright.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqright.
  apply Le.le_refl.
assert (SUB : SQSUB G (FPCCTXT2CTXT (pre N))).
  eapply leq_trans. apply HG. intros z. apply HGamma.
assert (SUB' : SQSUB G' (FPCCTXT2CTXT (pre N))).
  eapply leq_trans. apply HG'. intros z. apply HGamma.
clear HG HG' PreZ.
assert (PostNPreN : FSQSUB (post N) (pre N)).
  clear Hyp Hyp' H0.
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. 
        unfold FLUB; simpl. apply lubleft. apply lub_leqleft.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
  apply Le.le_refl.
remember (llub (Gamma pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) as p.  
assert (HN := H0 N). clear H0.
remember (HypC N) as HypCN. clear HeqHypCN.
remember (pre N) as PreN.
remember (post N) as PostN.
assert (P1 : ssqsub (GEval (fun x : PVar => PreN (progVar x)) E) p).
  rewrite Heqp. apply lub_leqright.
clear Heqp. clear PrePost HGamma Gamma. clear HypC. clear HeqPreN HeqPostN PreAboveN pre post. 
remember (TWhile E (theta C PreN PostN HypCN)) as CC.

inversion Hyp; subst; clear Hyp. destruct r as [r F]. 
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [r' F']. 
  destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [MC1 MC2]. subst; clear H4.
  destruct (CompCode_Lemma0 _ _ _ _ _ _ H6) as [MC1' MC2']. subst; clear H6.
 
destruct (MC2 x) as [[HD HZ] | [HD HZ]]; clear MC2.
(*In x (modVarsL C)*)
  destruct (MC2' x) as [[HD' HZ'] | [HD' HZ']]; clear MC2'.
      Focus 2. rewrite HZ in HZ'. inversion HZ'.
  assert (AUX: t x = t' x /\ F x = F' x).
     rewrite HD in SIM.
     clear HZ' HD HD'. clear D D'.
     assert (HG_E_D: ssqsub (GEval G E) (llub (F x) p)). 
              eapply @leq_trans with (y:= (GEval (FPCCTXT2CTXT (PreN)) E)). 
                eapply GEval_mono. apply SUB.
             apply lubright. apply P1.
     assert (EV:= sim_eval E _ _ _ _ _ SIM HG_E_D). 
     remember (s, G) as ss. remember (t, F) as tt.
       generalize dependent G. generalize dependent s.
       generalize dependent F. generalize dependent t.
     remember (s', G') as ss'. remember (t', F') as tt'.
       generalize dependent G'. generalize dependent s'.
       generalize dependent F'. generalize dependent t'.
     remember (TWhile E (theta C PreN PostN HypCN)) as CC.
       generalize dependent tt'. generalize dependent ss'. 
     generalize dependent C.      
     induction H2; intros; try inversion HeqCC; subst; clear HeqCC; rename s0 into s. rename t0 into t. rename C0 into C.
     (*Case eval s E = true*)
         clear IHTntSem1. simpl in *.
         inversion H3; intros; subst; clear H3; simpl in *.
           Focus 2. (*While False*) clear HN IHTntSem2. rewrite EV in H. rewrite H in H4. inversion H4.
       (*While True*)
        destruct r as [r K]. destruct r0 as [r' K'].
        assert (Sim_KK_KK': sim (llub (F x) p) K K' r r').
          intros z Hz. apply (HN _ _ _ _ H2_ _ _ _ _ H5 SUB SUB').
            eapply sim_mono. apply SIM. apply Hz.
        assert (sub_K_PreN: SQSUB K (FPCCTXT2CTXT PreN)).
          clear IHTntSem2 HN.
          eapply leq_trans.
              apply (TntSem_FPC C PreN PostN HypCN _ _ _ _ H2_ SUB).
              intros z. apply PostNPreN.
        assert (sub_K'_PreN: SQSUB K' (FPCCTXT2CTXT PreN)).
          clear IHTntSem2 HN.
          eapply leq_trans.
              apply (TntSem_FPC C PreN PostN HypCN _ _ _ _ H5 SUB').
              intros z. apply PostNPreN.
        assert (K_E_Dx: ssqsub (GEval K E) (llub (F x) p)).
          eapply @leq_trans with (y:=(GEval (FPCCTXT2CTXT PreN) E)).
             apply GEval_mono. apply sub_K_PreN. 
          eapply lubright. apply P1.
        eapply (IHTntSem2 _ HypCN HN (refl_equal _) HZ); clear IHTntSem2.
          apply H7.
          reflexivity.
          apply sub_K'_PreN.
          reflexivity.
          reflexivity.
          eapply sim_eval. apply Sim_KK_KK'. apply K_E_Dx.
          apply Sim_KK_KK'.
          apply sub_K_PreN.
          reflexivity. 
          apply K_E_Dx.
     (*eval s E = false*)
        inversion Heqss; subst; clear Heqss. simpl in *.
        inversion H3; intros; subst; clear H3; simpl in *.
        (*While True*) clear HN. rewrite EV in H. rewrite H in H2. inversion H2.
        (*WhileFalse*) inversion H6; subst; clear H6.
          apply SIM.  apply lub_leqleft. 
  destruct AUX. rewrite HD. rewrite HD'. rewrite H. rewrite H0. split; trivial.
(*~ In x (modVarsL C)*)
  destruct (MC2' x) as [[HD' HZ'] | [HD' HZ']]; clear MC2'.
      rewrite HZ in HZ'. inversion HZ'.
  clear HN HZ'. 
      assert (MV: modVars (TErase (TWhile E (theta C PreN PostN HypCN))) x = false).
         simpl. rewrite modVars_theta; trivial.
      destruct (TSem1 _ _ _ _ _ H2 x MV) as [X Y]. 
        rewrite <- X in *. rewrite <- Y in *. clear X Y. 
      destruct (TSem1 _ _ _ _ _ H3 x MV) as [X Y].
        rewrite <- X in *. rewrite <- Y in *. clear X Y.
      rewrite HD. rewrite HD'. apply SIM. rewrite HD. apply leq_refl. 
Qed.

Theorem Theorem3: forall C Gamma Delta (tw: FPCT Gamma C Delta),
    dynamic_sec (FPCCTXT2CTXT Gamma) (theta _ _ _ tw).
Proof.
unfold dynamic_sec, sec. intros.
assert (SUB: SQSUB (FPCCTXT2CTXT Gamma) (FPCCTXT2CTXT Gamma)). apply leq_refl.
apply (Lemma4 _ _ _ tw _ _ _ _ Hyp _ _ _ _ Hyp' SUB SUB x).
split; trivial. apply H. apply H0.
Qed.

Lemma thetaSem_TntSem : forall C Gamma Delta (tw:FPCT Gamma C Delta) s t 
(HS: Sem s C t)  
R (HR: R = (theta _ _ _ tw) )
G (SUB:SQSUB G (FPCCTXT2CTXT Gamma)), 
exists D,
TntSem (s,G) R (t,D). 
Proof. 
intros C G H tw.
induction tw; intros s t SEM.
(*Skip*)
  inversion SEM; intros; subst; simpl in *.
  eexists. constructor; eauto.
(*Assign*) 
  inversion SEM; intros; subst; simpl in *.
  eexists. econstructor; eauto.
(*Comp*)
  inversion SEM; intros; subst; simpl in *.
  destruct (IHtw1 _ _ H2 _ (eq_refl _) _ SUB) as  [K RS1]; clear IHtw1.
  assert (SQ: SQSUB K (FPCCTXT2CTXT G1)). 
      eapply TntSem_FPC. apply RS1. apply SUB.
  destruct (IHtw2 _ _ H4 _ (eq_refl _) _ SQ) as  [HH RS2]; clear IHtw2.
  eexists. econstructor; eauto.
(*Ite*)  
  inversion SEM; intros; subst; simpl in *.
  (*True*) clear IHtw2.
      remember (theta C1
          (fpcupd G pcVar
             (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E))) G1
          tw1) as R1.
      remember ((theta C2
               (fpcupd G pcVar
                  (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E)))
               G2 tw2)) as R2.
      remember (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E)) as p.
      assert (SQ: SQSUB (FPCCTXT2CTXT (mkFPCctxt G0 p)) (FPCCTXT2CTXT (fpcupd G pcVar p))). 
        rewrite (mkCtxtInverse2).  rewrite FPCCTXT2CTXT_upd_pcVar.  apply SUB. 

      destruct (IHtw1 _ _ H6 _ (eq_refl _) _ SQ) as [K RS1]; clear IHtw1.
      destruct (CompCode_exists (Ite E C1 C2) p t K) as [D RS2].
      subst. eexists. econstructor; try eassumption.  
        apply TS_True.  apply H5. rewrite (mkCtxtInverse2) in RS1. apply RS1.

  (*False*) clear IHtw1.
      remember (theta C1
          (fpcupd G pcVar
             (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E))) G1
          tw1) as R1.
      remember ((theta C2
               (fpcupd G pcVar
                  (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E)))
               G2 tw2)) as R2.
      remember (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E)) as p.
      assert (SQ: SQSUB (FPCCTXT2CTXT (mkFPCctxt G0 p)) (FPCCTXT2CTXT (fpcupd G pcVar p))). 
        rewrite (mkCtxtInverse2).  rewrite FPCCTXT2CTXT_upd_pcVar.  apply SUB. 

      destruct (IHtw2 _ _ H6 _ (eq_refl _) _ SQ) as [K RS1]; clear IHtw2.
      destruct (CompCode_exists (Ite E C1 C2) p t K) as [D RS2].
      subst. eexists. econstructor; try eassumption.  
        apply TS_False.  apply H5. rewrite (mkCtxtInverse2) in RS1. apply RS1.

(*While*)
assert (HG : FSQSUB G (pre N)).
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. 
        unfold FLUB, llub; simpl. eapply @leq_trans with (y:=(lub (post N pcVar) (G pcVar))). apply lub_leqright.
         apply lub_leqleft. 
  rewrite fpcupd_Diff. unfold FLUB, llub; simpl.  apply lub_leqright. discriminate.
  apply Le.le_refl.

assert (PostNPreN : FSQSUB (post N) (pre N)).
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. 
        unfold FLUB, llub; simpl. eapply @leq_trans with (y:=(lub (post N pcVar) (G pcVar))). apply lub_leqleft.
         apply lub_leqleft. 
  rewrite fpcupd_Diff. unfold FLUB, llub; simpl.  apply lub_leqleft. discriminate.
  apply Le.le_refl.
remember (llub (G pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) as q.  
assert (q_le: ssqsub q (pre N pcVar)).
  clear SEM. rewrite (PreAboveN N). rewrite PrePost. rewrite Heqq.
             unfold FLUB; simpl. rewrite fpcupd_Same. 
             apply lub_leq. eapply @leq_trans with (y:=llub (post N pcVar) (G pcVar)). apply lub_leqright. apply lub_leqleft.
                eapply @leq_trans with (y:=(GEval (fun x : PVar => llub (post N (progVar x)) (G (progVar x))) E)).
                  apply GEval_mono. intros y. rewrite (PreAboveN N). rewrite PrePost. rewrite fpcupd_Diff. apply leq_refl. discriminate. apply Le.le_refl. 
                  eapply lub_leqright. 
            apply Le.le_refl.
intros R HR.
assert (forall GG : CTXT,
SQSUB GG (FPCCTXT2CTXT (pre N)) ->
exists H1 : CTXT, TntSem (s, GG) R (t, H1)).
  Focus 2. intros. eapply H1. eapply leq_trans. apply SUB. intros y. apply HG. 
remember (While E C) as D.
simpl in HR.
generalize dependent C.
clear PreZ PreBelowN PreAboveN PrePost.
induction SEM; intros; subst; simpl in *; try inversion HeqD.
(*WhileTrue*)
clear IHSEM1. subst. clear HeqD.
  destruct (H1 N _ _ SEM1 _ (eq_refl _) _ H2) as [H RS1].
  assert (SUB_H: SQSUB H (FPCCTXT2CTXT (pre N))).
    eapply @leq_trans with (y:=(FPCCTXT2CTXT (post N))).
      eapply TntSem_FPC. apply RS1. apply H2. intros y. apply PostNPreN.
  destruct (IHSEM2 _ HypC H1 (eq_refl _) (eq_refl _) _ SUB_H) as [K RS2].
  exists K. clear IHSEM2.
  inversion RS2; subst.
  econstructor.
    eapply TS_WTrue. apply H0. apply RS1. apply H6. apply H8.
(*WhileFalse*)
    inversion HeqD; subst. clear H6 H5 HeqD H2.
    destruct (CompCode_exists C0 (llub (G pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) t GG) as [D RS2].
    exists D.
    econstructor. apply TS_WFalse. apply H0.  reflexivity.
    assumption. 
Qed.

Lemma thetaSem_TntSem_Inv: forall C Gamma Delta (tw:FPCT Gamma C Delta)
R (HR: R = (theta _ _ _ tw) ) s t 
G D (HS: TntSem (s,G) R (t,D)), 
Sem s C t. 
Proof. 
intros C Gamma Delta tw.
induction tw; intros R HR s t; rename G into Gamma; intros G D SEM; subst; inversion SEM; subst; clear SEM.
(*Skip*)
  inversion H; subst. econstructor; eauto.
(*Assign*)
  econstructor; eauto. 
(*Comp*)
  destruct r as [r K].
  apply (IHtw1 _ (eq_refl _)) in H2.
  apply (IHtw2 _ (eq_refl _)) in H4.
  econstructor; eauto.
(*Ite*)
destruct r as [r K].
destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [Ht _]; clear H4. subst.
inversion H2; subst; clear H2.
apply (IHtw1 _ (eq_refl _)) in H6. econstructor (solve [eauto]).
apply (IHtw2 _ (eq_refl _)) in H6. econstructor (solve [eauto]).
(*While*)
destruct r as [r K].
destruct (CompCode_Lemma0 _ _ _ _ _ _ H5) as [Ht _]; clear H5. subst.
clear D.
remember (TWhile E (theta C (pre N) (post N) (HypC N))) as R.
remember (t, K) as tK. remember (s,G) as sG.
generalize dependent K.
generalize dependent t.
generalize dependent G.
generalize dependent s.
induction H3; subst; try inversion HeqR.
  subst. intros; subst. clear IHTntSem1.
  destruct r as [r D].
  apply (H0 _ _ (eq_refl _)) in H3_. clear H0.
  assert (Z:= IHTntSem2 (eq_refl _) _ _ (eq_refl _) _ _ (eq_refl _)); clear IHTntSem2.
  econstructor; eauto.
intros; subst. inversion HeqtK; subst. eapply S_WFalse. apply H. trivial.
Qed. 

Theorem Corollary1: forall C Gamma Delta (tw: FPCT Gamma C Delta) 
s G t D (Hyp: TntSem (s,G) (theta C Gamma Delta tw) (t,D)) 
s' t' (Hyp': Sem s' C t')
(HG: G = FPCCTXT2CTXT Gamma) 
x
(SIM: eq_below (D x) G s s'),
t x = t' x.
Proof.
intros. subst.
destruct (thetaSem_TntSem _ _ _ tw _ _ Hyp' _ (eq_refl _) (FPCCTXT2CTXT Gamma)) as [K' tntHyp']. apply leq_refl.
  eapply Lemma4. apply Hyp. apply tntHyp'. apply leq_refl. apply leq_refl. 
   apply sim_eq_below. split. apply SIM. intros z Hz. trivial.
Qed.

(*The rest of the file contains two variants of Lemma 3.8, for comparing
pragrams stemming from different type derivations.*)

(*The first version uses a variant of sim where the pivot (G x) is
  an UPPER bound for (G' x).*)

Definition simsub {A:Type} (k:L) (G G': CTXT) (s s' : PVar -> A) :=
forall x, ssqsub (G x) k -> (s x = s' x /\ ssqsub (G' x) (G x)). 

Lemma simsub_eval: forall E k G G' s s' 
(Hyp: simsub k G G' s s') (SS:ssqsub (GEval G E) k),
eval s E = eval s' E. 
Proof. 
intros E. 
induction E; intros; simpl in *. 
  trivial. 

  apply Hyp.  assumption. 

  destruct (lub_leq_elim _ _ _ SS) as [K1 K2]. 
  rewrite (IHE1 _ _ _ _ _ Hyp K1). 
  rewrite (IHE2 _ _ _ _ _ Hyp K2). 
  trivial. 
Qed. 

Definition simsub_mono: forall {A} k k' G G' s s',
@simsub A k G G' s s' -> ssqsub k' k -> @simsub A k' G G' s s'. 
Proof. 
intros. 
intros x Hx. 
apply H. 
eapply (leq_trans Hx H0). 
Qed. 

Lemma simsub_GEval: forall {A} E k G G' s s' 
(Hyp: @simsub A k G G' s s') (SS:ssqsub (GEval G E) k),
ssqsub (GEval G' E) (GEval G E). 
Proof. 
intros A E. 
induction E; intros; simpl in *. 
  apply ssqsub_refl. 

  apply Hyp.  assumption. 

  destruct (lub_leq_elim _ _ _ SS) as [K1 K2]. 
  assert (KK1 :=IHE1 _ _ _ _ _ Hyp K1); clear IHE1 K1. 
  assert (KK2 :=IHE2 _ _ _ _ _ Hyp K2); clear IHE2 K2. 
  eapply leq_trans. 
    apply leq_monoleft. apply KK1.
    apply leq_monoright. apply KK2. 
Qed. 
 
(*As a result of using simsub, the final primed taint 
  D'x is bounded from above by the unprimed taint D x.
  The hypotheses are such that the primed execution stays below the nonprimed one.*)
Lemma Lemma4_DiffProg: forall Gamma Delta C (tw: FPCT Gamma C Delta) 
Gamma' Delta' (tw': FPCT Gamma' C Delta')
s G t D (Hyp: TntSem (s,G) (theta _ _ _ tw) (t,D)) 
s' G' t' D' (Hyp': TntSem (s',G') (theta _ _ _ tw') (t',D'))
(HGamma: FSQSUB Gamma' Gamma)
(HG: SQSUB G (FPCCTXT2CTXT Gamma)) 
(HG': SQSUB G' (FPCCTXT2CTXT Gamma'))
x
(SIM: simsub (D x) G G' s s'),
t x = t' x /\ ssqsub (D' x) (D x).
Proof.
intros Gamma Delta C tw. 
induction tw; intros. 
(*Skip*)
 subst; simpl in *; rename G into Gamma; rename G0 into G.
remember Skip as C.
induction tw'; simpl in *; intros; subst; try inversion HeqC; clear HeqC.
  inversion Hyp; subst; clear Hyp. 
  inversion H; subst; clear H.  
  inversion Hyp'; subst; clear Hyp'. 
  inversion H; subst; clear H. 
  apply (SIM x). apply leq_refl. 
(*Assign*) 
subst; simpl in *; rename G into Gamma; rename G0 into G.
remember (Assign x E) as C.
induction tw'; simpl in *; intros; subst; try inversion HeqC; clear HeqC.
subst. rename G0 into Gamma'.
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'. simpl in *. rewrite id_bbot in SIM. repeat rewrite id_bbot.
  remember (pvar_eq x x0) as b; destruct b; subst; simpl.
    repeat rewrite M_upd_Same. rewrite M_upd_Same in SIM.  
    split.  eapply simsub_eval. 
             apply SIM. 
             apply llub_sqleft.
    apply lub_leq. apply lubleft. eapply simsub_GEval. apply SIM. apply lub_leqleft.
       apply lubright. apply HGamma.
  repeat rewrite M_upd_Diff; trivial. rewrite M_upd_Diff in SIM; trivial.  
    apply SIM. apply ssqsub_refl.  
(*Comp*)
 subst; simpl in *; rename G into Gamma; rename G0 into G.
  remember (Comp C1 C2) as C. simpl in *.
  induction tw'; simpl in *; intros; subst; try inversion HeqC; clear HeqC.
  subst. rename G0 into Gamma'. rename G1 into Sigma. rename G2 into Delta.
          rename G3 into Sigma'. rename G4 into Delta'. rename tw'1 into tw1'. rename tw'2 into tw2'.
    clear IHtw'1 IHtw'2.
    inversion Hyp; subst; clear Hyp. destruct r as [r H].
    inversion Hyp'; subst; clear Hyp'. destruct r0 as [r' H'].
            eapply (IHtw2 _ _ tw2'); clear IHtw2.
              apply H4.
              apply H7.
              apply (FPC_mono _ _ _ (FPCT_FPC _ _ _ tw1) _ _ (FPCT_FPC _ _ _ tw1') HGamma).
              eapply TntSem_FPC. apply H2. apply HG. 
              eapply TntSem_FPC. apply H5. apply HG'. 
            intros y Hy. 
              eapply (IHtw1 _ _ tw1'); clear IHtw1.
                apply H2. 
                apply H5.
                apply HGamma.  
                apply HG.  
                apply HG'.
                intros z Hz. apply SIM. eapply (leq_trans Hz Hy).
 (*If*)
 subst; simpl in *; rename G into Gamma; rename G0 into G.
  remember (Ite E C1 C2) as C. simpl in *.
  induction tw'; simpl in *; intros; subst; try inversion HeqC.
  subst. autorewrite with fpcREWRITE in *.
    clear HeqC IHtw'1 IHtw'2. rename G0 into Gamma'. rename G1 into Delta1. rename G2 into Delta2.
        rename G3 into Delta1'. rename tw'1 into tw1'. rename G4 into Delta2'. rename tw'2 into tw2'.
    inversion Hyp; subst; clear Hyp. destruct r as [r F].
    inversion Hyp'; subst; clear Hyp'. destruct r0 as [r' F']. 
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [MC1 MC2]. subst.
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H6) as [MC1' MC2']. subst.
    clear H4 H6.
    assert(SUB: ssqsub (llub (Gamma' pcVar) (GEval (fun x0 : PVar => Gamma' (progVar x0)) E))
                       (llub (Gamma pcVar) (GEval (fun x0 : PVar => Gamma (progVar x0)) E))).
        apply lub_leq. apply lubleft. apply HGamma.
                   apply lubright. apply GEval_mono. intros z. apply HGamma.
    assert (HGammaUpd: FSQSUB
              (fpcupd Gamma' pcVar (llub (Gamma' pcVar) (GEval (fun x : PVar => Gamma' (progVar x)) E)))
              (fpcupd Gamma pcVar (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)))).
       clear H2 H3 IHtw1 IHtw2 MC2 MC2' tw1 tw2 tw1' tw2'.
       intros z. destruct z; simpl. repeat rewrite M_upd_Same. apply SUB. 
           repeat rewrite M_upd_Diff; try discriminate. apply HGamma.
    destruct (MC2 x) as [[HD HZ] | [HD HZ]]. 
    (*In  z (modVarsL C1 ++ modVarsL C2)*)
       destruct (MC2' x) as [[HD' _] | [HD' HZ']]. Focus 2. rewrite HZ in HZ'. inversion HZ'. 
        rewrite HD in *.  rewrite HD' in *. clear HD HD' MC2 MC2'. 
        assert (EV: eval s E = eval s' E). 
            clear IHtw1 IHtw2 tw1 tw2 H2 H3. 
            eapply simsub_eval.  apply SIM. apply lubright. apply lubright.
                eapply GEval_mono. apply HG.
          inversion H2; subst; clear H2; simpl in *. 
          (*True*) clear IHtw2. 
            inversion H3; subst; clear H3; simpl in *. 
            (*True*) 
               destruct (IHtw1 _ _ tw1' _ _ _ _ H7 _ _ _ _ H8 HGammaUpd HG HG' x) as [IH_S IH_C]; clear IHtw1 H7 H8 tw1 tw2.
                 eapply simsub_mono. apply SIM. apply lub_leqleft.
               rewrite <- IH_S; clear IH_S. 
               split; trivial.
               apply lub_leq. apply lubleft. apply IH_C.
                   apply lubright. apply SUB.
            (*False*)
              clear IHtw1 H7 H8 tw1 tw2. 
              rewrite EV in H6.  rewrite H6 in H5.  inversion H5. 
          (*False*) clear IHtw1. 
            inversion H3; subst; clear H3; simpl in *. 
            (*True*) 
              clear IHtw2 H7 H8 tw1 tw2. 
              rewrite EV in H6.  rewrite H6 in H5.  inversion H5. 
            (*False*)
               destruct (IHtw2 _ _ tw2' _ _ _ _ H7 _ _ _ _ H8 HGammaUpd HG HG' x) as [IH_S IH_C]; clear IHtw2 H7 H8 tw1 tw2.
                 eapply simsub_mono. apply SIM. apply lub_leqleft.
               rewrite <- IH_S; clear IH_S. 
               split; trivial.
               apply lub_leq. apply lubleft. apply IH_C.
                   apply lubright. apply SUB.
    (* ~ In  z (modVarsL C1 ++ modVarsL C2)*)
      destruct (MC2' x) as [[HD' HZ'] | [HD' _]]. rewrite HZ in HZ'. inversion HZ'. 
      rewrite HD in *.  rewrite HD' in *. clear HD HD' MC2 MC2'. 
 (*      simpl in HZ. apply Bool.orb_false_iff in HZ. destruct HZ as [MV1 MV2].*)
      assert (noMod1: s x = t x /\ G x = F x). 
        eapply TSem1.  apply H2. simpl. repeat rewrite modVars_theta. apply HZ.
      destruct noMod1 as [Y1 Y2] .  rewrite <- Y1 in *.  rewrite <- Y2 in *.  clear Y1 Y2 H2. 
      assert (noMod2: s' x = t' x /\ G' x = F' x). 
        eapply TSem1.  apply H3. simpl. repeat rewrite modVars_theta. apply HZ.
      destruct noMod2 as [Y1 Y2] .  rewrite <- Y1 in *.  rewrite <- Y2 in *.  clear Y1 Y2 H3. 

      clear IHtw1 IHtw2 tw1 tw2. 
      apply SIM. apply ssqsub_refl.
(*While*)
rename H into Delta. rename G into Gamma. rename G0 into G.
assert (HDelta: FSQSUB Delta' Delta).
  eapply (FPC_mono (While E C) Gamma Delta).
     eapply (FPC_While (Gamma pcVar) Gamma E C Delta). reflexivity. apply PreBelowN. apply PreAboveN. apply PreZ. apply PreN. apply PrePost. 
         intros. eapply FPCT_FPC. apply (HypC i).
  eapply FPCT_FPC. apply tw'. apply HGamma.
subst; simpl in *.
clear PreBelowN.
assert (HGammaN : FSQSUB Gamma (pre N)).
  clear Hyp Hyp'.
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. apply lubleft.
        unfold FLUB; simpl.  apply lub_leqright.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqright.
  apply Le.le_refl.
assert (SUB : SQSUB G (FPCCTXT2CTXT (pre N))).
  eapply leq_trans. apply HG. intros z. apply HGammaN.
assert (PostNPreN : FSQSUB (post N) (pre N)).
  clear Hyp Hyp' H0.
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. 
        unfold FLUB; simpl. apply lubleft. apply lub_leqleft.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
  apply Le.le_refl.
remember (llub (Gamma pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) as p.  
assert (HN := H0 N). clear H0.

remember (TComp (TWhile E (theta C (pre N) (post N) (HypC N))) (CompCode p C))  as R.
remember (theta (While E C) _ _ tw') as R'.

cut (forall s GG t DD (HT: TntSem (s, GG) R (t, DD))
            s' GG' t' DD' (HT': TntSem (s', GG') R' (t', DD')) x,
            SQSUB GG (FPCCTXT2CTXT (pre N)) ->
SQSUB GG' (FPCCTXT2CTXT Gamma') ->
simsub (DD x) GG GG' s s' -> t x = t' x /\ ssqsub (DD' x) (DD x)).
  intros IH. eapply IH. apply Hyp. apply Hyp'. apply SUB. apply HG'. apply SIM.
clear Hyp Hyp'. clear SIM x s s' t t'.
remember (While E C) as DD. simpl in *.
induction tw'; simpl in *; intros; subst; try inversion HeqDD. subst.
    clear HeqDD H0. rename pre0 into pre'. rename post0 into post'. rename G0 into Gamma'.
    autorewrite with fpcREWRITE in *. 
    inversion HT; subst; clear HT. destruct r as [r F].
    inversion HT'; subst; clear HT'. destruct r0 as [r' F']. 
rename N0 into N'.

cut (ssqsub (pre' N' pcVar) (pre N pcVar)).
  intros SUB_pc.
   
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H7) as [MC1 MC2].  rewrite <- MC1 in *.  clear MC1.  
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H9) as [MC1' MC2'].  rewrite <- MC1' in *.  clear MC1'. 
    clear H7 H9. 

    assert(SUBGG': ssqsub (llub (Gamma' pcVar) (GEval (fun x0 : PVar => Gamma' (progVar x0)) E))
                       (llub (Gamma pcVar) (GEval (fun x0 : PVar => Gamma (progVar x0)) E))).
        apply lub_leq. apply lubleft. apply HGamma.
                   apply lubright. apply GEval_mono. intros y. apply HGamma.
    assert (HGammaUpd: FSQSUB
              (fpcupd Gamma' pcVar (llub (Gamma' pcVar) (GEval (fun x : PVar => Gamma' (progVar x)) E)))
              (fpcupd Gamma pcVar (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)))).
       intros y. destruct y; simpl. repeat rewrite M_upd_Same. apply SUBGG'. 
           repeat rewrite M_upd_Diff; try discriminate. apply HGamma.
  destruct (MC2 x). destruct H as [HD HZ].
  (*In x (modVarsL C)*)
    destruct (MC2' x); destruct H as [HD' HZ']. 
    (*In x (modVarsL C)*) clear HZ'.
      cut (r x = r' x /\ ssqsub (F' x) (F x)).
         intros HH. destruct HH. rewrite HD. rewrite HD'.
         split. assumption.
         apply lub_leq. apply lubleft. assumption.
         apply lubright.
         eapply leq_trans. apply leq_monoleft. apply HGamma.
            apply leq_monoright. apply GEval_mono.
               intros z. apply (HDelta (progVar z)).
      rewrite HD in *. clear HD HD'. clear MC2 MC2'.
      remember (theta C (pre N) (post N) (HypC N)) as R.
      remember (s, GG) as ss.
      remember (r, F) as tt.
      assert (SUB_G' : FSQSUB Gamma' (pre' N')).
        rewrite (PreAboveN0 N'). rewrite PrePost0. 
        intros z. destruct z. rewrite fpcupd_Same. 
          unfold FLUB, llub; simpl. apply lubleft.  apply lub_leqright.
        rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl.  apply lub_leqright.
        apply Le.le_refl.
 
      assert (Post0NPre0N : FSQSUB (post' N') (pre' N')).
        rewrite (PreAboveN0 N'). rewrite PrePost0. 
        intros z. destruct z. rewrite fpcupd_Same. 
          unfold FLUB, llub; simpl. apply lubleft.  apply lub_leqleft.  
         rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
         apply Le.le_refl.
      remember (llub (Gamma' pcVar) (GEval (fun x : PVar => pre' N' (progVar x)) E)) as q'.  
      assert (q_le': ssqsub q' (pre' N' pcVar)).
      rewrite (PreAboveN0 N'). rewrite PrePost0. rewrite Heqq'.
             unfold FLUB; simpl. rewrite fpcupd_Same. 
             apply lub_leq. apply lubleft. apply lub_leqright.
              apply lubright. apply GEval_mono. intros y. rewrite (PreAboveN0 N'). rewrite PrePost0. apply leq_refl. apply Le.le_refl.
            apply Le.le_refl.
      clear PreZ0 PreAboveN0 PrePost0.
      assert (SUB': SQSUB GG' (FPCCTXT2CTXT(pre' N'))). eapply leq_trans. apply H2. intros z. apply SUB_G'.
      clear H2 SUB_G'.

      generalize dependent GG'. generalize dependent s'.
      generalize dependent GG. generalize dependent s. clear t t'.

      remember (TWhile E R) as RW. generalize dependent R.
      induction H5; intros; try inversion HeqRW; subst; clear HeqRW.
      (*Whiletrue*)
        clear H5_0 IHTntSem1.
        rename s0 into s.
        destruct r0 as [t K].
        assert (EV: eval s E = eval s' E). clear IHTntSem2 HN.
          eapply simsub_eval. apply H3. apply lubright. apply lubright . apply GEval_mono. apply H1. 
         inversion H6; subst; simpl in *.
        (*True*)
          destruct r0 as [t' K'].
          assert (SUBK: SQSUB K (FPCCTXT2CTXT (pre N))). clear HN IHTntSem2.
            eapply leq_trans. eapply TntSem_FPC. apply H5_. assumption.
            intro y. apply PostNPreN.      
          assert (SUBK': SQSUB K' (FPCCTXT2CTXT (pre' N'))). clear HN IHTntSem2.
            eapply leq_trans. eapply TntSem_FPC. apply H7. assumption.
            intro y. apply Post0NPre0N.         
          eapply (IHTntSem2 (eq_refl _) _ (eq_refl _) HN (eq_refl _) t K SUBK (eq_refl _) t' K'); clear IHTntSem2.
             intros y Hy. 
             eapply HN. apply H5_. apply H7.
                 intros z. assert (Z:= HDelta z). destruct z. apply SUB_pc. apply Z.
             assumption. assumption.
             eapply simsub_mono. apply H3. apply Hy.
             apply H9.
             apply SUBK'.
         (*False*) rewrite EV in H. rewrite H in H5. inversion H5.
      (*WhileFalse*) rename s0 into s.
          inversion Heqss; subst; clear Heqss HN.
          assert (EV: eval s E = eval s' E).
            eapply simsub_eval. apply H3. apply lubright. apply lubright. apply GEval_mono. apply H1. 
          inversion H6; subst; clear H6; simpl in *.
          (*True*) rewrite EV in H. rewrite H in H4. inversion H4.
          (*False*) apply eq_sym in H8; inversion H8; subst; clear H8.
                    apply H3. apply lub_leqleft. 
    (*~ In x (modVarsL C)*) rewrite HZ in HZ'.  inversion HZ'.
  (*~ In x (modVarsL C)*) destruct H as [HD HZ].
    destruct (MC2' x); destruct H as [HD' HZ'].
    (*In x (modVarsL C*)  rewrite HZ in HZ'.  inversion HZ'.
    (*~In x (modVarsL C)*) clear HZ'.
        rewrite HD in *. rewrite HD'. clear HD HD' PreZ0 PrePost0 PreAboveN0.
         destruct (TSem1 _ _ _ _ _ H6 x) as [X Y]. simpl. rewrite modVars_theta. apply HZ.
         rewrite <- X. rewrite <- Y in *. clear X Y.
         destruct (TSem1 _ _ _ _ _ H5 x) as [X Y]. simpl. rewrite modVars_theta. apply HZ.
         rewrite <- X. rewrite <- Y in *. clear X Y.
        apply H3. apply leq_refl.
(*discharge of sqsub (pre' N' pcVar) (pre N pcVar)*)
  clear H5 H7 H6 H9 HN. clear r F r' F' H3 s s' t t' DD DD' x.
 clear SUB HG HG' H1 H2 GG GG' G G' D D'.
    assert(SUBGG': ssqsub (llub (Gamma' pcVar) (GEval (fun x0 : PVar => Gamma' (progVar x0)) E))
                       (llub (Gamma pcVar) (GEval (fun x0 : PVar => Gamma (progVar x0)) E))).
        apply lub_leq. apply lubleft. apply HGamma.
                   apply lubright. apply GEval_mono. intros y. apply HGamma.
    assert (HGammaUpd: FSQSUB
              (fpcupd Gamma' pcVar (llub (Gamma' pcVar) (GEval (fun x : PVar => Gamma' (progVar x)) E)))
              (fpcupd Gamma pcVar (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)))).
         intros y. destruct y; simpl. repeat rewrite M_upd_Same. apply SUBGG'. 
             repeat rewrite M_upd_Diff; try discriminate. apply HGamma.
    assert (SUBprej: forall j, FSQSUB (pre' j) (pre j)).
      intros j. induction j; intros.
       (*Base*) rewrite PreZ. rewrite PreZ0. apply HGammaUpd.
       (*Step*) assert (POST: FSQSUB (post' j) (post j)).
                       eapply FPC_mono. eapply FPCT_FPC. apply HypC. eapply FPCT_FPC. apply HypC0. apply IHj. 
            assert (Z:FSQSUB (FLUB (post' j) Gamma')   (FLUB (post j) Gamma) ).
               intros z. unfold FLUB; simpl.
               apply lub_leq. apply lubleft. apply POST. apply lubright. apply HGamma.
           rewrite PrePost. rewrite PrePost0.
           intros z. destruct z. repeat rewrite M_upd_Same.
             apply lub_leq. apply lubleft. apply Z. 
             apply lubright. apply GEval_mono. intros z. apply Z.
             repeat rewrite M_upd_Diff; try discriminate. apply Z.
    assert (SB': forall i j, i <= j -> ssqsub (pre' i pcVar) (pre' j pcVar)).
       intros. clear HGammaUpd SUBGG'. 
          generalize dependent i. 
          induction j; intros. rewrite (Le.le_n_0_eq _ H). apply leq_refl.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             eapply leq_trans. apply IHj. apply Lt.lt_n_Sm_le. apply H0.  rewrite PrePost0. repeat  rewrite M_upd_Same. apply lubleft.  unfold FLUB; simpl. 
                     rewrite (FPC_Lemma1 _ _ _ (FPCT_FPC _ _ _ (HypC0 _))). apply lub_leqleft. 
          subst. apply leq_refl.

   assert (SB: forall i j, i <= j -> ssqsub (pre i pcVar) (pre j pcVar)).
       intros. clear HGammaUpd SUBGG'. 
          generalize dependent i. 
          induction j; intros. rewrite (Le.le_n_0_eq _ H). apply leq_refl.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             eapply leq_trans. apply IHj. apply Lt.lt_n_Sm_le. apply H0.  rewrite PrePost. repeat  rewrite M_upd_Same. apply lubleft.  unfold FLUB; simpl. 
                     rewrite (FPC_Lemma1 _ _ _ (FPCT_FPC _ _ _ (HypC _))). apply lub_leqleft. 
          subst. apply leq_refl.
    assert (PreAboveNStrong : forall i : nat, N <= i -> pre N = pre i).
           intros i; induction i; intros. apply Le.le_n_0_eq in H. rewrite <- H.  trivial.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             apply Lt.lt_n_Sm_le in H0. rewrite IHi. apply PreAboveN. apply H0.  apply H0. 
           subst. trivial. 
    assert (PreAboveNStrong' : forall i : nat, N' <= i -> pre' N' = pre' i).
           intros i; induction i; intros. apply Le.le_n_0_eq in H. rewrite <- H.  trivial.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             apply Lt.lt_n_Sm_le in H0. rewrite IHi. apply PreAboveN0. apply H0.  apply H0. 
           subst. trivial.            
  remember (MinMax.min N N') as M.
  destruct (Min.min_dec N N').
  (*min N N' = N*)
      apply MinMax.MMP.min_l_iff in e.
      assert (Z:= PreAboveNStrong _ e). rewrite Z. apply SUBprej.
  (*min N N' = N'*)
      apply MinMax.MMP.min_r_iff in e.
      assert (Z:= PreAboveNStrong' _ e). rewrite Z.  apply SUBprej.
Qed.

(*The second version uses a variant of sim where the pivot (G x) is
  an LOWER bound for (G' x).*)

Definition ssimsub {A:Type} (k:L) (G G': CTXT) (s s' : PVar -> A) :=
forall x, ssqsub (G x) k -> (s x = s' x /\ ssqsub (G x) (G' x)). 

Lemma ssimsub_eval: forall E k G G' s s' 
(Hyp: ssimsub k G G' s s') (SS:ssqsub (GEval G E) k),
eval s E = eval s' E. 
Proof. 
intros E. 
induction E; intros; simpl in *. 
  trivial. 

  apply Hyp.  assumption. 

  destruct (lub_leq_elim _ _ _ SS) as [K1 K2]. 
  rewrite (IHE1 _ _ _ _ _ Hyp K1). 
  rewrite (IHE2 _ _ _ _ _ Hyp K2). 
  trivial. 
Qed. 

Definition ssimsub_mono: forall {A} k k' G G' s s',
@ssimsub A k G G' s s' -> ssqsub k' k -> @ssimsub A k' G G' s s'. 
Proof. 
intros. 
intros x Hx. 
apply H. 
eapply (leq_trans Hx H0). 
Qed. 

Lemma ssimsub_GEval: forall {A} E k G G' s s' 
(Hyp: @ssimsub A k G G' s s') (SS:ssqsub (GEval G E) k),
ssqsub (GEval G E) (GEval G' E). 
Proof. 
intros A E. 
induction E; intros; simpl in *. 
  apply ssqsub_refl. 

  apply Hyp.  assumption. 

  destruct (lub_leq_elim _ _ _ SS) as [K1 K2]. 
  assert (KK1 :=IHE1 _ _ _ _ _ Hyp K1); clear IHE1 K1. 
  assert (KK2 :=IHE2 _ _ _ _ _ Hyp K2); clear IHE2 K2. 
  eapply @leq_trans with (y:=llub (GEval G' E1) (GEval G E2)). 
    apply leq_monoleft.  assumption. 
    apply leq_monoright.  assumption. 
Qed. 

(*As a result of using ssimsub, the final primed taint 
  D'x is bounded from BELOW by the unprimed taint D x.
  The hypotheses are such that the primed execution stays ABOVE the nonprimed one.*)
Lemma sLemma4_DiffProg: forall Gamma Delta C (tw: FPCT Gamma C Delta) 
Gamma' Delta' (tw': FPCT Gamma' C Delta')
s G t D (Hyp: TntSem (s,G) (theta _ _ _ tw) (t,D)) 
s' G' t' D' (Hyp': TntSem (s',G') (theta _ _ _ tw') (t',D'))
(HGamma: FSQSUB Gamma Gamma')
(HG: SQSUB G (FPCCTXT2CTXT Gamma)) 
(HG': SQSUB G' (FPCCTXT2CTXT Gamma'))
x
(SIM: ssimsub (D x) G G' s s'),
t x = t' x /\ ssqsub (D x) (D' x).
Proof.
intros Gamma Delta C tw. 
induction tw; intros; subst; simpl in *; rename G into Gamma; rename G0 into G. 
(*Skip*)
remember Skip as C.
induction tw'; simpl in *; intros; subst; try inversion HeqC; clear HeqC.
  inversion Hyp; subst; clear Hyp. 
  inversion H; subst; clear H.  
  inversion Hyp'; subst; clear Hyp'. 
  inversion H; subst; clear H. 
  apply (SIM x). apply leq_refl. 
(*Assign*)
remember (Assign x E) as C.
induction tw'; simpl in *; intros; subst; try inversion HeqC; clear HeqC.
subst. rename G0 into Gamma'.
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'. simpl in *. rewrite id_bbot in SIM. repeat rewrite id_bbot.
  remember (pvar_eq x x0) as b; destruct b; subst; simpl.
    repeat rewrite M_upd_Same. rewrite M_upd_Same in SIM.  
    split.  eapply ssimsub_eval. 
             apply SIM. 
             apply llub_sqleft.
    apply lub_leq. apply lubleft. eapply ssimsub_GEval. apply SIM. apply lub_leqleft.
       apply lubright. apply HGamma.
  repeat rewrite M_upd_Diff; trivial. rewrite M_upd_Diff in SIM; trivial.  
    apply SIM. apply ssqsub_refl.  
(*Comp*)
  remember (Comp C1 C2) as C. simpl in *.
  induction tw'; simpl in *; intros; subst; try inversion HeqC; clear HeqC.
  subst. rename G0 into Gamma'. rename G1 into Sigma. rename G2 into Delta.
          rename G3 into Sigma'. rename G4 into Delta'. rename tw'1 into tw1'. rename tw'2 into tw2'.
    clear IHtw'1 IHtw'2.
    inversion Hyp; subst; clear Hyp. destruct r as [r H].
    inversion Hyp'; subst; clear Hyp'. destruct r0 as [r' H'].
            eapply (IHtw2 _ _ tw2'); clear IHtw2.
              apply H4.
              apply H7.
              apply (FPC_mono _ _ _ (FPCT_FPC _ _ _ tw1') _ _ (FPCT_FPC _ _ _ tw1) HGamma).
              eapply TntSem_FPC. apply H2. apply HG. 
              eapply TntSem_FPC. apply H5. apply HG'. 
            intros y Hy. 
              eapply (IHtw1 _ _ tw1'); clear IHtw1.
                apply H2. 
                apply H5.
                apply HGamma.  
                apply HG.  
                apply HG'.
                intros z Hz. apply SIM. eapply (leq_trans Hz Hy).
 (*If*)
  remember (Ite E C1 C2) as C. simpl in *.
  induction tw'; simpl in *; intros; subst; try inversion HeqC.
  subst. autorewrite with fpcREWRITE in *.
    clear HeqC IHtw'1 IHtw'2. rename G0 into Gamma'. rename G1 into Delta1. rename G2 into Delta2.
        rename G3 into Delta1'. rename tw'1 into tw1'. rename G4 into Delta2'. rename tw'2 into tw2'.
    inversion Hyp; subst; clear Hyp. destruct r as [r F].
    inversion Hyp'; subst; clear Hyp'. destruct r0 as [r' F']. 
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H4) as [MC1 MC2]. subst.
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H6) as [MC1' MC2']. subst.
    clear H4 H6.
    assert(SUB: ssqsub (llub (Gamma pcVar) (GEval (fun x0 : PVar => Gamma (progVar x0)) E))
                       (llub (Gamma' pcVar) (GEval (fun x0 : PVar => Gamma' (progVar x0)) E))).
        apply lub_leq. apply lubleft. apply HGamma.
                   apply lubright. apply GEval_mono. intros z. apply HGamma.
    assert (HGammaUpd: FSQSUB
              (fpcupd Gamma pcVar (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)))
              (fpcupd Gamma' pcVar (llub (Gamma' pcVar) (GEval (fun x : PVar => Gamma' (progVar x)) E)))).
       clear H2 H3 IHtw1 IHtw2 MC2 MC2' tw1 tw2 tw1' tw2'.
       intros z. destruct z; simpl. repeat rewrite M_upd_Same. apply SUB. 
           repeat rewrite M_upd_Diff; try discriminate. apply HGamma.
    destruct (MC2 x) as [[HD HZ] | [HD HZ]]. 
    (*In  z (modVarsL C1 ++ modVarsL C2)*)
       destruct (MC2' x) as [[HD' _] | [HD' HZ']]. Focus 2. rewrite HZ in HZ'. inversion HZ'. 
        rewrite HD in *.  rewrite HD' in *. clear HD HD' MC2 MC2'. 
        assert (EV: eval s E = eval s' E). 
            clear IHtw1 IHtw2 tw1 tw2 H2 H3. 
            eapply ssimsub_eval.  apply SIM. apply lubright. apply lubright.
                eapply GEval_mono. apply HG.
          inversion H2; subst; clear H2; simpl in *. 
          (*True*) clear IHtw2. 
            inversion H3; subst; clear H3; simpl in *. 
            (*True*) 
               destruct (IHtw1 _ _ tw1' _ _ _ _ H7 _ _ _ _ H8 HGammaUpd HG HG' x) as [IH_S IH_C]; clear IHtw1 H7 H8 tw1 tw2.
                 eapply ssimsub_mono. apply SIM. apply lub_leqleft.
               rewrite <- IH_S; clear IH_S. 
               split; trivial.
               apply lub_leq. apply lubleft. apply IH_C.
                   apply lubright. apply SUB.
            (*False*)
              clear IHtw1 H7 H8 tw1 tw2. 
              rewrite EV in H6.  rewrite H6 in H5.  inversion H5. 
          (*False*) clear IHtw1. 
            inversion H3; subst; clear H3; simpl in *. 
            (*True*) 
              clear IHtw2 H7 H8 tw1 tw2. 
              rewrite EV in H6.  rewrite H6 in H5.  inversion H5. 
            (*False*)
               destruct (IHtw2 _ _ tw2' _ _ _ _ H7 _ _ _ _ H8 HGammaUpd HG HG' x) as [IH_S IH_C]; clear IHtw2 H7 H8 tw1 tw2.
                 eapply ssimsub_mono. apply SIM. apply lub_leqleft.
               rewrite <- IH_S; clear IH_S. 
               split; trivial.
               apply lub_leq. apply lubleft. apply IH_C.
                   apply lubright. apply SUB.
    (* ~ In  z (modVarsL C1 ++ modVarsL C2)*)
      destruct (MC2' x) as [[HD' HZ'] | [HD' _]]. rewrite HZ in HZ'. inversion HZ'. 
      rewrite HD in *.  rewrite HD' in *. clear HD HD' MC2 MC2'. 
 (*      simpl in HZ. apply Bool.orb_false_iff in HZ. destruct HZ as [MV1 MV2].*)
      assert (noMod1: s x = t x /\ G x = F x). 
        eapply TSem1.  apply H2. simpl. repeat rewrite modVars_theta. apply HZ.
      destruct noMod1 as [Y1 Y2] .  rewrite <- Y1 in *.  rewrite <- Y2 in *.  clear Y1 Y2 H2. 
      assert (noMod2: s' x = t' x /\ G' x = F' x). 
        eapply TSem1.  apply H3. simpl. repeat rewrite modVars_theta. apply HZ.
      destruct noMod2 as [Y1 Y2] .  rewrite <- Y1 in *.  rewrite <- Y2 in *.  clear Y1 Y2 H3. 

      clear IHtw1 IHtw2 tw1 tw2. 
      apply SIM. apply ssqsub_refl.
(*While*)
remember (fpcupd (pre N) pcVar (Gamma pcVar)) as Delta.
assert (HDelta: FSQSUB Delta Delta').
  eapply (FPC_mono (While E C) Gamma' Delta'). 
    eapply FPCT_FPC. apply tw'.
     eapply (FPC_While (Gamma pcVar) Gamma E C Delta). reflexivity. apply PreBelowN. apply PreAboveN. apply PreZ. subst.  trivial. apply PrePost. 
         intros. eapply FPCT_FPC. apply (HypC i).
   apply HGamma.
clear PreBelowN.
assert (HGammaN : FSQSUB Gamma (pre N)).
  clear Hyp Hyp'.
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. apply lubleft.
        unfold FLUB; simpl.  apply lub_leqright.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqright.
  apply Le.le_refl.
assert (SUB : SQSUB G (FPCCTXT2CTXT (pre N))).
  eapply leq_trans. apply HG. intros z. apply HGammaN.
assert (PostNPreN : FSQSUB (post N) (pre N)).
  clear Hyp Hyp' H0.
  rewrite (PreAboveN N). rewrite PrePost. 
  intros z. destruct z. rewrite fpcupd_Same. 
        unfold FLUB; simpl. apply lubleft. apply lub_leqleft.
  rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
  apply Le.le_refl.
remember (llub (Gamma pcVar) (GEval (fun x : PVar => pre N (progVar x)) E)) as p.  
assert (HN := H0 N). clear H0.

remember (TComp (TWhile E (theta C (pre N) (post N) (HypC N))) (CompCode p C))  as R.
remember (theta (While E C) _ _ tw') as R'.

cut (forall s GG t DD (HT: TntSem (s, GG) R (t, DD))
            s' GG' t' DD' (HT': TntSem (s', GG') R' (t', DD')) x,
            SQSUB GG (FPCCTXT2CTXT (pre N)) ->
SQSUB GG' (FPCCTXT2CTXT Gamma') ->
ssimsub (DD x) GG GG' s s' -> t x = t' x /\ ssqsub (DD x) (DD' x)).
  intros IH. eapply IH. apply Hyp. apply Hyp'. apply SUB. apply HG'. apply SIM.
clear Hyp Hyp'. clear SIM x s s' t t'.
remember (While E C) as DD. simpl in *.
induction tw'; simpl in *; intros; subst; try inversion HeqDD. subst.
    clear HeqDD H0. rename pre0 into pre'. rename post0 into post'. rename G0 into Gamma'.
    autorewrite with fpcREWRITE in *. 
    inversion HT; subst; clear HT. destruct r as [r F].
    inversion HT'; subst; clear HT'. destruct r0 as [r' F']. 
rename N0 into N'.
cut (ssqsub (pre N pcVar) (pre' N' pcVar)).
  intros SUB_pc.
   
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H7) as [MC1 MC2].  rewrite <- MC1 in *.  clear MC1.  
    destruct (CompCode_Lemma0 _ _ _ _ _ _ H9) as [MC1' MC2'].  rewrite <- MC1' in *.  clear MC1'. 
    clear H7 H9. 

    assert(SUBGG': ssqsub (llub (Gamma pcVar) (GEval (fun x0 : PVar => Gamma (progVar x0)) E))
                       (llub (Gamma' pcVar) (GEval (fun x0 : PVar => Gamma' (progVar x0)) E))).
        apply lub_leq. apply lubleft. apply HGamma.
                   apply lubright. apply GEval_mono. intros y. apply HGamma.
    assert (HGammaUpd: FSQSUB
              (fpcupd Gamma pcVar (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)))
              (fpcupd Gamma' pcVar (llub (Gamma' pcVar) (GEval (fun x : PVar => Gamma' (progVar x)) E)))).
       intros y. destruct y; simpl. repeat rewrite M_upd_Same. apply SUBGG'. 
           repeat rewrite M_upd_Diff; try discriminate. apply HGamma.
  destruct (MC2 x). destruct H as [HD HZ].
  (*In x (modVarsL C)*)
    destruct (MC2' x); destruct H as [HD' HZ']. 
    (*In x (modVarsL C)*) clear HZ'.
       cut (r x = r' x /\ ssqsub (F x) (F' x)).
         intros HH. destruct HH. rewrite HD. rewrite HD'.
           split. assumption.
           apply lub_leq. apply lubleft. assumption.
           apply lubright.
           eapply leq_trans. apply leq_monoleft. apply HGamma.
              apply leq_monoright. apply GEval_mono.
                 intros z. apply (HDelta (progVar z)).
      rewrite HD in *. clear HD HD'. clear MC2 MC2'.
      remember (theta C (pre N) (post N) (HypC N)) as R.
      remember (s, GG) as ss.
      remember (r, F) as tt.
      assert (SUB_G' : FSQSUB Gamma' (pre' N')).
        rewrite (PreAboveN0 N'). rewrite PrePost0. 
        intros z. destruct z. rewrite fpcupd_Same. 
          unfold FLUB, llub; simpl. apply lubleft.  apply lub_leqright.
        rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl.  apply lub_leqright.
        apply Le.le_refl.
 
      assert (Post0NPre0N : FSQSUB (post' N') (pre' N')).
        rewrite (PreAboveN0 N'). rewrite PrePost0. 
        intros z. destruct z. rewrite fpcupd_Same. 
          unfold FLUB, llub; simpl. apply lubleft.  apply lub_leqleft.  
         rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
         apply Le.le_refl.
      remember (llub (Gamma' pcVar) (GEval (fun x : PVar => pre' N' (progVar x)) E)) as q'.  
      assert (q_le': ssqsub q' (pre' N' pcVar)).
        rewrite (PreAboveN0 N'). rewrite PrePost0. rewrite Heqq'.
             unfold FLUB; simpl. rewrite fpcupd_Same. 
             apply lub_leq. apply lubleft. apply lub_leqright.
              apply lubright. apply GEval_mono. intros y. rewrite (PreAboveN0 N'). rewrite PrePost0. apply leq_refl. apply Le.le_refl.
            apply Le.le_refl.
      clear PreZ0 PreAboveN0 PrePost0.
      assert (SUB': SQSUB GG' (FPCCTXT2CTXT(pre' N'))). eapply leq_trans. apply H2. intros z. apply SUB_G'.
      clear H2 SUB_G'.

      generalize dependent GG'. generalize dependent s'.
      generalize dependent GG. generalize dependent s. clear t t'.

      remember (TWhile E R) as RW. generalize dependent R.
      induction H5; intros; try inversion HeqRW; subst; clear HeqRW.
      (*Whiletrue*)
        clear H5_0 IHTntSem1.
        rename s0 into s.
        destruct r0 as [t K].
        assert (EV: eval s E = eval s' E). clear IHTntSem2 HN.
          eapply ssimsub_eval. apply H3. apply lubright. apply lubright . apply GEval_mono. apply H1. 
        inversion H6; subst; simpl in *.
        (*True*)
          destruct r0 as [t' K'].
          assert (SUBK: SQSUB K (FPCCTXT2CTXT (pre N))). clear HN IHTntSem2.
            eapply leq_trans. eapply TntSem_FPC. apply H5_. assumption.
            intro y. apply PostNPreN.      
          assert (SUBK': SQSUB K' (FPCCTXT2CTXT (pre' N'))). clear HN IHTntSem2.
            eapply leq_trans. eapply TntSem_FPC. apply H7. assumption.
            intro y. apply Post0NPre0N.         
          eapply (IHTntSem2 (eq_refl _) _ (eq_refl _) HN (eq_refl _) t K SUBK (eq_refl _) t' K'); clear IHTntSem2.
             intros y Hy. 
             eapply HN. apply H5_. apply H7.
                 intros z. assert (Z:= HDelta z). destruct z. apply SUB_pc. apply Z.
             assumption. assumption.
             eapply ssimsub_mono. apply H3. apply Hy.
             apply H9.
             apply SUBK'.
        (*False*) rewrite EV in H. rewrite H in H5. inversion H5.
      (*WhileFalse*) rename s0 into s.
         inversion Heqss; subst; clear Heqss HN.
         assert (EV: eval s E = eval s' E).
           eapply ssimsub_eval. apply H3. apply lubright. apply lubright. apply GEval_mono. apply H1. 
         inversion H6; subst; clear H6; simpl in *.
         (*True*) rewrite EV in H. rewrite H in H4. inversion H4.
         (*False*) apply eq_sym in H8; inversion H8; subst; clear H8.
                   apply H3. apply lub_leqleft. 
    (*~ In x (modVarsL C)*) rewrite HZ in HZ'.  inversion HZ'.
  (*~ In x (modVarsL C)*) destruct H as [HD HZ].
     destruct (MC2' x); destruct H as [HD' HZ'].
   (*In x (modVarsL C*)  rewrite HZ in HZ'.  inversion HZ'.
    (*~In x (modVarsL C)*) clear HZ'.
        rewrite HD in *. rewrite HD'. clear HD HD' PreZ0 PrePost0 PreAboveN0.
         destruct (TSem1 _ _ _ _ _ H6 x) as [X Y]. simpl. rewrite modVars_theta. apply HZ.
         rewrite <- X. rewrite <- Y in *. clear X Y.
         destruct (TSem1 _ _ _ _ _ H5 x) as [X Y]. simpl. rewrite modVars_theta. apply HZ.
         rewrite <- X. rewrite <- Y in *. clear X Y.
        apply H3. apply leq_refl.
(*discharge of FSQSUB (pre' N' pcVar) (pre N pcVar)*)
  clear H5 H7 H6 H9 HN. clear r F r' F' H3 s s' t t' DD DD' x.
  clear SUB HG HG' H1 H2 GG GG' G G' D D'.
    assert(SUBGG': ssqsub (llub (Gamma pcVar) (GEval (fun x0 : PVar => Gamma (progVar x0)) E))
                       (llub (Gamma' pcVar) (GEval (fun x0 : PVar => Gamma' (progVar x0)) E))).
        apply lub_leq. apply lubleft. apply HGamma.
                   apply lubright. apply GEval_mono. intros y. apply HGamma.
    assert (HGammaUpd: FSQSUB
              (fpcupd Gamma pcVar (llub (Gamma pcVar) (GEval (fun x : PVar => Gamma (progVar x)) E)))
              (fpcupd Gamma' pcVar (llub (Gamma' pcVar) (GEval (fun x : PVar => Gamma' (progVar x)) E)))).
         intros y. destruct y; simpl. repeat rewrite M_upd_Same. apply SUBGG'. 
             repeat rewrite M_upd_Diff; try discriminate. apply HGamma.
    assert (SUBprej: forall j, FSQSUB (pre j) (pre' j)).
      intros j. induction j; intros.
       (*Base*) rewrite PreZ. rewrite PreZ0. apply HGammaUpd.
       (*Step*) assert (POST: FSQSUB (post j) (post' j)).
                       eapply FPC_mono. eapply FPCT_FPC. apply HypC0. eapply FPCT_FPC. apply HypC. apply IHj. 
            assert (Z:FSQSUB (FLUB (post j) Gamma)   (FLUB (post' j) Gamma') ).
               intros z. unfold FLUB; simpl.
               apply lub_leq. apply lubleft. apply POST. apply lubright. apply HGamma.
           rewrite PrePost. rewrite PrePost0.
           intros z. destruct z. repeat rewrite M_upd_Same.
             apply lub_leq. apply lubleft. apply Z. 
             apply lubright. apply GEval_mono. intros z. apply Z.
             repeat rewrite M_upd_Diff; try discriminate. apply Z.
    assert (SB': forall i j, i <= j -> ssqsub (pre' i pcVar) (pre' j pcVar)).
       intros. clear HGammaUpd SUBGG'. 
          generalize dependent i. 
          induction j; intros. rewrite (Le.le_n_0_eq _ H). apply leq_refl.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             eapply leq_trans. apply IHj. apply Lt.lt_n_Sm_le. apply H0.  rewrite PrePost0. repeat  rewrite M_upd_Same. apply lubleft.  unfold FLUB; simpl. 
                     rewrite (FPC_Lemma1 _ _ _ (FPCT_FPC _ _ _ (HypC0 _))). apply lub_leqleft. 
          subst. apply leq_refl.

   assert (SB: forall i j, i <= j -> ssqsub (pre i pcVar) (pre j pcVar)).
       intros. clear HGammaUpd SUBGG'. 
          generalize dependent i. 
          induction j; intros. rewrite (Le.le_n_0_eq _ H). apply leq_refl.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             eapply leq_trans. apply IHj. apply Lt.lt_n_Sm_le. apply H0.  rewrite PrePost. repeat  rewrite M_upd_Same. apply lubleft.  unfold FLUB; simpl. 
                     rewrite (FPC_Lemma1 _ _ _ (FPCT_FPC _ _ _ (HypC _))). apply lub_leqleft. 
          subst. apply leq_refl.
    assert (PreAboveNStrong : forall i : nat, N <= i -> pre N = pre i).
           intros i; induction i; intros. apply Le.le_n_0_eq in H. rewrite <- H.  trivial.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             apply Lt.lt_n_Sm_le in H0. rewrite IHi. apply PreAboveN. apply H0.  apply H0. 
           subst. trivial. 
    assert (PreAboveNStrong' : forall i : nat, N' <= i -> pre' N' = pre' i).
           intros i; induction i; intros. apply Le.le_n_0_eq in H. rewrite <- H.  trivial.
          destruct (Lt.le_lt_or_eq _ _ H); clear H.             
             apply Lt.lt_n_Sm_le in H0. rewrite IHi. apply PreAboveN0. apply H0.  apply H0. 
           subst. trivial.            
  remember (MinMax.min N N') as M.
  destruct (Min.min_dec N N').
  (*min N N' = N*)
      apply MinMax.MMP.min_l_iff in e.
      assert (Z:= PreAboveNStrong _ e). rewrite Z. apply SUBprej.
  (*min N N' = N'*)
      apply MinMax.MMP.min_r_iff in e.
      assert (Z:= PreAboveNStrong' _ e). rewrite Z.  apply SUBprej.
Qed.