(*based on mkU_Term*)
Require Export JoinLanguage. 
Require Export SynthesisIota. 
(*towards the end of the file we also import IotaThea*)

Fixpoint xi (C:Com) : (list nat) -> JCom SVar :=
match C with
  Skip => fun F => (JSkip SVar)
| Assign x e => fun F => 
       JComp SVar (JAssign _ x e) 
                  (JJoin _ (var_svar (progVar x)) nil ((allPCs F)++(mkSVars e)))
| Comp C1 C2 => fun F => JComp _ (xi C1 F) (xi C2 F)
| Ite e C1 C2 => (fun F => 
      match (fresh F) with exist n Pn => 
        let termRegs := (var_svar pcVar) :: (mkSVars e) in
        let secregsPre := (allPCs F) ++ (mkSVars e) 
        in JComp _ (JJoin _ (var_svar pcVar) (e::nil) termRegs)
           (JComp _ (JJoin _ (cond_svar n) (e::nil) secregsPre  )
                    (JIte _ e (JComp _ (xi C1 (n::F)) (mkJoins n (modVarsL C2)))
                             (JComp _ (xi C2 (n::F)) (mkJoins n (modVarsL C1)))))
      end)
| While e C => fun F => 
      match (fresh F) with exist n Pn => 
        let termRegs := (var_svar pcVar) :: (mkSVars e) in
        let secregsPre := (allPCs F) ++ (mkSVars e) in
        let secregsInner := (allPCs (n::F))  ++ (mkSVars e)
        in JComp _ (JJoin _ (var_svar pcVar) (e::nil) termRegs)
           (JComp _ (JJoin _ (cond_svar n) (e::nil) secregsPre)
                     (JComp _ (JWhile _ e (JComp _ (xi C (n::F))
                                            (JComp _ (mkJoins n (modVarsL C))
                                              (JComp _ (JJoin _ (var_svar pcVar) (e::nil) termRegs)
                                                        (JJoin _ (cond_svar n) (e::nil) secregsInner)))))
                              (mkJoins n (modVarsL C))))
        end
end.

Lemma modvars_xi: forall C F,
modVars (JErase _ (xi C F)) = modVars C.
Proof.
intros C. 
induction C; simpl; intros.
(*Skip*)
  reflexivity.
(*Assign*)
  apply extensionality. intros z.
  rewrite Bool.orb_false_r. trivial.
(*Comp*)
  apply extensionality. intros z.
  rewrite IHC1. rewrite IHC2. trivial.
(*Ite*)
  remember (fresh F) as FRESH; destruct FRESH as [n Hn].
  apply extensionality. intros z.
  simpl in *.    
  rewrite IHC1. clear IHC1.
  rewrite IHC2. clear IHC2.
  repeat rewrite modVars_mkJoins.
  repeat rewrite Bool.orb_false_r. trivial.
(*While*)
  remember (fresh F) as FRESH; destruct FRESH as [n Hn].
  apply extensionality. intros z.
  simpl in *.    
  rewrite IHC. clear IHC.
  rewrite modVars_mkJoins.
  repeat rewrite Bool.orb_false_r. trivial.
Qed.

Lemma xi_noMod: forall C F U
  (HU: U = xi C F) 
  s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x,
  modVars (JErase _ U) x = false ->
  M (var_svar (progVar x)) = N (var_svar (progVar x)).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem. trivial.
(*Assign*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H3; subst; clear H3. 
  repeat rewrite Bool.orb_false_r in H. apply pvar_eq_ElimF in H. 
  inversion H5; subst; clear H5.
  rewrite supd_Diff. trivial.
  intros Z. inversion Z; subst. apply H. trivial. 
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  apply Bool.orb_false_elim in H.
  destruct H as [MV1 MV2].
  rewrite <- (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H5 x MV2).
  apply (IHC1 _ _ (eq_refl _) _ _ _ _ _ _  H3 x MV1).
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  simpl in H.
  apply Bool.orb_false_elim in H.
  destruct H as [MV1 MV2].
  apply Bool.orb_false_elim in MV2.
  destruct MV2 as [MV2a _].
  apply Bool.orb_false_elim in MV1.
  destruct MV1 as [MV1a _].
(*  rewrite modVars_iota in *.*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5; simpl in *.
     (*True*) clear H4 IHC2.
        inversion H6; subst; clear H6. 
        destruct r0 as [[s K] Q].
        rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H4). clear H4.
        rewrite <- (IHC1 _ _ (eq_refl _) _ _ _ _ _ _  H2 x). clear IHC1 H2.
        repeat rewrite supd_Diff; try discriminate. trivial.
        apply MV1a.
        apply modVarsChar'. rewrite modvars_xi in MV2a. apply MV2a.
     (*False*) clear H4 IHC1.
        inversion H6; subst; clear H6. 
        destruct r0 as [[s K] Q].
        rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H4). clear H4.
        rewrite <- (IHC2 _ _ (eq_refl _) _ _ _ _ _ _  H2 x). clear IHC2 H2.
        repeat rewrite supd_Diff; try discriminate. trivial.
        apply MV2a.
        apply modVarsChar'. rewrite modvars_xi in MV1a. apply MV1a.
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  simpl in H.
  apply Bool.orb_false_elim in H.
  destruct H as [MV1 _].
  apply Bool.orb_false_elim in MV1.
  destruct MV1 as [MV2 _].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4. 
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r0 as [[s G] P].
  rewrite <- (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H4). clear H4.
  assert (IHCC := IHC (n::F) _ (eq_refl _)). clear IHC.
  assert (IHW : forall (s : Store) (G : CTXT) (M : ZStore SVar) (t : Store) (D : CTXT)
         (N : ZStore SVar),
       JoinSem SVar SVar_eq (s, G, M) (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e)))))) (t, D, N) ->
       forall x : PVar,
       modVars (JErase SVar (xi C (n :: F))) x = false ->
       M (var_svar (progVar x)) = N (var_svar (progVar x))).
    clear H2 HeqFRESH FRn. clear s G M t D N. clear r P E.
    remember (xi C (n :: F)) as R.
    intros s G M t D N HRS.
    remember (s, G, M) as ST.
    remember (t, D, N) as TT.
    remember (JWhile SVar e
           (JComp SVar R
              (JComp SVar (mkJoins n (modVarsL C))
                 (JComp SVar
                    (JJoin SVar (var_svar pcVar) (e :: nil)
                       (var_svar pcVar :: mkSVars e))
                    (JJoin SVar (cond_svar n) (e :: nil)
                       (cond_svar n :: allPCs F ++ mkSVars e)))))) as PP.
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction HRS; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear IHHRS1. destruct r as [[r A] Q].
       inversion HRS1; subst; clear HRS1.  destruct r0 as [[u B] O].
       inversion H6; subst; clear H6. 
       inversion H8; subst; clear H8. 
       inversion H9; subst; clear H9. 
       inversion H6; subst; clear H6. 
       rewrite (IHCC _ _ _ _ _ _ H4 _ H0). clear H4 IHCC.
       assert (LP := IHHRS2 (eq_refl _)). clear IHHRS2.
       rewrite <- (LP _ _ _ (eq_refl _) _ _ _ (eq_refl _) _ H0). clear LP.
       repeat rewrite supd_Diff; try discriminate.
       apply (mkJoins_Notin_SecStore1 _ _ _ _ _ _ _ _ H5).
         rewrite modvars_xi in H0. apply modVarsChar'. assumption.
    (*False*)
       inversion HeqTT; subst. trivial.
  clear IHCC.
    rewrite <- (IHW _ _ _ _ _ _ H2  _ MV2). clear IHW H2.
    repeat rewrite supd_Diff; try discriminate. trivial.

  rewrite modvars_xi in MV2.
     intros Z. apply modVarsChar' in MV2. apply (MV2 Z).
Qed.

Lemma xi_ControlLabelsPreserved: forall C F U
  (HR: U = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N))
  n (IN: In n F),
  M (cond_svar n) = N (cond_svar n).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem. trivial.
(*Assign*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.  
  inversion H4; subst; clear H4. 
  repeat rewrite supd_Diff; try discriminate. trivial.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  rewrite (IHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2 _ IN).
  apply (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H4 _ IN).
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5; simpl in *.
  (*True*)
     clear IHC2 H4.
     inversion H6; subst; clear H6. destruct r0 as [[u B] Q].
     rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4 n). clear H4.
       rewrite <- (IHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2). clear IHC1 H2.
       repeat rewrite supd_Diff; try discriminate. trivial.
     intros Z. inversion Z; subst. apply FRm. trivial.
     right; trivial.
  (*False*)
     clear IHC1 H4.
     inversion H6; subst; clear H6. destruct r0 as [[u B] Q].
     rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4 n). clear H4.
       rewrite <- (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H2). clear IHC2 H2.
       repeat rewrite supd_Diff; try discriminate. trivial.
     intros Z. inversion Z; subst. apply FRm. trivial.
     right; trivial. 
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r0 as [[s G] P].
  rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4 n). clear H4.
  assert (HC := IHC (m::F) _ (eq_refl _)). clear IHC.
  assert (IH: forall s G M t D N,
      JoinSem _ SVar_eq (s, G, M)  (JWhile SVar e
          (JComp SVar (xi C (m :: F))
             (JComp SVar (mkJoins m (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar m) (e :: nil)
                      (cond_svar m :: allPCs F ++ mkSVars e)))))) (t, D, N) ->
      forall n : nat, In n F -> M (cond_svar n) = N (cond_svar n)).

    clear H2 s G M r E P HeqFRESH t D N IN.
    intros s G M t D N HRS.
    remember (s, G, M) as ST.
    remember (t, D, N) as TT.
    remember (JWhile SVar e
           (JComp SVar (xi C (m :: F))
              (JComp SVar (mkJoins m (modVarsL C))
                 (JComp SVar
                    (JJoin SVar (var_svar pcVar) (e :: nil)
                       (var_svar pcVar :: mkSVars e))
                    (JJoin SVar (cond_svar m) (e :: nil)
                       (cond_svar m :: allPCs F ++ mkSVars e)))))) as PP. 
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction HRS; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear IHHRS1. destruct r as [[r A] Q].
       inversion HRS1; subst; clear HRS1.
       inversion H6; subst; clear H6.
       inversion H8; subst; clear H8.
       inversion H9; subst; clear H9.
       inversion H6; subst; clear H6.
       rewrite <- (IHHRS2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _) _ H0). clear IHHRS2 HRS2.
       repeat rewrite supd_Diff; try discriminate.
         destruct r0 as [[u B] Q].
         rewrite (HC _ _ _ _ _ _ H4); clear HC H4.
         rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H5 n0). trivial.
         right; assumption.
       intros Z. inversion Z; subst. apply (FRm H0).
    (*False*)
       inversion HeqTT; subst. trivial.
  clear HC.
    rewrite <- (IH _ _ _ _ _ _ H2). clear IH H2.
    repeat rewrite supd_Diff; try discriminate. trivial.
     intros Z. inversion Z; subst. apply FRm. apply IN.
    assumption. 
Qed.

Lemma xi_Mod: forall C F U
  (HmkT: U = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x
  (MVx: modVars (JErase _ U) x = true) n
  (IN: In n F),
  ssqsub (N (cond_svar n)) (N (var_svar (progVar x))).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion MVx.
(*Assign*)
  rewrite Bool.orb_false_r in MVx. apply pvar_eq_ElimT in MVx. subst.
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  rewrite supd_Same.
  rewrite supd_Diff; try discriminate.
    apply lubright. 
    apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. apply IN.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  apply Bool.orb_true_iff in MVx.
  remember (modVars (JErase _ (xi C2 F)) x) as b2; destruct b2; subst; simpl in *; apply eq_sym in Heqb2. 
  (*modVars SVar R2 x = true*)
     eapply (IHC2 F _ (eq_refl _)). apply H4. apply Heqb2. apply IN. 
  (*modVars SVar R2 x = false*)
     rewrite <- (xi_noMod C2 F _ (eq_refl _) _ _ _ _ _ _ H4 _ Heqb2).
     rewrite <- (xi_ControlLabelsPreserved C2 F _ (eq_refl _) _ _ _ _ _ _ H4 _ IN).
     clear H4 IHC2.
     destruct MVx as [MV | MV]; try inversion MV.
     eapply (IHC1 F _ (eq_refl _)). apply H2. apply MV. apply IN.
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  simpl in MVx.
  repeat rewrite modVars_mkJoins in MVx. 
  repeat rewrite Bool.orb_false_r in MVx.
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5; simpl in *.
  (*True*) clear IHC2.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
       destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.

       remember (in_dec pvar_eq x (modVarsL C2)) as b; destruct b; simpl.
          remember (in_dec pvar_eq x (modVarsL C1)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modvars_xi. apply modVarsChar. assumption.
                               right; assumption.
                            apply lub_leqright.

         assert (Z:= xi_ControlLabelsPreserved C1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2).
           rewrite <- (Z n). rewrite supd_Diff. Focus 2. intros ZZ. inversion ZZ. subst. apply (FRm IN).
         assert (Y : In m (m::F)). left. trivial.
           assert (ZZ := Z _ Y). clear Z Y. rewrite supd_Same in ZZ.
           rewrite supd_Diff; try discriminate.
           rewrite <- zeval_app_pc_supd in ZZ.
           eapply (@leq_trans L LatL _ (zeval SVar M (allPCs F ++ mkSVars e))).
             apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. assumption.
           rewrite <- ZZ. apply lubleft. apply lub_leqright.
        right. assumption.

       remember (in_dec pvar_eq x (modVarsL C1)) as bb; destruct bb; simpl.
         eapply leq_trans. eapply (IHC1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modvars_xi. apply modVarsChar. assumption.
                               right; assumption.
                            apply leq_refl.
         remember (modVars C1 x) as cc; destruct cc; apply eq_sym in Heqcc. rewrite modVarsChar in Heqcc. exfalso. apply (n1 Heqcc).          
         remember (modVars C2 x) as dd; destruct dd; apply eq_sym in Heqdd. rewrite modVarsChar in Heqdd. exfalso. apply (n0 Heqdd).
         repeat rewrite modvars_xi in MVx. rewrite Heqcc in MVx. rewrite Heqdd in MVx. simpl in MVx. inversion MVx.

   (*False*) clear IHC1.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
       destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.

       remember (in_dec pvar_eq x (modVarsL C1)) as b; destruct b; simpl.
          remember (in_dec pvar_eq x (modVarsL C2)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modvars_xi. apply modVarsChar. assumption.
                               right; assumption.
                            apply lub_leqright.


         assert (Z:= xi_ControlLabelsPreserved C2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2).
         rewrite <- (Z n). rewrite supd_Diff. Focus 2. intros ZZ. inversion ZZ. subst. apply (FRm IN).
         assert (Y : In m (m::F)). left. trivial.
           assert (ZZ := Z _ Y). clear Z Y. rewrite supd_Same in ZZ.
           rewrite supd_Diff; try discriminate.
           rewrite <- zeval_app_pc_supd in ZZ.
           eapply (@leq_trans L LatL _ (zeval SVar M (allPCs F ++ mkSVars e))).
             apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. assumption.
           rewrite <- ZZ. apply lubleft. apply lub_leqright.
        right. assumption.

       remember (in_dec pvar_eq x (modVarsL C2)) as bb; destruct bb; simpl.
          eapply leq_trans. eapply (IHC2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H2 x). simpl. 
                               rewrite modvars_xi. apply modVarsChar. assumption.
                               right; assumption.
                            apply leq_refl.
         remember (modVars C2 x) as cc; destruct cc; apply eq_sym in Heqcc. rewrite modVarsChar in Heqcc. exfalso. apply (n1 Heqcc).          
         remember (modVars C1 x) as dd; destruct dd; apply eq_sym in Heqdd. rewrite modVarsChar in Heqdd. exfalso. apply (n0 Heqdd).
         repeat rewrite modvars_xi in MVx. rewrite Heqcc in MVx. rewrite Heqdd in MVx. simpl in MVx. inversion MVx.
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  simpl in MVx. repeat rewrite modvars_xi in MVx.
  repeat rewrite modVars_mkJoins in MVx. 
  repeat rewrite Bool.orb_false_l in MVx. 
  repeat rewrite Bool.orb_false_r in MVx. 
  assert (IH := IHC (m::F) _ (eq_refl _)). clear IHC.
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4. 
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r0 as [[s G] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
  remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl in *.
    Focus 2. apply modVarsChar in MVx. exfalso. apply (n0 MVx).
  apply lubleft.
  clear IH.
  assert (Z : forall ST TT,
         JoinSem _ SVar_eq ST (JWhile SVar e
          (JComp SVar (xi C (m :: F))
             (JComp SVar (mkJoins m (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar m) (e :: nil)
                      (cond_svar m :: allPCs F ++ mkSVars e)))))) TT ->
           ((snd ST) (cond_svar n) = (snd TT) (cond_svar n) /\
            ssqsub ((snd ST) (cond_svar m)) ((snd TT) (cond_svar m)))).
    clear H2. clear Heqb i r E P HeqFRESH MVx x s G M.
    intros ST TT HRS. remember 
           (JWhile SVar e
     (JComp SVar (xi C (m :: F))
        (JComp SVar (mkJoins m (modVarsL C))
           (JComp SVar
              (JJoin SVar (var_svar pcVar) (e :: nil)
                 (var_svar pcVar :: mkSVars e))
              (JJoin SVar (cond_svar m) (e :: nil)
                 (cond_svar m :: allPCs F ++ mkSVars e)))))) as PP.
    induction HRS; subst; try inversion HeqPP; subst; clear HeqPP; simpl in *.
     (*True*) clear IHHRS1.
              destruct (IHHRS2 (eq_refl _)) as [Z1 Z2]. clear IHHRS2 HRS2.
              destruct s as [[s G] M]. destruct r as [[r D] N].
              inversion HRS1; subst; clear HRS1. destruct r0 as [[u E] P]. simpl in *.
              inversion H5; subst; clear H5.
              inversion H7; subst; clear H7.
              inversion H5; subst; clear H5.
              inversion H8; subst; clear H8.
              simpl in *. rewrite supd_Same in Z2.
              rewrite supd_Diff in Z2; try discriminate.
              rewrite <- zeval_app_pc_supd in Z2.
              rewrite supd_Diff in Z1.
                rewrite supd_Diff in Z1; try discriminate.
                assert (Z:= xi_ControlLabelsPreserved C (m::F) _ (eq_refl _) _ _ _ _ _ _ H3). clear H3.
                assert (ZZ:= mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4). clear H4.
                rewrite (Z n). rewrite <- Z1. split. apply ZZ.
                   rewrite (Z m). eapply leq_snart. apply Z2. apply lubright. rewrite (ZZ m). apply lub_leqleft. left. trivial.
                right; assumption.
              intros Z. inversion Z; subst. apply (FRm IN).
      (*False*) split. reflexivity. apply leq_refl.
   destruct (Z _ _ H2) as [Z1 Z2]; clear Z H2. simpl in *.
   rewrite supd_Same in Z2. 
   rewrite <- zeval_app_pc_supd in Z2.
   rewrite supd_Diff in Z1.
     rewrite supd_Diff in Z1; try discriminate.
     eapply leq_snart. apply Z2. apply lubright. rewrite <- Z1. apply zeval_IN_svar. apply in_or_app. left. apply allPCs_In. trivial.        
   intros Z. inversion Z; subst. apply (FRm IN).
Qed.

Lemma xi_Mod': forall C F U
  (HmkT: U = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N)) x
  (MVx: modVars (JErase _ U) x = true) n
  (IN: In n F),
  ssqsub (M (cond_svar n)) (N (var_svar (progVar x))).
Proof.
intros. subst.
rewrite (xi_ControlLabelsPreserved C F _ (eq_refl _) _ _ _ _ _ _ HRSem).
  eapply (xi_Mod C F _ (eq_refl _)). apply HRSem. assumption. assumption. 
assumption.
Qed.

Lemma xi_Sem: forall C F R
  (HmkT: R = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) R (t,D,N)) s' t'
  (HSem: Sem s' C t') x,
  (forall y, ssqsub (llub (G y) (M (var_svar (progVar y))))
                    (llub (D x) (N (var_svar (progVar x)))) 
             -> s y = s' y) ->
  t x = t' x.
Proof.
intros C. 
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem.
  inversion HSem; subst; clear HSem.
  apply H. apply leq_refl.
(*Assign*) 
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  inversion HSem; subst; clear HSem.
  apply M_upd_split_at; intros; subst.
     rewrite M_upd_Same. rewrite M_upd_Same in H. rewrite supd_Same in H.
     eapply (eq_below_evalE _ (fun z => llub (G z) (P (var_svar (progVar z)))))
          with (kappa:=(GEval (fun z : PVar => llub (G z) (P (var_svar (progVar z)))) e)).
       unfold eq_below. intros z Hz. apply H.
         eapply leq_trans. apply Hz. clear Hz H.
         rewrite GEval_sep.
         apply lub_leq. apply lub_leqleft.
         apply lubright. apply lubright. rewrite zeval_app.   
         rewrite zeval_GEval. apply lub_leqright.
       apply leq_refl.
   rewrite M_upd_Diff; try assumption.
     rewrite M_upd_Diff in H; try assumption.
     rewrite supd_Diff in H. apply H; clear H. apply leq_refl.
      intros Z; inversion Z; subst. apply H0; trivial.
(*Comp*)
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion HSem; subst; clear HSem.
  rename r0 into r'.
  eapply (IHC2 F _ (eq_refl _)); clear IHC2. apply H5. apply H7.
  intros y Hy.
  eapply (IHC1 F  _ (eq_refl _)); clear IHC1. apply H3. apply H4.
  intros z Hz.
  apply H; clear H.
  eapply leq_trans. apply Hz. apply Hy.
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5. destruct r0 as [[s G] P].
  inversion H3; subst; clear H3.
  inversion H6; subst; clear H6; simpl in *.
  (*eval s e = True*) clear IHC2.
     inversion H7; subst; clear H7. destruct r as [[r E] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]; clear H6.
     subst. repeat rewrite supd_Same in H.
     inversion HSem; subst; clear HSem.
    (*eval s' e = True*)
        eapply (IHC1 (n::F) _ (eq_refl _)). apply H3. apply H8.
        intros y Hy. apply H; clear H.
        rewrite supd_Diff in Hy; try discriminate.
        rewrite supd_Diff in Hy; try discriminate.
        eapply leq_trans. apply Hy. clear Hy.
        apply lub_leq. apply lub_leqleft.
        apply lubright. remember (in_dec pvar_eq x (modVarsL C2)) as b; destruct b; subst; simpl in *.
           apply lub_leqright.
           apply leq_refl.
     (*eval s' e = False*)
        clear IHC1.
        remember (in_dec pvar_eq x (modVarsL C1)) as b; destruct b; subst; simpl in *.
        (*In x modVars C1*) 
             assert (MVx : modVars (JErase _ (xi C1 (n :: F))) x = true).
               rewrite modvars_xi. apply modVarsChar. assumption.
             assert (Z:= xi_Mod' C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx n).
             rewrite supd_Same in Z.
             rewrite <- zeval_app_pc_supd in Z.
             assert (EV: eval s e = eval s' e).
             eapply (eq_below_evalE _ (fun z => llub (G z) (M (var_svar (progVar z)))))
                 with (kappa:=(GEval (fun z : PVar => llub (G z) (M (var_svar (progVar z)))) e)).
                 intros z Hz. apply H. clear H. 
                 eapply leq_trans. apply Hz. clear Hz.
                 apply lubright.
                 apply @leq_snart with (y:= P (var_svar (progVar x))).
                   remember (in_dec pvar_eq x (modVarsL C2)) as bb; destruct bb; subst; simpl in *.
                      apply lub_leqright.
                      apply leq_refl.
                 eapply leq_snart. apply Z. left. trivial.
                   clear Z. rewrite GEval_sep. 
                   apply lub_leq. apply lubleft. rewrite lab_eval_GEval. apply leq_refl.
                   apply lubright. rewrite zeval_app. apply lubright. rewrite zeval_GEval. apply leq_refl.
               apply leq_refl.
             rewrite EV in H5. rewrite H5 in H7. inversion H7.
         (*~ In x (MV C1*)
           assert (MVx :modVars (JErase _ (xi C1 (n :: F))) x = false).
             rewrite modvars_xi. apply modVarsChar'. assumption.
           destruct (JSem1 _ _ _ _ _ _ _ _ _ H3 _ MVx) as [Z1 Z2]. 
           rewrite <- Z1 in *. clear Z1.
           rewrite <- Z2 in *. clear Z2. 
           remember (modVars C2 x) as bb; destruct bb; subst; simpl in *; apply eq_sym in Heqbb.
             remember (in_dec pvar_eq x (modVarsL C2)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
               assert (EV: eval s e = eval s' e).                      
                 eapply (eq_below_evalE _ (fun z => llub (G z) (M (var_svar (progVar z)))))
                     with (kappa:=(GEval (fun z : PVar => llub (G z) (M (var_svar (progVar z)))) e)).
                   intros z Hz. apply H. clear H. 
                   eapply leq_trans. apply Hz. clear Hz.
                   assert (Z:= xi_ControlLabelsPreserved C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 n). clear H3.
                   rewrite supd_Same in Z. 
                   rewrite <- zeval_app_pc_supd in Z.
                   apply lubright. apply lubleft.
                   rewrite <- Z.
                     rewrite lab_eval_GEval. 
                     rewrite GEval_sep. apply lub_leq. apply lub_leqleft.
                       rewrite zeval_app. rewrite <- zeval_GEval. 
                       apply lubright. apply lub_leqright.
                   left; trivial. 
                 apply leq_refl.
               rewrite EV in H5. rewrite H5 in H7. inversion H7.
             exfalso. apply n1. apply modVarsChar. apply Heqbb.
           rewrite <- (Sem_nomod _ _ _ H8 _ Heqbb). 
             apply H. clear H H8.  
             remember (in_dec pvar_eq x (modVarsL C2)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
                exfalso. clear Heqbbb. apply modVarsChar in i. rewrite i in Heqbb. inversion Heqbb. 
             assert (Z:= xi_noMod C1 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx).
                    rewrite supd_Diff in Z; try discriminate.
                    rewrite supd_Diff in Z; try discriminate. rewrite Z. apply leq_refl. 
  (*eval s e = False*) clear IHC1.
     inversion H7; subst; clear H7. destruct r as [[r E] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]; clear H6.
     subst. repeat rewrite supd_Same in H.
     inversion HSem; subst; clear HSem.
    (*eval s' e = True*)
        clear IHC2.
        remember (in_dec pvar_eq x (modVarsL C2)) as b; destruct b; subst; simpl in *.
        (*In x modVars C2*) 
             assert (MVx :modVars (JErase _ (xi C2 (n :: F))) x = true).
               rewrite modvars_xi. apply modVarsChar. assumption.
             assert (Z:= xi_Mod' C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx n).
             rewrite supd_Same in Z. 
             rewrite <- zeval_app_pc_supd in Z.
             assert (EV: eval s e = eval s' e).
             eapply (eq_below_evalE _ (fun z => llub (G z) (M (var_svar (progVar z)))))
                 with (kappa:=(GEval (fun z : PVar => llub (G z) (M (var_svar (progVar z)))) e)).
                 intros z Hz. apply H. clear H. 
                 eapply leq_trans. apply Hz. clear Hz.
                 apply lubright.
                 apply @leq_snart with (y:= P (var_svar (progVar x))).
                   remember (in_dec pvar_eq x (modVarsL C1)) as bb; destruct bb; subst; simpl in *.
                      apply lub_leqright.
                      apply leq_refl.
                 eapply leq_snart. apply Z. left. trivial. 
                     rewrite lab_eval_GEval. 
                     rewrite GEval_sep. apply lub_leq. apply lub_leqleft.
                       rewrite zeval_app. rewrite <- zeval_GEval. 
                       apply lubright. apply lub_leqright. 
               apply leq_refl.
             rewrite EV in H5. rewrite H5 in H7. inversion H7.
         (*~ In x (MV C2*)
           assert (MVx :modVars (JErase _ (xi C2 (n :: F))) x = false).
             rewrite modvars_xi. apply modVarsChar'. assumption.
           destruct (JSem1 _ _ _ _ _ _ _ _ _ H3 _ MVx) as [Z1 Z2]. 
           rewrite <- Z1 in *. clear Z1.
           rewrite <- Z2 in *. clear Z2. 
           remember (modVars C1 x) as bb; destruct bb; subst; simpl in *; apply eq_sym in Heqbb.
             remember (in_dec pvar_eq x (modVarsL C1)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
               assert (EV: eval s e = eval s' e).                      
                 eapply (eq_below_evalE _ (fun z => llub (G z) (M (var_svar (progVar z)))))
                     with (kappa:=(GEval (fun z : PVar => llub (G z) (M (var_svar (progVar z)))) e)).
                   intros z Hz. apply H. clear H. 
                   eapply leq_trans. apply Hz. clear Hz.
                   assert (Z:= xi_ControlLabelsPreserved C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 n). clear H3.
                   rewrite supd_Same in Z.
                   rewrite <- zeval_app_pc_supd in Z.
                   apply lubright. apply lubleft.
                   rewrite <- Z. 
                     rewrite lab_eval_GEval.
                     rewrite GEval_sep. apply lub_leq. apply lub_leqleft.
                     apply lubright. rewrite zeval_app. rewrite <- zeval_GEval. 
                        apply lub_leqright. 
                   left; trivial. 
                 apply leq_refl.
               rewrite EV in H5. rewrite H5 in H7. inversion H7.
             exfalso. apply n1. apply modVarsChar. apply Heqbb.
           rewrite <- (Sem_nomod _ _ _ H8 _ Heqbb). 
             apply H. clear H H8.  
             remember (in_dec pvar_eq x (modVarsL C1)) as bbb; destruct bbb; subst; simpl in *; apply eq_sym in Heqbbb.
                exfalso. clear Heqbbb. apply modVarsChar in i. rewrite i in Heqbb. inversion Heqbb. 
             assert (Z:= xi_noMod C2 (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 _ MVx).
                    rewrite supd_Diff in Z; try discriminate.
                    rewrite supd_Diff in Z; try discriminate. rewrite Z. apply leq_refl. 
     (*eval s' e = False*)
        eapply (IHC2 (n::F) _ (eq_refl _)). apply H3. apply H8.
        intros y Hy. apply H; clear H.
        rewrite supd_Diff in Hy; try discriminate.
        rewrite supd_Diff in Hy; try discriminate.
        eapply leq_trans. apply Hy. clear Hy.
        apply lub_leq. apply lub_leqleft.
        apply lubright. remember (in_dec pvar_eq x (modVarsL C1)) as b; destruct b; subst; simpl in *.
           apply lub_leqright.
           apply leq_refl.
(*While*)
  remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl in *.
Focus 2. 
(*~In x (modVarsL C)*)
  clear IHC.
  assert (Z:= xi_noMod (While e C) F _ (eq_refl _)). simpl in Z.
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  assert (ZZ:= Z _ _ _ _ _ _ HRSem x). clear Z.
  simpl in ZZ. repeat rewrite modVars_mkJoins in ZZ. simpl in ZZ. repeat rewrite Bool.orb_false_r in ZZ. rewrite modvars_xi in ZZ.
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  inversion H3; subst; clear H3.  
  inversion H6; subst; clear H6. destruct r0 as [[s G] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5. 
  apply eq_sym in Z1; apply eq_sym in Z2; subst.
  clear Heqb. 
  assert (MV: modVars C x = false). apply modVarsChar'. apply n. 
  assert (ZZZ := ZZ MV). clear ZZ. rename ZZZ into ZZ.  
  destruct (JSem1 _ _ _ _ _ _ _ _ _ H3 x) as [X1 X2].
     simpl. rewrite modvars_xi. rewrite modVars_mkJoins. simpl. rewrite Bool.orb_false_r. apply MV.
   rewrite <- X1. clear X1.
  rewrite <- (Sem_nomod _ _ _ HSem).
     clear H3. apply H. rewrite X2. 
     remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl.
       exfalso. apply (n i).
       rewrite ZZ. apply leq_refl.
  apply MV.
(*In x (modVarsL C)*)
  remember (fresh F) as FRESH. destruct FRESH as [n FRn].
  inversion HRSem; subst; clear HRSem.
  destruct r as [[r E] P].
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  inversion H3; subst; clear H3.
  inversion H6; subst; clear H6. destruct r0 as [[s G] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5.
  apply eq_sym in Z1; apply eq_sym in Z2; subst.
  repeat rewrite <- Heqb in *.
  assert (IH := IHC (n::F) _ (eq_refl _)); clear IHC.

  assert (SUB: forall s G M t E P 
              (HR: JoinSem _ SVar_eq (s, G, M)
                       (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e))))))
                       (t, E, P)),
                   ssqsub (M (cond_svar n)) (P (cond_svar n))).
  (*Proof of SUB*)
    clear H3 H HSem r E M t E P s' t'. intros.
      remember (s, G, M) as ST.
      remember (t, E, P) as TT.
      remember (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e)))))) as PP.
      generalize dependent P. generalize dependent E. generalize dependent t.
      generalize dependent M. generalize dependent G. generalize dependent s.
      induction HR; intros; subst; try inversion HeqPP; subst; clear HeqPP.
      (*True*) clear IHHR1 HR2 IH; simpl in *.
               inversion HR1; subst; clear HR1.
               inversion H5; subst; clear H5.
               inversion H7; subst; clear H7.
               inversion H5; subst; clear H5.
               inversion H8; subst; clear H8.
               assert (IHH := IHHR2 (eq_refl _) _ _ _  (eq_refl _) _ _ _ (eq_refl _)). clear IHHR2.
               rewrite supd_Same in IHH. simpl in IHH. rewrite <- zeval_app_pc_supd in IHH. rewrite supd_Diff in IHH; try discriminate.
               eapply leq_snart. apply IHH.
               clear IHH.
               destruct r0 as [[u B] O].
               rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H4).
               rewrite (xi_ControlLabelsPreserved C (n::F) _ (eq_refl _) _ _ _ _ _ _ H3 n).
               apply lubright. apply lub_leqleft.
               left; trivial.
       (*False*) inversion HeqTT; subst; clear HeqTT.
                 apply leq_refl.
  (*Use of SUB*)
  remember (JWhile SVar e
           (JComp SVar (xi C (n :: F))
              (JComp SVar (mkJoins n (modVarsL C))
                 (JComp SVar
                    (JJoin SVar (var_svar pcVar) (e :: nil)
                       (var_svar pcVar :: mkSVars e))
                    (JJoin SVar (cond_svar n) (e :: nil)
                       (cond_svar n :: allPCs F ++ mkSVars e)))))) as PP.
  cut (forall s GG MM t D N
     (HR: JoinSem _ SVar_eq (s, GG, MM) PP (t, D, N))
     s' t' 
     (HS: Sem s' (While e C) t')
     (HM: ssqsub (llub (lab_eval GG (e :: nil)) (zeval SVar MM (allPCs F ++ mkSVars e)))
                 (MM (cond_svar n))),
     (forall y : PVar,
      ssqsub (llub (GG y) (MM (var_svar (progVar y))))
        (llub (D x) (llub (N (cond_svar n)) (N (var_svar (progVar x))))) -> s y = s' y) -> t x = t' x).
  (*Use OF CUT*)
     intros CUT.
     eapply CUT; clear CUT.
       apply H3.
       apply HSem.
       rewrite supd_Same.
       rewrite zeval_all_app.
         repeat rewrite <- zeval_app_pc_supd. apply leq_refl. 
       intros Z. apply (FRn Z). 
     intros y Hy. 
       rewrite supd_Diff in Hy; try discriminate.
       rewrite supd_Diff in Hy; try discriminate.
       apply H. apply Hy.
  (*Proof OF CUT*)
   clear H3 HSem H r s' t t'.
   intros ss GG MM tt DD NN HR.
   remember (ss, GG,MM) as ST. remember (tt, DD,NN) as TT.
   generalize dependent NN. generalize dependent DD. generalize dependent tt.
   generalize dependent MM. generalize dependent GG. generalize dependent ss.
   induction HR; intros; subst; try inversion HeqPP; subst; clear HeqPP.
   (*OUTER True*) clear IHHR1; simpl in *. destruct r as [[r B] Q].
       assert (IHH := IHHR2 (eq_refl _)); clear IHHR2.
       assert (SUB_IH := SUB _ _ _ _ _ _ HR2). clear HR2.
       inversion HR1; subst; clear HR1. destruct r0 as [[u A] K].
       inversion H6; subst; clear H6.
       inversion H8; subst; clear H8.
       inversion H9; subst; clear H9.
       rewrite supd_Same in *.
       inversion H6; subst; clear H6.
       simpl in *. rewrite supd_Diff in SUB_IH; try discriminate. rewrite <- zeval_app_pc_supd in SUB_IH.

       assert (EV: eval ss e = eval s' e). clear IHH SUB IH HS.
         eapply eq_below_evalE with (G:= (fun z => llub (GG z) (MM (var_svar (progVar z)))))
              (kappa:=(GEval (fun z : PVar => llub (GG z) (MM (var_svar (progVar z)))) e)).
           intros y Hy. apply H0; clear H0.
             apply (leq_trans Hy); clear Hy.
             apply lubright. apply lubleft. 
              eapply leq_snart. apply SUB_IH.
              clear SUB_IH.
              rewrite (xi_ControlLabelsPreserved C (n::F) _ (eq_refl _) _ _ _ _ _ _ H4 n) in HM;  clear H4.
              rewrite <- (mkJoins_ControlLabelsPreserved _ _ _ _ _ _ _ _ H5 n). clear H5.
              rewrite GEval_sep. rewrite lab_eval_GEval in HM. 
              apply lubright. apply lubleft. 
              eapply leq_snart. apply HM. rewrite zeval_app. 
                           apply lub_leq. apply lub_leqleft.
                             apply lubright. rewrite zeval_GEval. apply lub_leqright.
             left; trivial.
           apply leq_refl.
       inversion HS; subst; clear HS SUB_IH HM.
       (*INNER True*) rename r into u'.
             apply (IHH SUB _ _ _ (eq_refl _) _ _ _ (eq_refl _) _ _ H9); clear IHH SUB H9.
                rewrite supd_Same. 
                rewrite zeval_all_app. 
                  repeat rewrite <- zeval_app_pc_supd.
                  repeat rewrite supd_Diff; try discriminate.
                  apply lub_leq. apply lub_leqleft.
                    apply lubright. apply lub_leqright.
                intros Z. apply (FRn Z).
             intros y Hy. 
               rewrite supd_Diff in Hy; try discriminate.
               rewrite supd_Diff in Hy; try discriminate.
               destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. 
               simpl in *. 
               eapply IH; clear IH. apply H4. apply H7.
                 intros z Hz. apply H0; clear H0.
                   apply (leq_trans Hz); clear Hz. 
                   eapply leq_snart. apply Hy.
                    apply lub_leq. apply lub_leqleft.
                    apply lubright. remember (in_dec pvar_eq y (modVarsL C)) as bb; destruct bb; subst; simpl.
                        apply lub_leqright. apply leq_refl.
       (*INNER False*) clear IHH SUB IH H0 H4.
              rewrite EV in H. rewrite H in H6. inversion H6.
   (*OUTER False*) clear SUB IH.
       apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT.
       simpl in *.
       assert (EV: eval ss e = eval s' e).
         apply eq_below_evalE with (G:= (fun z => llub (GG z) (MM (var_svar (progVar z)))))
                      (kappa:=(GEval (fun z : PVar => llub (GG z) (MM (var_svar (progVar z)))) e)).
           intros y Hy. apply H1; clear H1.
             apply (leq_trans Hy); clear Hy.
             apply lubright. apply lubleft.
             eapply leq_snart. apply HM.
             rewrite GEval_sep.
             apply lub_leq. rewrite lab_eval_GEval. apply lub_leqleft.
                   apply lubright. rewrite zeval_app. rewrite zeval_GEval. apply lub_leqright.
           apply leq_refl.
       inversion HS; subst; clear HS HM.
       (*INNER True*) rewrite EV in H. rewrite H in H3. inversion H3.
       (*INNER False*) apply H1; clear H1. 
         apply lub_leq. apply lub_leqleft.
         apply lubright. apply lub_leqright.
Qed.


(*Let's again fold the definition of eq_below in mkUT_Sem*)
Goal forall C F s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N)) s' t'
  (HSem: Sem s' C t') x,
  eq_below (MERGE D N x) (MERGE G M) s s' ->
  t x = t' x.
Proof.
intros.
eapply (xi_Sem C F _ (eq_refl _)).
  apply HRSem. apply HSem. apply H.
Qed.

Lemma xi_TermLabelsIncreases: forall C F R 
  (HR: R = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) R (t,D,N)),
  ssqsub (M (var_svar pcVar)) (N (var_svar pcVar)).
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem.
  apply leq_refl.
(*Assign*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2. 
  inversion H4; subst; clear H4. 
  rewrite supd_Diff; try discriminate. apply leq_refl.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  eapply leq_trans.
    apply (IHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2).
  apply (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H4).
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5.
  (*True*)
     clear IHC2 H4.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
       eapply leq_snart. apply (IHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2).
         clear IHC1 H2.
         rewrite supd_Diff; try discriminate.
           rewrite supd_Same. apply lubright. apply lub_leqleft.
  (*False*)
     clear IHC1 H4.
     inversion H6; subst; clear H6. destruct r0 as [[s G] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
       eapply leq_snart. apply (IHC2 _ _ (eq_refl _) _ _ _ _ _ _ H2).
         clear IHC2 H2.
         rewrite supd_Diff; try discriminate.
           rewrite supd_Same. apply lubright. apply lub_leqleft.
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r E] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r0 as [[s G] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
  assert (HC := IHC (m::F) _ (eq_refl _)). clear IHC.
 
  remember (JWhile SVar e
          (JComp SVar (xi C (m :: F))
             (JComp SVar (mkJoins m (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar m) (e :: nil)
                      (cond_svar m :: allPCs F ++ mkSVars e)))))) as PP.

  assert (IH: forall s G M t D N,
     JoinSem _ SVar_eq (s, G, M) PP (t, D, N) ->
     ssqsub (M (var_svar pcVar)) (N (var_svar pcVar))).
    clear s G M r E P H2. intros.
    remember (s, G, M) as ST. remember (t,D,N) as TT.
    generalize dependent N. generalize dependent D. generalize dependent t.
    generalize dependent M. generalize dependent G. generalize dependent s.
    induction H; intros; subst; try inversion HeqPP; subst; clear HeqPP.
    (*True*)
       simpl in *. clear H1 IHJoinSem1. destruct r as [[r A] Q].
       eapply leq_snart. eapply IHJoinSem2; try reflexivity. clear IHJoinSem2. 
       inversion H0; subst; clear H0.
       inversion H6; subst; clear H6.
       inversion H7; subst; clear H7.
       inversion H8; subst; clear H8.
       inversion H5; subst; clear H5. destruct r0 as [[t E] P].
       destruct (mkJoins_effect _ _ _ _ _ _ _ _ H3) as [Z1 [Z2 Z3]]. subst. clear H3.
       rewrite supd_Diff; try discriminate.
       rewrite supd_Same. 
       eapply leq_trans. apply (HC _ _ _ _ _ _ H4). clear HC H4.
       apply lubright. simpl. apply lub_leqleft.
    (*False*)
       inversion HeqTT; subst. apply leq_refl.
(*Use of loop assert *)
  eapply leq_snart. apply (IH _ _ _ _ _ _ H2). clear H2 IH HC.
  rewrite supd_Diff; try discriminate.
  rewrite supd_Same.
    apply lubright. apply lub_leqleft. 
Qed.

Lemma Theorem6: forall C F R
  (HmkT: R = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) R (t,D,N)) s',
  (forall y, ssqsub (llub (G y) (M (var_svar (progVar y)))) (N (var_svar pcVar))
             -> s y = s' y) ->
  exists t', 
   (Sem s' C t' /\ 
    forall x, 
      (forall y, ssqsub (llub (G y) (M (var_svar (progVar y))))
                    (llub (D x) (N (var_svar (progVar x)))) 
                 -> s y = s' y) ->
      t x = t' x).
Proof.
intros C. 
induction C; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst.
  assert (EE: exists t', Sem s' Skip t' ).
    eexists. constructor. reflexivity.
  destruct EE as [t' Ht]. 
  exists t'; split; trivial.
  apply (xi_Sem Skip F _ (eq_refl _) _ _ _ _ _ _ HRSem _ _ Ht).
(*Assign*)
  inversion HRSem; subst.
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  assert (EE: exists t', Sem s' (Assign p e) t' ).
    eexists. constructor. reflexivity.
  destruct EE as [t' Ht]. 
  exists t'; split; trivial.
  apply (xi_Sem (Assign p e) F _ (eq_refl _) _ _ _ _ _ _ HRSem _ _ Ht).
(*Comp*)
  inversion HRSem; subst.
  destruct r as [[r E] P].
  assert (IH1:= IHC1 F _ (eq_refl _) _ _ _ _ _ _ H3). clear IHC1.
  assert (IH2:= IHC2 F _ (eq_refl _) _ _ _ _ _ _ H5). clear IHC2.
  assert (E1: exists r', Sem s' C1 r'). clear IH2.
     destruct (IH1 s') as [r' [Hr1' HR2']]. 
       intros. apply H; clear H. eapply (leq_trans H0).  eapply (xi_TermLabelsIncreases C2 _ _ (eq_refl _)). apply H5.
     exists r'; assumption.
  destruct E1 as [r' Hr'].
  assert (E2: exists t', Sem r' C2 t').
     destruct (IH2 r') as [t' [Ht1' HRt']]. clear IH2.
       intros. eapply (xi_Sem C1 F). reflexivity. apply H3. apply Hr'.
                 intros z Hz. apply H; clear H. eapply (leq_trans Hz). apply H0.
     exists t'; assumption.
  destruct E2 as [t' Ht'].
  exists t'.
  split. econstructor. apply Hr'. apply Ht'.
  eapply (xi_Sem (Comp C1 C2) F). reflexivity. apply HRSem.
     econstructor. apply Hr'. apply Ht'.
(*Ite*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst.
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  inversion H3; subst; clear H3.
  inversion H6; subst; clear H6.
  (*True*) clear IHC2.
    inversion H7; subst; clear H7. destruct r as [[r A] P].
    destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
    assert (E: exists t', Sem s' C1 t').
      destruct (IHC1 (m::F) _ (eq_refl _) _ _ _ _ _ _ H3 s') as [r' [Hr1' HR2']]. 
        intros z Hz. apply H; clear H HRSem.
          rewrite supd_Diff in Hz; try discriminate. 
          rewrite supd_Diff in Hz; try discriminate.
          apply Hz. 
      exists r'; assumption.
    destruct E as [t' Ht'].
    assert (EV: eval s e = eval s' e).
      eapply eq_below_evalE with (G:=fun y => (llub (G y) (M (var_svar (progVar y))))).
      Focus 2. apply leq_refl.
      intros z Hz. apply H; clear H.
         eapply (leq_trans Hz); clear Hz.
         eapply leq_snart. eapply (xi_TermLabelsIncreases C1 (m::F) _ (eq_refl _)). apply H3.
                              rewrite supd_Diff; try discriminate. rewrite supd_Same. 
                              rewrite GEval_sep. apply lub_leq. rewrite lab_eval_GEval. apply lub_leqleft.
                              apply lubright. rewrite zeval_GEval. apply lub_leqright.
    assert (HSem: Sem s' (Ite e C1 C2) t').
      eapply S_True. rewrite <- EV. apply H5. 
                     assumption.
    exists t'. split; trivial.
    assert (TT:= xi_Sem (Ite e C1 C2) F _ (eq_refl _)).
    simpl in TT. rewrite <- HeqFRESH in TT.
    apply (TT _ _ _ _ _ _ HRSem _ _ HSem).
  (*False*) clear IHC1.
    inversion H7; subst; clear H7. destruct r as [[r A] P].
    destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
    assert (E: exists t', Sem s' C2 t').
      destruct (IHC2 (m::F) _ (eq_refl _) _ _ _ _ _ _ H3 s') as [r' [Hr1' HR2']]. 
        intros z Hz. apply H; clear H HRSem.
          rewrite supd_Diff in Hz; try discriminate. 
          rewrite supd_Diff in Hz; try discriminate. 
          apply Hz.
      exists r'; assumption.
    destruct E as [t' Ht'].
    assert (EV: eval s e = eval s' e).
      eapply eq_below_evalE with (G:=fun y => (llub (G y) (M (var_svar (progVar y))))).
      Focus 2. apply leq_refl.
      intros z Hz. apply H; clear H.
         eapply (leq_trans Hz); clear Hz.
         eapply leq_snart. eapply (xi_TermLabelsIncreases C2 (m::F) _ (eq_refl _)). apply H3.
                              rewrite supd_Diff; try discriminate. rewrite supd_Same. 
                              rewrite GEval_sep. apply lub_leq. rewrite lab_eval_GEval. apply lub_leqleft.
                              apply lubright. rewrite zeval_GEval. apply lub_leqright.
    assert (HSem: Sem s' (Ite e C1 C2) t').
      eapply S_False. rewrite <- EV. apply H5. 
                     assumption.
    exists t'. split; trivial.
    assert (TT:= xi_Sem (Ite e C1 C2) F _ (eq_refl _)).
    simpl in TT. rewrite <- HeqFRESH in TT.
    apply (TT _ _ _ _ _ _ HRSem _ _ HSem).
(*While*)
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst.
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5.
  inversion H3; subst; clear H3.
  inversion H6; subst; clear H6. destruct r as [[r A] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
  assert (IH := IHC (m::F) _ (eq_refl _)). clear IHC.
  remember (JWhile SVar e
                      (JComp SVar (xi C (m :: F))
                         (JComp SVar (mkJoins m (modVarsL C))
                            (JComp SVar
                               (JJoin SVar (var_svar pcVar) (e :: nil)
                                  (var_svar pcVar :: mkSVars e))
                               (JJoin SVar (cond_svar m) (e :: nil)
                                  (cond_svar m :: allPCs F ++ mkSVars e)))))) as PP.
  assert (X: forall s G M t D N
     (HRS: JoinSem _ SVar_eq (s, G, M) PP (t, D, N))
     s'
     (HGM: ssqsub (llub (lab_eval G (e::nil)) (zeval SVar M (mkSVars e))) (M (var_svar pcVar))),
     (forall y : PVar,
      ssqsub (llub (G y) (M (var_svar (progVar y)))) (N (var_svar pcVar)) ->
      s y = s' y) ->
     exists t' : Store,
       Sem s' (While e C) t' /\
       (forall x : PVar,
        (forall y : PVar,
         ssqsub (llub (G y) (M (var_svar (progVar y))))
           (llub (D x) (N (var_svar (progVar x)))) -> s y = s' y) ->
        t x = t' x)).
      clear H3 H s'.
      intros ss GG MM t D N HRS.
      remember (ss, GG, MM) as ST. remember (t, D, N) as TT.
       generalize dependent N. generalize dependent D. generalize dependent t.
       generalize dependent MM. generalize dependent GG. generalize dependent ss.
       induction HRS; subst; try inversion HeqPP; subst; clear HeqPP.
       (*True*) clear IHHRS1. intros. subst.
          assert (IHH:= IHHRS2 (eq_refl _)); clear IHHRS2.
          destruct r0 as [[u B] Q].
          assert (IHHH := IHH HRSem _ _ _ (eq_refl _) _ _ _ (eq_refl _)); clear IHH.
          assert (ssqsub (Q (var_svar pcVar)) (N (var_svar pcVar))).
          (*Proof of ssqsub (Q treg) (N treg)*)
              clear IHHH HRS1 HRSem IH H0 H s G M r A P s' ss.
              remember ((u, B, Q)) as ST. remember (t0, D, N) as TT.
              remember (JWhile SVar e
            (JComp SVar (xi C (m :: F))
               (JComp SVar (mkJoins m (modVarsL C))
                  (JComp SVar
                     (JJoin SVar (var_svar pcVar) (e :: nil)
                        (var_svar pcVar :: mkSVars e))
                     (JJoin SVar (cond_svar m) (e :: nil)
                        (cond_svar m :: allPCs F ++ mkSVars e)))))) as PP.
             generalize dependent N. generalize dependent D. generalize dependent t0.
             generalize dependent Q. generalize dependent B. generalize dependent u.
             induction HRS2; subst; try inversion HeqPP; subst; clear HeqPP.
             (*True*) clear  HRS2_2 IHHRS2_1.
               intros. subst. inversion HRS2_1; subst; clear HRS2_1.
                              inversion H5; subst; clear H5.
                              inversion H7; subst; clear H7.
                              inversion H5; subst; clear H5.
                              inversion H8; subst; clear H8. destruct r0 as [[r K] P].
                              eapply leq_trans.
                                eapply (xi_TermLabelsIncreases C (m::F) _ (eq_refl _) _ _ _ _ _ _ H3); clear H3.
                              destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
                              eapply leq_snart.
                                apply (IHHRS2_2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)).
                                clear IHHRS2_2.
                                rewrite supd_Diff; try discriminate.
                                rewrite supd_Same. apply lubright. apply lub_leqleft. 
              (*False*) intros; subst. inversion HeqTT; subst; clear HeqTT. apply leq_refl.
          (*Use of ssqsub (Q treg) (N treg)*)
          assert (exists u', Sem s' C u'). clear IHHH HRS2. 
            inversion HRS1; subst; clear HRS1. destruct r0 as [[v Hv1] Hv2].
            destruct (IH _ _ _ _ _ _ H5 s') as [v' [Hv1' Hv2']]. clear H5.
               intros z Hz. apply H0; clear H0. 
                  apply (leq_trans Hz); clear Hz.
                  inversion H7; subst; clear H7.
                  inversion H6; subst; clear H6.
                  inversion H5; subst; clear H5.
                  inversion H8; subst; clear H8.
                  rewrite supd_Diff in H1; try discriminate.
                  rewrite supd_Same in H1.
                  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
                  eapply leq_snart. apply H1.
                  apply lubright. apply lub_leqleft. 
            exists v'. apply Hv1'.
          destruct H2 as [u' Hu'].
          inversion HRS1; subst; clear HRS1.
                  inversion H7; subst; clear H7.
                  inversion H9; subst; clear H9.
                  inversion H7; subst; clear H7.
                  inversion H10; subst; clear H10.
                  rewrite supd_Diff in H1; try discriminate.
                  rewrite supd_Same in H1. destruct r0 as [[uu Huu1] Huu2].
                  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
          assert (IHC := xi_Sem C (m::F) _ (eq_refl _) _ _ _ _ _ _ H5 _ _ Hu').
          clear IH.    
          assert (EV: eval ss e = eval s' e).
             clear IHHH IHC HRSem HRS2.
             eapply eq_below_evalE with (G:= fun z => llub (GG z) (MM (var_svar (progVar z)))). Focus 2. apply leq_refl.
                intros z Hz. apply H0; clear H0.
                  apply (leq_trans Hz); clear Hz.      
                  eapply leq_snart. apply H1. clear H1.
                  rewrite lab_eval_GEval in HGM. rewrite zeval_GEval in HGM.
                  rewrite GEval_sep. eapply (leq_trans HGM); clear HGM.
                  eapply leq_trans. eapply (xi_TermLabelsIncreases C (m::F) _ (eq_refl _)). apply H5.
                    apply lubright. apply lub_leqleft.  
          simpl in H.
          rewrite supd_Diff in IHHH; try discriminate.
          rewrite supd_Same in IHHH.
          destruct (IHHH u') as [t' [Ht1' Ht2']]; clear IHHH.
             clear IHC HRS2.
               apply lub_leq. apply lub_leqleft.
                rewrite <- zeval_zupd. rewrite <- zeval_pc_supd. apply lubright. apply lub_leqright.
             intros z Hz. rewrite supd_Diff in Hz; try discriminate. rewrite supd_Diff in Hz; try discriminate.
               apply IHC; clear IHC.
               intros y Hy. apply H0; clear H0.
                 eapply (leq_trans Hy); clear Hy.
                 eapply leq_snart. apply Hz. clear Hz.
                 apply lub_leq. apply lub_leqleft.
                   apply lubright. remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b; subst; simpl. apply lub_leqright. apply leq_refl.
          exists t'.
          assert (EX: Sem s' (While e C) t').
             eapply S_WTrue. clear Ht1' Ht2' IHC HRS2. rewrite EV in H. apply H.
             apply Hu'.
             apply Ht1'.
          split; try assumption.
          intros x Hx. apply Ht2'; clear Ht2'.
            intros u Hu. rewrite supd_Diff in Hu; try discriminate. rewrite supd_Diff in Hu; try discriminate.
              apply IHC. intros z Hz. apply Hx; clear Hx.
                eapply (leq_trans Hz); clear Hz.
                eapply leq_snart. apply Hu; clear Hu.
                apply lub_leq. apply lub_leqleft.
                  apply lubright. remember (in_dec pvar_eq u (modVarsL C)) as b; destruct b; subst; simpl. apply lub_leqright. apply leq_refl.
       (*False*) clear HRSem IH; intros; subst.
          inversion HeqTT; subst; clear HeqTT. simpl in H.
          assert (EV: eval t0 e = eval s' e).
             eapply eq_below_evalE with (G:= fun z => llub (D z) (N (var_svar (progVar z)))). Focus 2. apply leq_refl.
                intros z Hz. apply H0; clear H0.
                  apply (leq_trans Hz); clear Hz.   
                  rewrite lab_eval_GEval in HGM. rewrite zeval_GEval in HGM.
                  rewrite GEval_sep. apply HGM. 
          assert (EX: Sem s' (While e C) s').
             eapply S_WFalse. rewrite EV in H. apply H. 
             trivial.
          exists s'.
          split; try assumption.
          intros x Hx. apply Hx. apply leq_refl. 
  (*Use of X*) clear HRSem.
    assert (XX:= X _ _ _ _ _ _ H3). clear X H3.
    destruct (XX s') as [t' [Ht1' Ht2']]; clear XX IH.
      rewrite supd_Diff; try discriminate.
      rewrite supd_Same. 
        apply lub_leq. apply lub_leqleft.
          repeat rewrite <- zeval_zupd. rewrite <- zeval_pc_supd. apply lubright. apply lub_leqright.
      intros y Hy. apply H; clear H. rewrite supd_Diff in Hy; try discriminate. rewrite supd_Diff in Hy; try discriminate. apply Hy.
    exists t'.
    split; trivial.
    intros x Hx. apply Ht2'; clear Ht2'.
      intros y Hy. rewrite supd_Diff in Hy; try discriminate. rewrite supd_Diff in Hy; try discriminate. 
        apply Hx; clear Hx.
        eapply (leq_trans Hy); clear Hy. 
        apply lub_leq. apply lub_leqleft.
          apply lubright. remember (in_dec pvar_eq x (modVarsL C)) as b; destruct b; subst; simpl.
            apply lub_leqright. apply leq_refl.
Qed.

(*same result, but formulated using MERGE operator*)
Theorem Theorem6_MERGE: forall C F s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N)) s'
  (EB: eq_below (N (var_svar pcVar)) (MERGE G M) s s'),
  exists t', 
   (Sem s' C t' /\ 
    forall x, eq_below (MERGE D N x) (MERGE G M) s s' ->
      t x = t' x).
Proof.
intros.
eapply (Theorem6 C F _ (eq_refl _)).
  apply HRSem.
  apply EB.
Qed.

(*The weaker claim alluded to in the text presceding Theorem6 in the paper, 
  (indistinghuishability below (MERGE G M x) lub (N z^pcVar))*)
Theorem Theorem6_MERGE_pcVar: forall C F s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N)) s'
  (EB: eq_below (N (var_svar pcVar)) (MERGE G M) s s'),
  exists t', 
   (Sem s' C t' /\ 
    forall x, eq_below (llub (MERGE D N x) (N (var_svar pcVar))) (MERGE G M) s s' ->
      t x = t' x).
Proof.
intros.
destruct (Theorem6_MERGE _ _ _ _ _ _ _ _ HRSem _ EB) as [t' [HSem' Hind]].
exists t'. split; trivial.
intros. apply Hind. eapply eq_below_mono. apply H. apply lub_leqleft. apply leq_refl.   
Qed.

Lemma xi_Term: forall C F R
  (HmkT: R = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) R (t,D,N)) s' x,
  (forall y, ssqsub (llub (G y) (M (var_svar (progVar y)))) 
             (llub (N (var_svar pcVar)) (llub (D x) (N (var_svar (progVar x)))))
                 -> s y = s' y) ->
  exists t', Sem s' C t' /\
      t x = t' x.
Proof.
intros.
assert (SUB1: forall y : PVar,
    ssqsub (llub (G y) (M (var_svar (progVar y))))
           (N (var_svar pcVar)) ->
    s y = s' y).
  intros. apply H. apply (leq_trans H0). apply lub_leqleft.
subst.
destruct (Theorem6 C F _ (eq_refl _) _ _ _ _ _ _ HRSem _ SUB1) as [t' [Ht1' Ht2']].
exists t'; split; trivial.
apply Ht2'. intros. apply H; clear H.
  apply (leq_trans H0). apply lub_leqright.
Qed.

(*Let's again fold the definition of eq_below in Theorem6*)
Goal  forall C F s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N)) s'
  (HB: eq_below (N (var_svar pcVar)) (MERGE G M) s s'),
  exists t', 
   (Sem s' C t' /\ 
    (forall x, eq_below (MERGE D N x) (MERGE G M) s s' -> t x = t' x)).
Proof.
intros.
apply (Theorem6 C F _ (eq_refl _) _ _ _ _ _ _ HRSem _ HB). 
Qed.  

Goal  forall C F s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N)) s'
  (HB: eq_below (N (var_svar pcVar)) (MERGE G M) s s'),
  exists t', 
   Sem s' C t' /\ 
   forall kappa, eq_below kappa (MERGE G M) s s' -> eq_below kappa (MERGE D N) t t'.
Proof.
intros.
destruct (Theorem6 C F _ (eq_refl _) _ _ _ _ _ _ HRSem _ HB) as [t' [Ht' EQ]].
exists t'. split; trivial.
intros kappa HEQ y Hy.
  eapply EQ.
  intros. apply HEQ. apply (leq_trans H). apply Hy.  
Qed.  

Lemma xi_JoinSem_Sem: forall C F s G M t D N
(Hyp: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N)), 
  Sem s C t. 
Proof.
intros C.
induction C; intros; subst; simpl in *.
(*Skip*)  
  inversion Hyp; subst; simpl; clear Hyp.
  constructor.  trivial. 
 (*Assign*)
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4.
  constructor.  trivial. 
(*Comp*)
  inversion Hyp; subst; simpl; clear Hyp. 
  destruct r as [[r K] P]. 
  assert (IH1 := IHC1 _ _ _ _ _ _ _ H2). 
  assert (IH2 := IHC2 _ _ _ _ _ _ _ H4). 
  econstructor.  apply IH1.  assumption. 
(*If*)
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4.
  inversion H2; subst; simpl; clear H2.
  inversion H5; subst; simpl; clear H5; simpl in *.
  (*True*) clear IHC2.
     inversion H6; subst; simpl; clear H6. destruct r as [[r K] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5.
     subst.
     assert (IH1 := IHC1 _ _ _ _ _ _ _ H2).
     apply S_True; assumption.
  (*False*) clear IHC1.
     inversion H6; subst; simpl; clear H6. destruct r as [[r K] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]; clear H5.
     subst.
     assert (IH1 := IHC2 _ _ _ _ _ _ _ H2).
     apply S_False; assumption.
(*While*)
  remember (fresh F) as FRn. destruct FRn as [n Hn].
  inversion Hyp; subst; simpl; clear Hyp.
  inversion H2; subst; simpl; clear H2.
  inversion H4; subst; simpl; clear H4.
  inversion H2; subst; simpl; clear H2.
  inversion H5; subst; simpl; clear H5. destruct r as [[r K] P].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]; clear H4.
  subst.
  remember (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e)))))) as LOOP.
  remember (s, G,
       zupd SVar SVar_eq
         (zupd SVar SVar_eq M (var_svar pcVar)
            (llub (lab_eval G (e :: nil))
               (llub (M (var_svar pcVar)) (zeval SVar M (mkSVars e)))))
         (cond_svar n)
         (llub (lab_eval G (e :: nil))
            (zeval SVar
               (zupd SVar SVar_eq M (var_svar pcVar)
                  (llub (lab_eval G (e :: nil))
                     (llub (M (var_svar pcVar)) (zeval SVar M (mkSVars e)))))
               (allPCs F ++ mkSVars e)))) as ST.
  remember  (r, K, P) as TT.
  assert (IH: Sem (fst (fst ST)) (While e C) (fst (fst TT))).
    clear HeqST HeqTT M G s P K r.
    induction H2; intros; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
    (*True*) clear H2_0 IHJoinSem1.
             destruct r as [[r K] P]. destruct s as [[s G] M]. destruct t as [[t D] N]; simpl in *.
             inversion H2_; subst; clear H2_.
             inversion H5; subst; clear H5.
             inversion H7; subst; clear H7.
             inversion H5; subst; clear H5.
             inversion H8; subst; clear H8. destruct r0 as [[u A] Q].
             destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]; clear H4.
             subst.
             assert (IH := IHC _ _ _ _ _ _ _ H3).
             eapply S_WTrue; try assumption. apply IH. apply (IHJoinSem2 (eq_refl _)).
    (*False*) destruct t as [[t D] N]; simpl in *.
             eapply S_WFalse; trivial.
  subst; simpl in *. apply IH.
Qed.

(*written approx in paper*)
Definition ZStore_sim (M M': ZStore SVar) : Prop :=
  forall z, z <> var_svar pcVar -> M z = M' z.

Lemma ZStore_sim_refl: forall M, ZStore_sim M M.
Proof.
intros M x Hx; trivial.
Qed.

Lemma ZStore_sim_trans: forall {M P} N
(HMN: ZStore_sim M N) (HNP:ZStore_sim N P), ZStore_sim M P.
Proof.
intros. intros x Hx.
rewrite (HMN x Hx). apply (HNP x Hx).
Qed.

Lemma ZStore_sim_sym: forall {M N} (HMN:ZStore_sim M N), ZStore_sim N M.
Proof.
intros M N HMN x Hx.
rewrite (HMN x Hx); trivial.
Qed.

Lemma ZStore_sim_allPCc: forall M M' (HMM':ZStore_sim M M') F,
  zeval SVar M (allPCs F) = zeval SVar M' (allPCs F).
Proof.
intros.
apply allPCs_pcVar_eq.
intros. apply HMM'. apply H.
Qed.

Lemma ZStore_sim_mkSVars: forall M M' (HMM':ZStore_sim M M') e,
  zeval SVar M (mkSVars e) = zeval SVar M' (mkSVars e).
Proof.
intros.
apply mkSVars_pcVar_eq. 
intros. apply HMM'. apply H. 
Qed.

Lemma ZStore_sim_app: forall M M' (HMM':ZStore_sim M M') F e,
  zeval SVar M (allPCs F ++ mkSVars e) =
  zeval SVar M' (allPCs F ++ mkSVars e).
Proof.
intros. 
repeat rewrite zeval_app.
rewrite (ZStore_sim_allPCc _ _ HMM').
rewrite (ZStore_sim_mkSVars _ _ HMM').
trivial.
Qed.

(*Agreement of xi and iota on all variables and tains except for the pcTaint *)
Theorem xi_iota_equiv: forall C F s G M t D N
  (Hyp: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N))
   M' t' D' N'
  (Hyp': JoinSem _ SVar_eq (s, G, M') (iota C F) (t', D', N'))
  (HMM': ZStore_sim M M'), t=t' /\ D=D' /\ ZStore_sim N N'.
Proof.
intros C.
induction C; simpl; intros; subst.
(*Skip*)
  inversion Hyp; subst; clear Hyp.
  inversion Hyp'; subst; clear Hyp'.
  split; trivial.
  split; trivial.
(*Assign*)
  inversion Hyp; subst; clear Hyp.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion Hyp'; subst; clear Hyp'.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  split; trivial.
  split; trivial.
  rewrite (ZStore_sim_app _ _ HMM').
  intros z Hz.
  apply supd_split_at; intros.
    subst. rewrite supd_Same. trivial.
  rewrite supd_Diff; try discriminate. apply HMM'. apply Hz. apply H.
(*Comp*)
  inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [[r' K'] P'].
  destruct (IHC1 _ _ _ _ _ _ _ H2 _ _ _ _ H3 HMM') as [IH1a [IH1b IH1c]]. subst.
  apply (IHC2 _ _ _ _ _ _ _ H4 _ _ _ _ H6 IH1c). 
(*Ite*)
  remember (fresh F) as Fr. destruct Fr as [n Fn].
  inversion Hyp; subst; clear Hyp. destruct r as [[r K] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion Hyp'; subst; clear Hyp'. destruct r0 as [[r' K'] P'].
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5; simpl in *.
  (*True*) clear IHC2.
     inversion H7; subst; clear H7. destruct r as [[u G] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
     inversion H4; subst; clear H4; simpl in *.
     (*True*)
        inversion H8; subst; clear H8. destruct r as [[u' G'] P'].
        destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
        destruct (IHC1 _ _ _ _ _ _ _ H2 _ _ _ _ H3) as [IHa [IHb IHc]]; clear IHC1 H2 H3.
          intros z Hz.
            apply supd_split_at; intros.
              subst. rewrite supd_Same. rewrite <- zeval_app_pc_supd. rewrite (ZStore_sim_app _ _ HMM'). trivial.
              repeat rewrite supd_Diff. apply HMM'. apply Hz. apply H. intros Z. subst.  apply Hz. trivial.
        subst.
        split; trivial.
        split; trivial.
        intros z Hz. destruct z. destruct v. exfalso. apply Hz. trivial.
                                 rewrite (IHc (cond_svar n)); try discriminate.
                                 rewrite (IHc (var_svar (progVar p))); trivial.
                     apply IHc; trivial.
     (*False*) rewrite H7 in H6. inversion H6.
  (*False*) clear IHC1.
     inversion H7; subst; clear H7. destruct r as [[u G] P].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
     inversion H4; subst; clear H4; simpl in *.
     (*True*) rewrite H7 in H6. inversion H6.
     (*False*)
        inversion H8; subst; clear H8. destruct r as [[u' G'] P'].
        destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
        destruct (IHC2 _ _ _ _ _ _ _ H2 _ _ _ _ H3) as [IHa [IHb IHc]]; clear IHC2 H2 H3.
          intros z Hz.
            apply supd_split_at; intros.
              subst. rewrite supd_Same. rewrite <- zeval_app_pc_supd. rewrite (ZStore_sim_app _ _ HMM'). trivial.
              repeat rewrite supd_Diff. apply HMM'. apply Hz. apply H. intros Z. subst.  apply Hz. trivial.
        subst.
        split; trivial.
        split; trivial.
        intros z Hz. destruct z. destruct v. exfalso. apply Hz. trivial.
                                 rewrite (IHc (cond_svar n)); try discriminate.
                                 rewrite (IHc (var_svar (progVar p))); trivial.
                     apply IHc; trivial.
(*While*)
  remember (fresh F) as Fr. destruct Fr as [n Fn].
  inversion Hyp; subst; clear Hyp.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5.  destruct r as [[r K] P]. 
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
  inversion Hyp'; subst; clear Hyp'.
  inversion H3; subst; clear H3.
  inversion H5; subst; clear H5. destruct r0 as [[r' K'] P'].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
  remember (JWhile SVar e
          (JComp SVar (xi C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (e :: nil)
                      (var_svar pcVar :: mkSVars e))
                   (JJoin SVar (cond_svar n) (e :: nil)
                      (cond_svar n :: allPCs F ++ mkSVars e)))))) as LOOP.
  remember (JWhile SVar e
          (JComp SVar (iota C (n :: F))
             (JComp SVar (mkJoins n (modVarsL C))
                (JJoin SVar (cond_svar n) (e :: nil)
                   (cond_svar n :: allPCs F ++ mkSVars e))))) as LOOP'.
  assert (IH: forall s G M t D N (HL: JoinSem SVar SVar_eq (s, G, M) LOOP (t, D, N))
                 M' t' D' N' (HL': JoinSem SVar SVar_eq (s, G, M') LOOP' (t', D', N'))
                 (HS: ZStore_sim M M'), t=t' /\ D=D' /\ ZStore_sim N N').
     clear HMM' H2 H3 s G M M' r K P r' K' P'.
     intros s G M t D N Hyp.
     remember (s,G,M) as ST. remember (t,D,N) as TT.
     generalize dependent N. generalize dependent D. generalize dependent t.
     generalize dependent M. generalize dependent G. generalize dependent s.
     induction Hyp; try inversion HeqLOOP; subst; clear HeqLOOP; simpl in *.
     (*True*) clear IHHyp1 Hyp2. destruct s as [[s G] M]. destruct r as [[u K] P]. destruct t as [[t D] N]. simpl in *.
              intros.
              apply eq_sym in HeqST. inversion HeqST; subst; clear HeqST.
              apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT.
              assert (IH := IHHyp2 (eq_refl _) _ _ _ (eq_refl _) _ _ _ (eq_refl _)). clear IHHyp2.
              inversion Hyp1; subst; clear Hyp1.
              inversion H5; subst; clear H5.
              inversion H7; subst; clear H7. 
              inversion H5; subst; clear H5.
              inversion H8; subst; clear H8.
              destruct r as [[v A] Q].
              destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
              inversion HL'; subst; clear HL'; simpl in *.
              (*True*) 
                 inversion H5; subst; clear H5.
                 destruct r0 as [[v' A'] Q'].
                 inversion H9; subst; clear H9.
                 inversion H10; subst; clear H10. 
                 destruct (mkJoins_effect _ _ _ _ _ _ _ _ H5) as [Z1 [Z2 Z3]]. subst. clear H5.
                 destruct (IHC _ _ _ _ _ _ _ H3 _ _ _ _ H6 HS) as [Z1 [Z2 Z3]]; clear IHC H3 H6.
                 apply eq_sym in Z1; apply eq_sym in Z2. subst.
                 apply (IH _ _ _ _ H7); clear IH H7.
                 intros z Hz. destruct z. rewrite supd_Diff; try discriminate.
                                          destruct v0. exfalso. apply Hz; trivial.
                                                       rewrite supd_Diff.
                                                       rewrite supd_Diff; try discriminate.
                                                         rewrite (Z3 (cond_svar n)); try discriminate.
                                                         rewrite (Z3 (var_svar (progVar p))); try discriminate. trivial.
                                                       intros Z. inversion Z.
                              rewrite <- zeval_app_pc_supd.
                              apply supd_split_at; intros Hnn0.
                                 inversion Hnn0; subst.
                                    simpl. repeat rewrite supd_Same.
                                    rewrite supd_Diff; try discriminate.
                                    rewrite (Z3 (cond_svar n0)).
                                    apply lub_triv. trivial.
                                    apply lub_triv. trivial.
                                    apply zeval_app_pcVar_eq; intros.
                                      destruct z. destruct v0. exfalso. apply H0; trivial. 
                                                   rewrite (Z3 (var_svar (progVar p))); trivial.
                                                   apply Z3. apply H0. 
                                                  apply Hz.
                                      repeat rewrite supd_Diff; try discriminate. apply Z3. apply Hz. apply Hnn0.
              (*False*) rewrite H4 in H. inversion H.
     (*False*) intros; subst. apply eq_sym in HeqTT. inversion HeqTT; subst; clear HeqTT. simpl in *.
               inversion HL'; subst; clear HL'; simpl in *.
               (*True*) rewrite H2 in H. inversion H.
               (*False*) apply eq_sym in H5; inversion H5; subst; clear H5.
                         split; trivial.
                         split; trivial.
  clear IHC. subst.
  destruct (IH _ _ _ _ _ _ H2 _ _ _ _ H3) as [Z1 [Z2 Z3]]; clear IH H2 H3.
    intros z Hz. destruct z. repeat rewrite supd_Diff; try discriminate. apply HMM'. apply Hz.
                              intros Z. inversion Z. subst. apply (Hz Z).
                 apply supd_split_at; intros.
                   inversion H; subst; clear H. 
                   rewrite supd_Same. rewrite <- zeval_app_pc_supd. rewrite (ZStore_sim_app _ _ HMM'). trivial.
                   rewrite supd_Diff; try discriminate. rewrite supd_Diff; trivial. apply HMM'. apply Hz.
  subst.
  split; trivial.
  split; trivial.
  intros z Hz. destruct z. destruct v. exfalso. apply Hz; trivial.
                                       rewrite (Z3 (cond_svar n)); try discriminate.
                                       rewrite (Z3 (var_svar (progVar p))); trivial.
                           apply Z3. apply Hz.
Qed. 


Require Export IotaTheta.

Lemma xi_below_theta_MERGE: 
forall C Gamma Delta (tw:FPCT Gamma C Delta) s' G M t' D' 
(HypT: TntSem (s',MERGE G M) (theta _ _ _ tw) (t',D'))
s t D N F
(HypJ: JoinSem _ SVar_eq (s,G,M) (xi C F) (t,D,N))
(SUBG': SQSUB (MERGE G M) (FPCCTXT2CTXT Gamma))
(SUBF: ssqsub (zeval SVar M (allPCs F)) (Gamma pcVar)) x
(HB: eq_below (MERGE D N x) (MERGE G M) s s'),
t x = t' x /\ ssqsub (MERGE D N x) (D' x). 
Proof.
intros.
assert (XI := xi_JoinSem_Sem _ _ _ _ _ _ _ _ HypJ).
destruct (iota_Sem_JoinSem _ _ _ XI F G M)  as [DD [MM HDDMM]]; clear XI.
destruct (xi_iota_equiv _ _ _ _ _ _ _ _ HypJ _ _ _ _ HDDMM) as [Htt' [HD HZS]].
  intros z Hz. trivial.
apply eq_sym in HD. subst.
assert (SQ: (MERGE D MM x) = (MERGE D N x)).
         unfold MERGE; simpl.
         rewrite (HZS (var_svar (progVar x))). trivial.
            intros Z. inversion Z.
rewrite <- SQ in *.
apply (iota_below_theta_MERGE _ _ _ tw _ _ _ _ _ HypT _ _ _ _ _ HDDMM SUBG' SUBF _ HB).
Qed.

Inductive FPTerm : L -> FPCCTXT -> Com -> FPCCTXT -> Prop :=
  FPTerm_Skip: forall q G, FPTerm q G Skip G
| FPTerm_Assign: forall p x E G H q, 
         p = G pcVar ->
         H = fpcupd G (progVar x) (llub p (FPCGEval G (progExpr E))) ->
         FPTerm q G (Assign x E) H
| FPTerm_Comp: forall G G1 G2 C1 C2 q, 
         FPTerm q G C1 G1 -> FPTerm q G1 C2 G2 ->
         FPTerm q G (Comp C1 C2) G2
| FPTerm_Ite: forall p G GG G1 G2 E C1 C2 a H HH q,
         a = FPCGEval G (progExpr E) ->
         p = G pcVar ->
         GG = fpcupd G pcVar (llub p a) ->
         FPTerm q GG C1 G1 -> 
         FPTerm q GG C2 G2 ->
         H = FLUB G1 G2 ->
         HH = fpcupd H pcVar p ->
         ssqsub (llub p a) q ->
         FPTerm q G (Ite E C1 C2) HH
| FPTerm_While: forall p G E C H N q
         (Hp: p = (G pcVar))
         (pre post : nat -> FPCCTXT)
         (PreBelowN: forall i, i < N -> (pre i) <> pre (S i))
         (PreAboveN: forall i, N <= i -> (pre i) = pre (S i))
         (PreZ: pre 0 = fpcupd G pcVar (llub p (FPCGEval G (progExpr E))))
         (PreN: fpcupd (pre N) pcVar p = H)
         (PrePost: forall i, pre (S i) = fpcupd (FLUB (post i) G) pcVar 
                                (llub ((FLUB (post i) G) pcVar) (FPCGEval (FLUB (post i) G) (progExpr E))))
         (HypC: forall i,  FPTerm q (pre i) C (post i))
         (Hq: ssqsub ((pre N) pcVar) q),
         FPTerm q G (While E C) H.

Lemma FPTerm_FPC: forall q G C D (HF: FPTerm q G C D), FPC G C D.
Proof.
intros.
induction HF; intros; subst.
  constructor.
  econstructor. reflexivity. trivial.
  econstructor. apply IHHF1. apply IHHF2.
  econstructor. reflexivity. reflexivity. reflexivity.
       apply IHHF1. apply IHHF2. reflexivity. trivial.
  econstructor. reflexivity. apply PreBelowN. apply PreAboveN.
       apply PreZ. reflexivity. apply PrePost. apply H0.
Qed.

Lemma FPCTerm_sub: forall q G C D (HF: FPTerm q G C D) p (Hp: ssqsub q p), FPTerm p G C D.
Proof.
intros.
induction HF; intros; subst.
  constructor. (* eapply (leq_trans H Hp).*)
  econstructor. reflexivity. trivial. (* eapply (leq_trans H2 Hp).*)
  econstructor. apply IHHF1. apply Hp. apply IHHF2. apply Hp.
  econstructor. reflexivity. reflexivity. reflexivity.
       apply IHHF1. apply Hp. apply IHHF2. apply Hp. reflexivity. trivial.
       eapply leq_snart. apply Hp. apply H5.
  econstructor. reflexivity. apply PreBelowN. apply PreAboveN.
       apply PreZ. reflexivity. apply PrePost.
       intros. apply (H0 i Hp).
       eapply leq_snart. apply Hp. apply Hq.
Qed.

Definition term C s := exists t, Sem s C t.

Definition coterm (q:L) (G:CTXT) C:=
   forall s s', eq_below q G s s' -> (term C s <-> term C s').

Lemma coterm_mono: forall C q G q' G' (HC: coterm q G C) (Hq: ssqsub q q') 
        (HG: SQSUB G' G), coterm q' G' C.
Proof.
unfold coterm; intros.
apply HC.
eapply eq_below_mono. apply H. assumption. assumption.
Qed.

(*Corresponds to Lemma7 in paper*)
Lemma FPTerm_coterm: forall q G C D (HF:FPTerm q G C D), 
      coterm q (FPCCTXT2CTXT G) C.
Proof.
intros.
induction HF; unfold coterm; intros.
(*Skip*)
  unfold term; split; intros.
  exists s'. constructor. trivial.
  exists s. constructor. trivial.
(*Assign*) subst.
  unfold term; split; intros.
  eexists. constructor. reflexivity. 
  eexists. constructor. reflexivity.
(*Comp*)
  apply FPTerm_FPC in HF1.
  unfold term; split; intros.
  destruct H0 as [r Hr].
    inversion Hr; subst; clear Hr. remember r as t. clear Heqt r. remember r0 as r. clear Heqr r0. 
    assert (X1 := IHHF1 _ _ H). unfold term in X1; destruct X1 as [X _].
    destruct X as [r' Hr']. exists r; assumption.
    assert (EB1 := FPC_Sound3 _ _ _ HF1 _ _ H3 _ _ Hr' _ H).
    assert (X2 := IHHF2 _ _ EB1). unfold term in X2; destruct X2 as [X _].
    destruct X as [t' Ht']. exists t; assumption.
    exists t'. econstructor. apply Hr'. apply Ht'.
  destruct H0 as [r' Hr'].
    inversion Hr'; subst; clear Hr'. remember r' as t'. clear Heqt' r'. remember r as r'. clear Heqr' r.
    assert (X1 := IHHF1 _ _ H). unfold term in X1; destruct X1 as [_ X].
    destruct X as [r Hr]. exists r'; assumption.
    assert (EB1 := FPC_Sound3 _ _ _ HF1 _ _ Hr _ _ H3 _ H).
    assert (X2 := IHHF2 _ _ EB1). unfold term in X2; destruct X2 as [_ X].
    destruct X as [t Ht]. exists t'; assumption.
    exists t. econstructor. apply Hr. apply Ht.
(*Ite*)
  subst.
  clear HF1 HF2. rewrite FPCCTXT2CTXT_upd_pcVar in *.
  split; intros.
    destruct H as [t Ht].
    inversion Ht; subst; clear Ht.
    (*True*)
       clear IHHF2.
       unfold coterm in IHHF1.
       destruct (IHHF1 s s' H6) as [T1 _].
       destruct T1 as [t' Ht']. exists t; apply H7.
       exists t'. apply S_True; try assumption.
         rewrite (eq_below_evalE E _ _ _ _ H6) in H4. apply H4.
         eapply leq_snart. apply H5. apply lub_leqright.        
    (*False*)
       clear IHHF1.
       unfold coterm in IHHF2.
       destruct (IHHF2 s s' H6) as [T1 _].
       destruct T1 as [t' Ht']. exists t; apply H7.
       exists t'. apply S_False; try assumption.
         rewrite (eq_below_evalE E _ _ _ _ H6) in H4. apply H4.
         eapply leq_snart. apply H5. apply lub_leqright.
    destruct H as [t' Ht'].
    inversion Ht'; subst; clear Ht'.
    (*True*)
       clear IHHF2.
       unfold coterm in IHHF1.
       destruct (IHHF1 s s' H6) as [_ T1].
       destruct T1 as [t Ht]. exists t'; apply H7.
       exists t. apply S_True; try assumption.
         rewrite (eq_below_evalE E _ _ _ _ H6). apply H4.
         eapply leq_snart. apply H5. apply lub_leqright.        
    (*False*)
       clear IHHF1.
       unfold coterm in IHHF2.
       destruct (IHHF2 s s' H6) as [_ T1].
       destruct T1 as [t Ht]. exists t'; apply H7.
       exists t. apply S_False; try assumption.
         rewrite (eq_below_evalE E _ _ _ _ H6). apply H4.
         eapply leq_snart. apply H5. apply lub_leqright.
(*While*)
  subst.
  assert (HypC0 := HypC 0).
  apply FPTerm_FPC in HypC0. 
  assert (INCPre: forall i, FSQSUB (pre i) (pre (S i))).
    intros j. induction j. (*rewrite PreZ. apply bot_leq.*)
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ HypC0).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). unfold FLUB; simpl. apply lub_leqleft. 
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff. unfold FLUB; simpl. apply lub_leqright.
                       discriminate. discriminate.
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply FPC_mono. assert (HypSj := HypC (S j)). apply FPTerm_FPC in HypSj. apply HypSj.
                assert (Hypj := HypC j). apply FPTerm_FPC in Hypj. apply Hypj. apply IHj.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                          eapply leq_trans. apply leq_monoright. apply GEval_mono. intros z. apply POST_FL.
                          apply leq_monoleft. apply POST_FL.
                      repeat rewrite fpcupd_Diff. apply POST_FL.
                        discriminate. discriminate.
 assert (INCPreStrong: forall z i, z <= i -> FSQSUB (pre z) (pre i)).
   intros z. 
   induction z; simpl.
     intros i.
     induction i; intros; simpl. apply leq_refl.
     eapply leq_trans. apply IHi. apply Le.le_0_n. apply INCPre.
   intros i.
   induction i; intros; simpl. inversion H.
   destruct (Lt.le_lt_or_eq _ _ H).
     eapply leq_trans. apply IHi. unfold lt in H2. apply Le.le_S_n. apply H2.
            apply INCPre.
     rewrite H2. apply leq_refl. 

  assert (SUBG: FSQSUB G (pre N)).
    eapply leq_snart. apply (INCPreStrong 0). apply Le.le_O_n.
    rewrite PreZ. intros z. destruct z. rewrite fpcupd_Same. apply lub_leqleft.
         rewrite fpcupd_Diff; try discriminate. apply leq_refl.
  assert (SUBN: FSQSUB (post N) (pre N)).
    rewrite (PreAboveN N). rewrite PrePost. 
         intros z; simpl. destruct z. rewrite fpcupd_Same. apply lubleft. unfold FLUB; simpl. apply lub_leqleft.
          rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
       apply Le.le_refl.
  clear INCPre INCPreStrong.
  assert (HypCN:= HypC N). clear HypC. apply FPTerm_FPC in HypCN. clear HypC0.
  assert (cotermC:= (H0 N)).
  assert (Hqq: ssqsub (GEval (FPCCTXT2CTXT (pre N)) E) q).
    rewrite PreAboveN. rewrite PrePost. rewrite FPCCTXT2CTXT_upd_pcVar.
    eapply leq_snart. apply Hq. rewrite PreAboveN. rewrite PrePost.
         rewrite fpcupd_Same. apply lub_leqright. 
       apply Le.le_refl. 
       apply Le.le_refl. 
  clear H0 PreBelowN PreAboveN PrePost PreZ. 
  
  assert (EB:  eq_below q (FPCCTXT2CTXT (pre N)) s s').
    eapply eq_below_mono. apply H1. apply leq_refl. intros z. apply SUBG.
  clear H1 SUBG Hq G.

  split; intros.
    destruct H as [t Ht]. remember (While E C) as D.
    generalize dependent s'.
    induction Ht; subst; simpl; try inversion HeqD; subst; simpl.
       intros.
       clear IHHt1.
       unfold coterm in cotermC. destruct (cotermC _ _ EB) as [X _].
       destruct X as [r' Hr']. exists r. apply Ht1.
       assert (EBr: eq_below q (FPCCTXT2CTXT (pre N)) r r').
         assert (EB1 := FPC_Sound3 _ _ _ HypCN _ _ Ht1 _ _ Hr' _ EB).
         eapply eq_below_mono. apply EB1. apply leq_refl. intros z. apply SUBN.
       destruct (IHHt2 HeqD _ EBr) as [t' Ht'].
       exists t'.
       eapply S_WTrue.
         rewrite (eq_below_evalE E _ _ _ _ EB) in H. apply H. assumption.
         apply Hr'. apply Ht'.
    intros.
      exists s'.
      eapply S_WFalse. 
         rewrite (eq_below_evalE E _ _ _ _ EB) in H. apply H. assumption.
         trivial.
 
  destruct H as [t' Ht']. remember (While E C) as D.
    generalize dependent s.
    induction Ht'; subst; simpl; try inversion HeqD; subst; simpl.
       remember s as s'. clear Heqs' s. remember r as r'. clear Heqr' r.  remember t as t'. clear Heqt' t.
       intros s EB.
       clear IHHt'1.
       unfold coterm in cotermC. destruct (cotermC _ _ EB) as [_ X].
       destruct X as [r Hr]. exists r'. apply Ht'1.
       assert (EBr: eq_below q (FPCCTXT2CTXT (pre N)) r r').
         assert (EB1 := FPC_Sound3 _ _ _ HypCN _ _ Hr _ _ Ht'1 _ EB).
         eapply eq_below_mono. apply EB1. apply leq_refl. intros z. apply SUBN.
       destruct (IHHt'2 HeqD _ EBr) as [t Ht].
       exists t.
       eapply S_WTrue.
         rewrite (eq_below_evalE E _ _ _ _ EB). apply H. assumption.
         apply Hr. apply Ht.
    intros.
      exists s.
      eapply S_WFalse. 
         rewrite (eq_below_evalE E _ _ _ _ EB). apply H. assumption.
         trivial.
Qed.

Lemma Theorem_7: 
forall q Gamma C Delta (HC:FPTerm q Gamma C Delta) F U
  (HmkTU: U = xi C F) s G M t D N
  (HRSem: JoinSem _ SVar_eq (s,G,M) U (t,D,N))
  (SUBG: SQSUB (MERGE G M) (FPCCTXT2CTXT Gamma))
  (HF: ssqsub (zeval _ M (allPCs F)) (Gamma pcVar))
  (SUBF: ssqsub (M (var_svar pcVar)) q),
ssqsub (N (var_svar pcVar)) q /\
  SQSUB (MERGE D N) (FPCCTXT2CTXT Delta).
Proof.
intros q Gamma C Delta HC.
induction HC; intros; subst; simpl in *.
(*Skip*)
  inversion HRSem; subst; clear HRSem.
  split; assumption.
(*Assign*)
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  unfold lab_eval; simpl.
  rewrite FPCCTXT2CTXT_upd_progVar.
  split. rewrite supd_Diff; try discriminate. assumption.
  intros z; unfold MERGE; simpl.
  unfold upd in *.
  remember (pvar_eq x z) as b; destruct b; subst; simpl in *.
    rewrite supd_Same. 
    apply lub_leq.
      apply lubright. apply GEval_mono. eapply leq_snart. apply SUBG. apply MERGE_L. 
      rewrite bbot_id. 
      rewrite zeval_app.
      apply lub_leq. eapply (leq_trans HF). apply lub_leqleft.
          apply lubright. rewrite zeval_GEval. apply GEval_mono. eapply leq_snart. apply SUBG. apply MERGE_R.
   rewrite supd_Diff. apply SUBG. intros Z. inversion Z; subst. apply n. trivial.
(*Comp*)
  inversion HRSem; subst; clear HRSem. destruct r as [[r K] P].
  destruct (IHHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2 SUBG HF SUBF) as [IH1a IH1b]; clear IHHC1.   
  apply (IHHC2 _ _ (eq_refl _) _ _ _ _ _ _ H4 IH1b).
    rewrite <- (allPCs_conds_eq _ M P).
      apply FPTerm_FPC in HC1. apply FPC_Lemma1 in HC1. rewrite <- HC1. assumption.
      intros. eapply (xi_ControlLabelsPreserved _ _ _ (eq_refl _) _ _ _ _ _ _ H2). apply H.
    assumption.
(*Ite*)
  rewrite FPCCTXT2CTXT_upd_pcVar.
  remember (fresh F) as FRESH. destruct FRESH as [m FRm].
  inversion HRSem; subst; clear HRSem. destruct r as [[r A] P].
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2. rewrite FPCCTXT2CTXT_upd_pcVar in *. rewrite <- zeval_app_pc_supd in *.
  
  assert (HL1: ssqsub (GEval A E) (GEval (fun x : PVar => G (progVar x)) E)).
     apply GEval_mono. eapply leq_trans. apply (MERGE_L _ M). apply SUBG.
  assert (HL2 : ssqsub (GEval (fun x : PVar => M (var_svar (progVar x))) E) (GEval (fun x : PVar => G (progVar x)) E)).
     apply GEval_mono. eapply leq_trans. apply (MERGE_R A). apply SUBG.
  assert (HL3: ssqsub (llub (GEval A E) (zeval SVar M (allPCs F ++ mkSVars E)))
                   (llub (G pcVar) (GEval (fun x : PVar => G (progVar x)) E))).
      clear H6 IHHC1 IHHC2 HC1 HC2 H5 SUBF.
      apply lub_leq. apply lubright. apply HL1.
      rewrite zeval_app. rewrite zeval_GEval.   
        apply lub_leq. apply lubleft. apply HF.
        apply lubright. apply HL2.
  inversion H6; subst; clear H6; simpl in H4.
  (*True*) clear IHHC2.
     inversion H7; subst; clear H7. destruct r0 as [[u B] Q].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
     assert (XX: Q (cond_svar m) = (llub (GEval A E) (llub (zeval SVar M (allPCs F)) (zeval SVar M (mkSVars E))))).
        rewrite <- (xi_ControlLabelsPreserved _ _ _ (eq_refl _) _ _ _ _ _ _ H2).
        rewrite supd_Same. rewrite lab_eval_GEval.
          rewrite zeval_app. trivial.
        left. trivial.
     assert (IH1 := IHHC1 _ _ (eq_refl _) _ _ _ _ _ _ H2). clear IHHC1 H2.
     rewrite zeval_GEval in XX. rewrite lab_eval_GEval in *.
     assert (ZZ: ssqsub (Q (var_svar pcVar)) q /\ SQSUB (MERGE B Q) (FPCCTXT2CTXT G1)).
       apply IH1; clear IH1.
         intros z; unfold MERGE; simpl.
           rewrite supd_Diff; try discriminate.
           rewrite supd_Diff; try discriminate.
           apply SUBG.
         rewrite fpcupd_Same. simpl.
           rewrite supd_Same.
           rewrite <- zeval_allPCs_supd. rewrite <- zeval_allPCs_pc_supd.
             apply lub_leq. apply HL3.
             apply lubleft. apply HF.
           intros Z. apply (FRm Z).

         rewrite supd_Diff; try discriminate. rewrite supd_Same.  
           apply lub_leq. eapply (leq_trans HL1). eapply leq_snart. apply H5. apply lub_leqright.
             apply lub_leq. apply SUBF.
               rewrite zeval_GEval. eapply leq_snart. apply H5. apply lubright. apply HL2. 
    clear IH1. destruct ZZ as [ZZ1 ZZ2].
     split; try assumption.
     intros z; unfold MERGE; simpl.
       apply lub_leq. unfold FLUB; simpl. apply lubleft.
            eapply leq_snart. apply ZZ2. apply lub_leqleft. 
       assert (HZ: ssqsub (Q (var_svar (progVar z)))(FPCCTXT2CTXT (FLUB G1 G2) z)).
         unfold FLUB; simpl. apply lubleft.
            eapply leq_snart. apply ZZ2. apply lub_leqright.        
       remember (in_dec pvar_eq z (modVarsL C2)) as b.
         destruct b.
            apply lub_leq; try apply HZ.
            rewrite XX; clear XX.
               rewrite zeval_app in HL3. rewrite zeval_GEval in HL3.
               eapply (leq_trans HL3). 
               apply FPTerm_FPC in HC2. eapply leq_snart. unfold FLUB. simpl. apply lubright.
                 eapply FPC_Mod2. apply HC2. apply modVarsChar. apply i.
               rewrite <- (FPC_Lemma1 _ _ _ HC2). rewrite fpcupd_Same. apply leq_refl.
         apply HZ. 
  (*False*) clear IHHC1.
     inversion H7; subst; clear H7. destruct r0 as [[u B] Q].
     destruct (mkJoins_effect _ _ _ _ _ _ _ _ H6) as [Z1 [Z2 Z3]]. subst. clear H6.
     assert (XX: Q (cond_svar m) = (llub (GEval A E) (llub (zeval SVar M (allPCs F)) (zeval SVar M (mkSVars E))))).
        rewrite <- (xi_ControlLabelsPreserved _ _ _ (eq_refl _) _ _ _ _ _ _ H2).
        rewrite supd_Same. rewrite lab_eval_GEval.
          rewrite zeval_app. trivial.
        left. trivial.
     assert (IH2 := IHHC2 _ _ (eq_refl _) _ _ _ _ _ _ H2). clear IHHC2 H2.
     rewrite zeval_GEval in XX. rewrite lab_eval_GEval in *.
     assert (ZZ: ssqsub (Q (var_svar pcVar)) q /\ SQSUB (MERGE B Q) (FPCCTXT2CTXT G2)).
       apply IH2; clear IH2.
         intros z; unfold MERGE; simpl.
           rewrite supd_Diff; try discriminate.
           rewrite supd_Diff; try discriminate.
           apply SUBG.
         rewrite fpcupd_Same. simpl.
           rewrite supd_Same.
           rewrite <- zeval_allPCs_supd. rewrite <- zeval_allPCs_pc_supd.
             apply lub_leq. apply HL3.
             apply lubleft. apply HF.
           intros Z. apply (FRm Z).

         rewrite supd_Diff; try discriminate. rewrite supd_Same.  
           apply lub_leq. eapply (leq_trans HL1). eapply leq_snart. apply H5. apply lub_leqright.
             apply lub_leq. apply SUBF.
               rewrite zeval_GEval. eapply leq_snart. apply H5. apply lubright. apply HL2. 
    clear IH2. destruct ZZ as [ZZ1 ZZ2].
     split; try assumption.
     intros z; unfold MERGE; simpl.
       apply lub_leq. unfold FLUB; simpl. apply lubright.
            eapply leq_snart. apply ZZ2. apply lub_leqleft. 
       assert (HZ: ssqsub (Q (var_svar (progVar z)))(FPCCTXT2CTXT (FLUB G1 G2) z)).
         unfold FLUB; simpl. apply lubright.
            eapply leq_snart. apply ZZ2. apply lub_leqright.        
       remember (in_dec pvar_eq z (modVarsL C1)) as b.
         destruct b.
            apply lub_leq; try apply HZ.
            rewrite XX; clear XX.
               rewrite zeval_app in HL3. rewrite zeval_GEval in HL3.
               eapply (leq_trans HL3). 
               apply FPTerm_FPC in HC1. eapply leq_snart. unfold FLUB. simpl. apply lubleft.
                 eapply FPC_Mod2. apply HC1. apply modVarsChar. apply i.
               rewrite <- (FPC_Lemma1 _ _ _ HC1). rewrite fpcupd_Same. apply leq_refl.
         apply HZ.
(*While*)
  assert (HypC0 := HypC 0).
  apply FPTerm_FPC in HypC0. 
  assert (INCPre: forall i, FSQSUB (pre i) (pre (S i))).
    intros j. induction j. (*rewrite PreZ. apply bot_leq.*)
               rewrite (PrePost 0).
               intros z. destruct z. rewrite fpcupd_Same.
                            rewrite (FPC_Lemma1 _ _ _ HypC0).
                            eapply @leq_trans with(y:= (FLUB (post 0) G pcVar)). unfold FLUB; simpl. apply lub_leqleft. 
                           apply lub_leqleft.
                    rewrite PreZ. repeat rewrite fpcupd_Diff. unfold FLUB; simpl. apply lub_leqright.
                       discriminate. discriminate.
            rewrite (PrePost j).
            rewrite (PrePost (S j)).
            assert (POST_jSj: FSQSUB (post j)(post (S j))).
              eapply FPC_mono. assert (HypSj := HypC (S j)). apply FPTerm_FPC in HypSj. apply HypSj.
                assert (Hypj := HypC j). apply FPTerm_FPC in Hypj. apply Hypj. apply IHj.
            assert (POST_FL: FSQSUB (FLUB (post j) G) (FLUB (post (S j)) G)).
                 apply leq_monoleft. apply POST_jSj.
            intros z. destruct z. repeat rewrite fpcupd_Same.
                          eapply leq_trans. apply leq_monoright. apply GEval_mono. intros z. apply POST_FL.
                          apply leq_monoleft. apply POST_FL.
                      repeat rewrite fpcupd_Diff. apply POST_FL.
                        discriminate. discriminate.
 assert (INCPreStrong: forall z i, z <= i -> FSQSUB (pre z) (pre i)).
   intros z. 
   induction z; simpl.
     intros i.
     induction i; intros; simpl. apply leq_refl.
     eapply leq_trans. apply IHi. apply Le.le_0_n. apply INCPre.
   intros i.
   induction i; intros; simpl. inversion H.
   destruct (Lt.le_lt_or_eq _ _ H).
     eapply leq_trans. apply IHi. unfold lt in H1. apply Le.le_S_n. apply H1.
            apply INCPre.
     rewrite H1. apply leq_refl. 

  assert (SUBGN: FSQSUB G (pre N)).
    eapply leq_snart. apply (INCPreStrong 0). apply Le.le_O_n.
    rewrite PreZ. intros z. destruct z. rewrite fpcupd_Same. apply lub_leqleft.
         rewrite fpcupd_Diff; try discriminate. apply leq_refl.
  assert (SUBN: FSQSUB (post N) (pre N)).
    rewrite (PreAboveN N). rewrite PrePost. 
         intros z; simpl. destruct z. rewrite fpcupd_Same. apply lubleft. unfold FLUB; simpl. apply lub_leqleft.
          rewrite fpcupd_Diff; try discriminate. unfold FLUB; simpl. apply lub_leqleft.
       apply Le.le_refl.
  assert (HpostE: ssqsub (GEval (FPCCTXT2CTXT (post N)) E) (pre N pcVar)).
     clear HRSem H0. rewrite PreAboveN. rewrite PrePost. rewrite fpcupd_Same. unfold FLUB; simpl. apply lubright. apply GEval_mono. intros z. apply lub_leqleft.
      apply Le.le_refl.
  assert (Hpvar: forall z, ssqsub (FPCCTXT2CTXT (post N) z) (FPCCTXT2CTXT (pre N) z)).
     rewrite PreAboveN. rewrite PrePost. intros.
     rewrite FPCCTXT2CTXT_upd_pcVar. unfold FLUB; simpl. apply lub_leqleft.
     apply Le.le_refl.
  clear INCPre INCPreStrong.
  assert (HypCN:= HypC N). clear HypC. apply FPTerm_FPC in HypCN. clear HypC0.

  remember (fresh F) as Fr. destruct Fr as [m Hm].
  assert (HC := H0 N (m::F) _ (eq_refl _)). clear H0.
  rewrite FPCCTXT2CTXT_upd_pcVar.
  assert (HE: ssqsub (GEval (FPCCTXT2CTXT (pre N)) E) (pre N pcVar)).
     clear HC HRSem. eapply leq_snart. rewrite PreAboveN. rewrite PrePost.
         rewrite fpcupd_Same. apply leq_refl. apply Le.le_refl. 
       rewrite PreAboveN. rewrite PrePost.  rewrite FPCCTXT2CTXT_upd_pcVar. apply lub_leqright. apply Le.le_refl. 
   clear PreBelowN PreAboveN PrePost PreZ.
  inversion HRSem; subst; clear HRSem.
  inversion H2; subst; clear H2.
  inversion H4; subst; clear H4.
  inversion H2; subst; clear H2.
  inversion H5; subst; clear H5. destruct r as [[r B] Q].
  destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
  assert (HL1: ssqsub (llub (GEval G0 E) (zeval SVar M (allPCs F ++ mkSVars E))) (pre N pcVar)).
    clear H2 HC. rewrite zeval_lub_MERGE.
    apply lub_leq. eapply leq_snart. apply HE. apply GEval_mono. eapply leq_trans. apply SUBG. intros z. apply SUBGN.
    apply (leq_trans HF). apply SUBGN.
  fold supd in *.
  remember (supd
         (zupd SVar SVar_eq M (var_svar pcVar)
            (llub (lab_eval G0 (E :: nil))
               (llub (M (var_svar pcVar)) (zeval SVar M (mkSVars E)))))
         (cond_svar m)
         (llub (lab_eval G0 (E :: nil))
            (zeval SVar
               (zupd SVar SVar_eq M (var_svar pcVar)
                  (llub (lab_eval G0 (E :: nil))
                     (llub (M (var_svar pcVar)) (zeval SVar M (mkSVars E)))))
               (allPCs F ++ mkSVars E)))) as MM.
    assert (MM1: SQSUB (MERGE G0 MM) (FPCCTXT2CTXT (pre N))).
      clear H2 HC; subst.
      intros z; unfold MERGE; simpl.
      apply lub_leq. eapply leq_trans. apply (MERGE_L _ M). eapply leq_trans. apply SUBG. apply SUBGN.
      rewrite supd_Diff. rewrite supd_Diff; try discriminate.  eapply leq_trans. apply (MERGE_R G0). eapply leq_trans. apply SUBG. apply SUBGN.
        intros Z. inversion Z; subst.  
    assert (MM2: ssqsub (zeval SVar MM (allPCs (m :: F))) (pre N pcVar)).
      clear H2 HC MM1; subst.
      simpl. rewrite supd_Same. rewrite lab_eval_GEval. fold supd. rewrite <- zeval_app_pc_supd.
      apply lub_leq. apply HL1. 
        rewrite <- zeval_allPCs_supd. rewrite <- zeval_allPCs_pc_supd. eapply (leq_trans HF). apply SUBGN.
        intros Z. apply (Hm Z).
    assert (MM3:ssqsub (MM (var_svar pcVar)) q).
      clear H2 HC MM1 MM2; subst. rewrite supd_Diff; try discriminate. rewrite supd_Same. rewrite lab_eval_GEval.
      apply lub_leq. eapply leq_snart. apply Hq. eapply leq_snart. apply HL1. apply lub_leqleft.
      apply lub_leq. apply SUBF.
      eapply leq_snart. apply Hq. eapply leq_snart. apply HL1. apply lubright. rewrite zeval_app. apply lub_leqright.

    assert (HL1' : ssqsub (llub (GEval G0 E) (zeval SVar MM (allPCs F ++ mkSVars E)))
        (pre N pcVar)).
      clear H2 HC. apply lub_leq. eapply leq_snart. apply HL1. apply lub_leqleft.
      rewrite zeval_app.
      apply lub_leq. eapply leq_snart. apply MM2. simpl. apply lub_leqright.
        subst. fold supd. rewrite mkSvars_upd_cond. rewrite <- zeval_pc_supd. eapply leq_snart. apply HL1. apply lubright. rewrite zeval_app. apply lub_leqright.
          
    clear HeqMM HL1.
    clear SUBG SUBGN. clear HF SUBF. clear G M.

    remember (JWhile SVar E
          (JComp SVar (xi C (m :: F))
             (JComp SVar (mkJoins m (modVarsL C))
                (JComp SVar
                   (JJoin SVar (var_svar pcVar) (E :: nil)
                      (var_svar pcVar :: mkSVars E))
                   (JJoin SVar (cond_svar m) (E :: nil)
                      (cond_svar m :: allPCs F ++ mkSVars E)))))) as PROG.
     remember (s, G0, MM) as PRE. remember (r, B, Q) as POST.
     generalize dependent Q. generalize dependent B. generalize dependent r. 
     generalize dependent MM. generalize dependent G0. generalize dependent s. 
     induction H2; subst; try inversion HeqPROG.
     (*True*) clear IHJoinSem1 HeqPROG; subst. intros; subst. 
              destruct r as [[r A] P].
              assert (IH := IHJoinSem2 (eq_refl _) _ _ _ (eq_refl _)); clear IHJoinSem2.
              clear H2_0.
              inversion H2_; subst; clear H2_. destruct r1 as [[u AA] PP].
              inversion H5; subst; clear H5. destruct r1 as [[u1 AA1] PP1].
              assert (XX: MM (cond_svar m) = PP (cond_svar m)).
                  eapply (xi_ControlLabelsPreserved _ _ _ (eq_refl _) _ _ _ _ _ _ H3). left; trivial.
              assert (YY: zeval SVar MM (allPCs F) = zeval SVar PP (allPCs F)).
                 apply allPCs_conds_eq.
                 intros. eapply (xi_ControlLabelsPreserved _ _ _ (eq_refl _) _ _ _ _ _ _ H3). right; trivial.
              destruct (mkJoins_effect _ _ _ _ _ _ _ _ H4) as [Z1 [Z2 Z3]]. subst. clear H4.
              inversion H7; subst; clear H7. destruct r1 as [[u1 AA1] PP1].
              inversion H6; subst; clear H6.
              inversion H4; subst; clear H4.
              destruct (HC _ _ _ _ _ _ H3) as [HS1 HS2]; clear HC.
                clear IH. assumption. 
                clear IH. assumption. 
                clear IH. assumption. 
              assert (CMod:= xi_Mod _ _ _ (eq_refl _) _ _ _ _ _ _ H3).
              eapply IH; clear IH.
                intros z; unfold MERGE; simpl. rewrite supd_Diff; try discriminate.
                  rewrite supd_Diff. 
                    apply lub_leq. eapply leq_trans. apply (MERGE_L _ PP). eapply leq_trans. apply HS2. apply SUBN.
                    assert (HPP: ssqsub (PP (var_svar (progVar z))) (FPCCTXT2CTXT (pre N) z)).
                      eapply leq_trans. apply (MERGE_R A). eapply leq_trans. apply HS2. apply SUBN.
                    remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b; try assumption.
                       assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar z)))).
                          apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.                    
                       apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz.
                       apply HPP.
                  intros Z; inversion Z.
                simpl. rewrite lab_eval_GEval. rewrite supd_Same.
                  apply lub_leq. 
                    apply lub_leq. eapply leq_snart. apply HpostE. apply GEval_mono. eapply leq_trans. apply (MERGE_L _ PP). apply HS2.
                    simpl. apply lub_leq. simpl. rewrite supd_Diff; try discriminate. rewrite <- XX. eapply leq_snart. apply MM2. simpl. apply lub_leqleft.
                    rewrite <- zeval_app_pc_supd. rewrite zeval_app.
                    apply lub_leq. rewrite <- (allPCs_pcVar_eq PP).
                         rewrite <- (allPCs_conds_eq _ MM). eapply leq_snart. apply MM2. simpl. apply lub_leqright.
                         intros. apply (xi_ControlLabelsPreserved _ _ _ (eq_refl _) _ _ _ _ _ _ H3). right; assumption.
                    intros. destruct z. destruct v. trivial. 
                             remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; trivial.
                               assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar p)))).
                                apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.                     
                                 apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz. trivial.
                            trivial.
                  rewrite zeval_GEval.
                     eapply leq_snart. apply HE. eapply GEval_mono. eapply leq_snart. intros z. apply SUBN.
                     eapply @leq_trans with (y:= (fun z : PVar => PP (var_svar (progVar z)))).
                       intros z. remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b; try apply leq_refl.
                           assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar z)))).
                                apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.                     
                                 apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz. apply leq_refl.
                       eapply leq_trans. apply (MERGE_R A). apply HS2.
                simpl. rewrite <- zeval_allPCs_supd. rewrite <- zeval_allPCs_pc_supd.
                           rewrite (allPCs_pcVar_eq _ PP). rewrite <- YY. eapply leq_snart. apply MM2. simpl. apply lub_leqright.
                       intros. destruct z; trivial. destruct v; trivial.
                        remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; try leq_refl.
                           assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar p)))).
                                apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.
                                 apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz. trivial.
                           trivial.
                      intros Z. apply (Hm Z).
                rewrite supd_Diff; try discriminate. rewrite supd_Same. rewrite lab_eval_GEval.
                     apply lub_leq. eapply leq_snart. apply Hq. eapply leq_snart. apply HpostE. apply GEval_mono. eapply leq_trans. apply (MERGE_L _ PP). apply HS2.
                     simpl.
                      apply lub_leq. apply HS1.
                      rewrite zeval_GEval. eapply @leq_trans with (y:=GEval (fun x => PP (var_svar (progVar x))) E).
                          apply GEval_mono. intros z. remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b.
                              assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar z)))).
                                apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.
                                 apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz. apply leq_refl.
                              apply leq_refl.
                         eapply leq_snart. apply Hq. eapply leq_snart. apply HpostE. apply GEval_mono. eapply leq_trans. apply (MERGE_R A). apply HS2.
                rewrite zeval_all_app. rewrite <- zeval_app_pc_supd.
                  apply lub_leq.  eapply leq_snart. apply HpostE. apply GEval_mono. eapply leq_trans. apply (MERGE_L _ PP). apply HS2.
                  simpl. rewrite zeval_app.
                    apply lub_leq. rewrite <- (allPCs_pcVar_eq PP). rewrite <- YY. eapply leq_snart. apply MM2. simpl. apply lub_leqright.
                       intros. destruct z; trivial. destruct v; trivial.
                         remember (in_dec pvar_eq p (modVarsL C)) as b; destruct b; trivial.
                            assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar p)))).
                                apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.
                                 apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz. trivial.
                      rewrite zeval_GEval. eapply @leq_trans with (y:=GEval (fun x => PP (var_svar (progVar x))) E).
                          apply GEval_mono. intros z. remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b.
                              assert (Hmodz: ssqsub (PP (cond_svar m)) (PP (var_svar (progVar z)))).
                                apply CMod. rewrite modvars_xi. apply modVarsChar. apply i. left; trivial.
                                 apply ccompat_llub in Hmodz. unfold llub. rewrite <- Hmodz. apply leq_refl.
                              apply leq_refl.
                         eapply leq_snart. apply HpostE. apply GEval_mono. eapply leq_trans. apply (MERGE_R A). apply HS2. 
                      intros Z. apply (Hm Z).
                reflexivity.
     (*False*) intros; subst. clear HeqPROG HC.
               inversion HeqPOST; subst; clear HeqPOST H.
               split; trivial.
               intros z; unfold MERGE; simpl.
               apply lub_leq. eapply leq_trans. apply (MERGE_L _ Q). apply MM1.
               assert (HQ: ssqsub (Q (var_svar (progVar z))) (FPCCTXT2CTXT (pre N) z)).
                 eapply leq_trans. apply (MERGE_R B). apply MM1.
               remember (in_dec pvar_eq z (modVarsL C)) as b; destruct b; try assumption.
                 apply lub_leq; try assumption.
                 eapply @leq_trans with (y:=pre N pcVar).
                    eapply leq_snart. apply MM2. simpl. apply lub_leqleft.
                 rewrite (FPC_Lemma1 _ _ _ HypCN). 
                   eapply leq_snart. apply Hpvar. eapply FPC_Mod2. apply HypCN. apply modVarsChar. apply i.
Qed.
