(*From RifleLight.v, but eliminated the 3rd state component*)
Require Export CTXT.

Section TaintLang.
Inductive TCom :=
  TSkip : TCom
| TAssign: list L -> PVar -> Expr -> TCom
| TComp: TCom -> TCom -> TCom
| TIte: Expr -> TCom -> TCom -> TCom
| TWhile: Expr -> TCom -> TCom.

Fixpoint TErase (C:TCom) : Com :=
match C with
  TSkip => Skip
| TAssign s x E => Assign x E
| TComp C1 C2 => Comp (TErase C1) (TErase C2)
| TIte e C1 C2 => Ite e (TErase C1) (TErase C2)
| TWhile e C => While e (TErase C)
end.

Definition lambda_eval (es: list L) :L :=
fold_right llub bbot es.

Lemma lambda_eval1: forall l es, In l es -> ssqsub l (lambda_eval es).
Proof.
intros l es.
induction es; simpl; intros; subst. contradiction.
destruct H; subst. apply lub_leqleft.
apply lubright. apply (IHes H).
Qed. 

Definition TStore:Type := prod Store CTXT. 

Inductive TntSem: TStore -> TCom -> TStore -> Prop :=
  TS_Skip: forall s t, s=t -> TntSem s TSkip t
| TS_Assign: forall x e lambda s G t D, 
     t = upd s x (eval s e) ->
     D = upd G x (llub (GEval G e) (lambda_eval lambda)) -> 
     TntSem (s,G) (TAssign lambda x e) (t,D)
| TS_Comp: forall s C1 C2 t r, TntSem s C1 r -> TntSem r C2 t -> TntSem s (TComp C1 C2) t
| TS_True: forall s E C1 C2 t, istrue (eval (fst s) E) = true ->
           TntSem s C1 t -> TntSem s (TIte E C1 C2) t
| TS_False: forall s E C1 C2 t, istrue (eval (fst s) E) = false -> 
              TntSem s C2 t -> TntSem s (TIte E C1 C2) t
| TS_WTrue: forall s b C r t, istrue (eval (fst s) b) = true -> 
           TntSem s C r -> TntSem r (TWhile b C) t -> 
           TntSem s (TWhile b C) t
| TS_WFalse: forall s b C t, istrue (eval (fst s) b) = false -> s=t -> 
           TntSem s (TWhile b C) t.  

Lemma TErase_Lemma: forall s R t (Hyp:TntSem s R t),
Sem (fst s) (TErase R) (fst t). 
Proof. 
intros s R t Hyp. 
induction Hyp; intros; subst; simpl in *.
(*Skip*) 
  constructor.  trivial. 
(*Assign*) 
  constructor.  trivial. 
(*Comp*)
  econstructor. apply IHHyp1. apply IHHyp2. 
(*IfTrue*)
    apply S_True. assumption. apply IHHyp.
(*IfFalse*)
    apply S_False. assumption. apply IHHyp.
(*WhileTrue*)
    eapply S_WTrue. assumption. apply IHHyp1. apply IHHyp2.
(*WhileFalse*)
    apply S_WFalse. assumption. trivial.
Qed.

Lemma TErase_Lemma_Inv: forall R s t (Hyp:Sem s (TErase R) t) G,
  exists H, TntSem (s,G) R (t,H). 
Proof. 
intros R.
induction R. 
(*Skip*) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  eexists.
  constructor.  reflexivity. 
(*Assign *) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  eexists.
  econstructor.  reflexivity.   reflexivity.
(*Comp*) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  destruct (IHR1 _ _ H2 G) as [H HR]. 
  destruct (IHR2 _ _ H4 H) as [K HR2]. 
  eexists.
  econstructor.  apply HR.  apply HR2. 
(*If*) intros.
  inversion Hyp; subst; simpl; clear Hyp. 
  (*True*) clear IHR2. 
     destruct (IHR1 _ _ H5 G) as [H HR]. 
     eexists.
     apply TS_True.  apply H4.  apply HR.  
  (*False*) clear IHR1. 
     destruct (IHR2 _ _ H5 G) as [H HR]. 
     eexists.
     apply TS_False.  apply H4.  apply HR.
(*While*) 
  intros s t SEM. simpl in SEM.
  remember (While e (TErase R)) as D.
  induction SEM; subst; try inversion HeqD; subst; clear HeqD; intros.
  (*WhileT*)
      destruct (IHR _ _ SEM1 G) as [K Kv].
      destruct (IHSEM2 (eq_refl _) K) as [L Lw].
      exists L. eapply TS_WTrue. apply H. apply Kv. apply Lw.
  (*WhileT*)
      exists G. eapply TS_WFalse. apply H. trivial.
Qed. 

Theorem RifleTErase: forall R s t G,
Sem s (TErase R) t <-> (exists H, TntSem (s,G) R (t,H)). 
Proof.
intros.
split; intros.
  apply TErase_Lemma_Inv. apply H.
destruct H as [H HH].
  apply (TErase_Lemma _ _ _ HH).
Qed.

Lemma TntSem_nomod: forall s R t (SEM:TntSem s R t) x 
 (MV:modVars (TErase R) x = false), fst s x = fst t x.
Proof.
intros s R t SEM.
induction SEM; intros; subst; simpl in *; trivial; simpl in MV. 
(*Assign*)
apply pvar_eq_ElimF in MV.
rewrite M_upd_Diff; trivial.
(*Comp*)
apply Bool.orb_false_iff in MV. destruct MV as [MV1 MV2].
rewrite <- (IHSEM2 x MV2).
apply (IHSEM1 _ MV1).
(*If-True*)
apply Bool.orb_false_iff in MV. destruct MV as [MV1 MV2].
apply (IHSEM x MV1).
(*If-False*)
apply Bool.orb_false_iff in MV. destruct MV as [MV1 MV2].
apply (IHSEM x MV2).
(*While*)
rewrite <- (IHSEM2 x MV).
apply (IHSEM1 _ MV).
Qed. 

Lemma TSem1_aux: forall R pre post
(Hyp: TntSem pre R post) x,
(modVars (TErase R) x = false) ->
((fst pre) x = (fst post) x /\ (snd pre) x = (snd post) x). 
Proof. 
intros R pre post Hyp. 
induction Hyp; subst; intros; simpl in *. 
(*Skip*)
 split; trivial.
(*Assign*)  
  apply pvar_eq_ElimF in H.
  repeat rewrite M_upd_Diff; trivial. split; trivial.
(*Comp*)
  apply Bool.orb_false_iff in H. destruct H as [MV1 MV2].
  destruct (IHHyp1 _ MV1) as [HH1 HH2].  rewrite HH1.  rewrite HH2. 
  apply (IHHyp2 _ MV2). 
(*True*)
  apply Bool.orb_false_iff in H0. destruct H0 as [MV1 MV2].
  apply (IHHyp _ MV1). 
(*False*)
  apply Bool.orb_false_iff in H0. destruct H0 as [MV1 MV2].
  apply (IHHyp _ MV2). 
(*WhileT*)
  intros. 
  destruct (IHHyp2 x H0). rewrite <- H1. rewrite <- H2. clear H1 H2.
  apply IHHyp1. apply H0.
(*False*)
  split; trivial. 
Qed. 

Lemma TSem1: forall R s G t H
(Hyp: TntSem (s,G) R (t,H)) x,
(modVars (TErase R) x = false) -> (s x = t x /\ G x = H x). 
Proof. 
intros. 
apply (TSem1_aux _ _ _ Hyp _ H0). 
Qed.
End TaintLang.

Fixpoint mkMergeTCom (p:L) (xs :list PVar): TCom :=
match xs with
  nil => TSkip
| (x::xss) => TComp (TAssign (cons p nil) x (varExpr x)) (mkMergeTCom p xss)
end.

Definition CompCode kappa C := mkMergeTCom kappa (modVarsL C).

(*Properties about CompCode and mkMergeTCom*)
Lemma mergeTntSemexists: 
forall V p s G, exists H, 
TntSem (s, G) (mkMergeTCom p V) (s,H). 
Proof. 
intros V. 
induction V; intros; subst; simpl. 
  eexists.  constructor.  reflexivity. 
destruct (IHV p (upd s a (eval s (varExpr a)))
         (upd G a (llub (GEval G (varExpr a)) p))) as [K HK].
eexists; econstructor. 
  econstructor; reflexivity. 
  assert (upd s a (eval s (varExpr a)) = s). 
    apply extensionality.  intros y. 
    unfold upd; simpl. 
    remember ( pvar_eq a y) as b; destruct b; subst; trivial. 
  rewrite H in *. simpl in *. rewrite id_bbot. apply HK.
Qed. 

Lemma MergeContinuationNotin: forall V s G t D p  
    y (MV : ~In y V)
   (Hyp: TntSem (s,G) (mkMergeTCom p V) (t,D)),
   s y = t y /\ G y = D y. 
Proof. 
intros V. 
induction V; intros; simpl in *; inversion Hyp; subst; clear Hyp.
  inversion H; subst; clear H.  split; trivial.
    inversion H2; subst; clear H2.
    remember (pvar_eq a y) as b; destruct b; subst.  simpl in *. 
       apply False_ind.  apply MV.  left.  trivial. 
    assert (~ In y V).  intros N.  apply MV.  right; assumption.  clear MV. 
    destruct (IHV _ _ _ _ _ y H H4) as [HH1 HH2].
    rewrite <- HH1.
    unfold upd in *.  rewrite <- Heqb in *.  
     split; trivial.
Qed. 

Lemma MergeContinuationIn: forall V s G t D p  
    y (MV : In y V)
    (Hyp: TntSem (s,G) (mkMergeTCom p V) (t,D)),
    s y = t y /\ 
    llub (G y) p = D y.
Proof. 
intros V. 
induction V; intros; simpl in *.  contradiction. 
inversion Hyp; subst; clear Hyp. 
  inversion H2;  subst; simpl in *;  clear H2. 
  remember (@in_dec PVar pvar_eq y V) as b; destruct b; subst; simpl in *. 
  (*In y V*)
  destruct (IHV _ _ _ _ _ _ i H4) as [K1 K2].  clear IHV H4. 
     rewrite <- K1 in *.  rewrite <- K2 in *. clear K1 K2. 
   split.  unfold upd.  remember (pvar_eq a y) as b; destruct b; subst; simpl in *; trivial.
   unfold upd. 
     remember (pvar_eq a y) as b; destruct b; subst; simpl in *. 
       rewrite id_bbot.
         rewrite <- llub_assoc.
         rewrite lub_refl. trivial.
       trivial.
  (*~In V *)
    destruct MV; try contradiction.  
      subst.  clear IHV. 
      destruct (MergeContinuationNotin _ _ _ _ _ _ _ n H4) as [K1 K2]; clear H4. 
      unfold upd in *; simpl in *. rewrite pvar_eq_refl in *.
        rewrite id_bbot in K2. split; trivial.
Qed. 

Lemma MergeContinuationLemma: forall V s G t D y p
  (Hyp: TntSem (s,G) (mkMergeTCom p V) (t,D)), 
   s y = t y /\ 
   ((D y = llub (G y) p /\ In y V) \/ (D y = G y /\ ~ In y V)). 
Proof. 
intros. 
remember (@in_dec PVar pvar_eq y V) as b; destruct b. 
  destruct (MergeContinuationIn _ _ _ _ _ _ _ i Hyp) as [K1 K2]. 
  split; trivial. apply eq_sym in K2.
  left.  split; trivial. 

  destruct (MergeContinuationNotin _ _ _ _ _ _ _ n Hyp) as [K1 K2]. 
  split; trivial. apply eq_sym in K2. 
  right.  split; trivial.
Qed. 

Lemma MergeContinuationLemma0: forall V s G t D p 
  (Hyp: TntSem (s,G) (mkMergeTCom p V) (t,D)),
   s = t /\ 
   (forall y, (D y = llub (G y) p /\ In y V) \/ (D y = G y /\ ~ In y V)). 
Proof. 
intros. 
split.  apply extensionality.  intros y. 
  remember (@in_dec PVar pvar_eq y V) as b; destruct b. 
    destruct (MergeContinuationIn _ _ _ _ _ _ _ i Hyp) as [K1 K2].  apply K1. 
    destruct (MergeContinuationNotin _ _ _ _ _ _ _ n Hyp) as [K1 K2].  apply K1. 
intros y. 
  remember (@in_dec PVar pvar_eq y V) as b; destruct b. 
    destruct (MergeContinuationIn _ _ _ _ _ _ _ i Hyp) as [K1 K2]. apply eq_sym in K2. left. split; assumption. 
    destruct (MergeContinuationNotin _ _ _ _ _ _ _ n Hyp) as [K1 K2]. apply eq_sym in K2. right. split; trivial. 
Qed. 

Lemma CompCode_exists: forall C p s G, 
exists H,  TntSem (s, G) (CompCode p C) (s,H). 
Proof.
intros.
apply (mergeTntSemexists (modVarsL C)).
Qed.

Lemma CompCode_Notin: forall C s G t D p  
    y (MV : modVars C y = false)
   (Hyp: TntSem (s,G) (CompCode p C) (t,D)),
   s y = t y /\ G y = D y. 
Proof.
intros.
apply modVarsChar' in MV.
apply (MergeContinuationNotin (modVarsL C) _ _ _ _ _ _ MV Hyp).
Qed.

Lemma CompCode_In: forall C s G t D p  
    y (MV : modVars C y = true)
    (Hyp: TntSem (s,G) (CompCode p C) (t,D)),
    s y = t y /\ 
    llub (G y) p = D y.
Proof.
intros.
apply modVarsChar in MV.
eapply MergeContinuationIn. apply MV. apply Hyp.
Qed.

Lemma CompCode_Lemma: forall C s G t D y p
  (Hyp: TntSem (s,G) (CompCode p C) (t,D)), 
   s y = t y /\ 
   ((D y = llub (G y) p /\ modVars C y = true) \/ (D y = G y /\ modVars C y = false)). 
Proof. 
intros.
destruct (MergeContinuationLemma (modVarsL C) _ _ _ _ y p Hyp).
split; trivial.
intros.
destruct H0 as [[HH1 HH2] | [HH1 HH2]].
left. apply modVarsChar in HH2. split; trivial.
right. apply modVarsChar' in HH2. split; trivial.
Qed.

Lemma CompCode_Lemma0: forall C s G t D p 
  (Hyp: TntSem (s,G) (CompCode p C) (t,D)),
   s = t /\ 
   (forall y, (D y = llub (G y) p /\ modVars C y = true) \/ 
              (D y = G y /\ modVars C y = false)). 
Proof. 
intros.
destruct (MergeContinuationLemma0 _ _ _ _ _ _ Hyp).
split; trivial.
intros.
destruct (H0 y) as [[HH1 HH2] | [HH1 HH2]]; clear H0.
left. apply modVarsChar in HH2. split; trivial.
right. apply modVarsChar' in HH2. split; trivial.
Qed.

Lemma modVarsMerge: forall q L x,
modVars (TErase (mkMergeTCom q L)) x = true <-> In x L. 
Proof. 
intros. 
induction L; simpl; split; intros. inversion H.  contradiction.
apply Bool.orb_true_iff in H.
destruct H. apply pvar_eq_ElimT in H. left; trivial.
 right. apply IHL. apply H.
destruct H; subst. 
  remember (pvar_eq x x) as b; destruct b; subst; simpl in *.  trivial. 
    apply False_ind.  apply n.  trivial.
  destruct IHL. rewrite (H1 H). apply Bool.orb_true_intro. right; trivial.
Qed. 

Lemma CompCode_modVars: forall q C,
modVars (TErase (CompCode q C)) = modVars C. 
Proof. 
intros. apply extensionality. intros.
remember (modVars C x). destruct b; apply eq_sym in Heqb.
 apply (modVarsMerge q (modVarsL C) x). apply modVarsChar. trivial.
apply modVarsChar' in Heqb.
  destruct (modVarsMerge q (modVarsL C) x). 
  remember (modVars (TErase (CompCode q C)) x) as bb.
  destruct bb; trivial. 
  unfold CompCode in Heqbb.
  rewrite <- Heqbb in *. 
  exfalso. apply Heqb. apply H. trivial. 
Qed.